(function () {

  const uniqueStylesheetRules = new Set();

  async function gatherExternalStyles() {
    const excludeExternal = document.getElementById('superspeed-exclude-outside-links-used-css') !== null;
    let allStylesheets = '';

    for (let sheet of Array.from(document.styleSheets)) {
      if (
        sheet.ownerNode.id === 'dashicons-css' ||
        sheet.ownerNode.id === 'admin-bar-css' ||
        sheet.ownerNode.id === 'superspeed-wp-ss-tools'
      ) {
        continue;
      }

      if (sheet.href && (!excludeExternal ? sheet.href.startsWith(window.location.origin) : true)) {
        try {
          const response = await fetch(sheet.href);

          if (response.ok) {
            const cssText = await response.text();
            allStylesheets += cssText + '\n';
          } else {
            console.error('Failed to load stylesheet:', sheet.href);
          }
        } catch (error) {
          console.error('Error fetching stylesheet:', sheet.href, '; Error:', error);
        }
      }
    }

    const minifiedCss = minifyCss(allStylesheets);
    const extractedStylesheetRules = extractCssRules(minifiedCss);
    const processedStylesheetRules = processRules(extractedStylesheetRules);
    processedStylesheetRules.forEach(stylesheetRule => {
      uniqueStylesheetRules.add(stylesheetRule);
    });

    return Array.from(uniqueStylesheetRules).join('\n');
  }

  function minifyCss(cssText) {
    cssText = cssText.replace(/\/\*[\s\S]*?\*\//g, '');
    let minified = cssText.replace(/\s+/g, ' ')
      .replace(/;\s*/g, ';')
      .replace(/\s*{\s*/g, '{')
      .replace(/\s*}\s*/g, '}');
    minified = minified.replace(/;(?=\s*})/g, '');
    return minified;
  }

  function extractCssRules(cssText) {
    const stylesheetRules = [];
    const cssRules = cssText.split('}');
    cssRules.forEach(rule => {
      rule = rule.trim();
      if (rule) {
        stylesheetRules.push(rule + '}');
      }
    });
    return stylesheetRules;
  }

  const essentialSelectors = ['*', '*::before', '*:before', '*::after', '*:after', 'html',
    'body', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'ul', 'ol', 'li', 'a', 'header',
    'footer', 'nav', 'section', 'article', 'aside', 'button', 'input', 'optgroup', 'select', 'textarea'];

  function isGenericTagSelector(selector) {
    return /^[a-z]+$/i.test(selector) && !essentialSelectors.includes(selector);
  }

  function isWpAdminBarSelector(selector) {
    return selector.includes('#wpadminbar');
  }

  function extractSelectors(stylesheetRule) {
    if (stylesheetRule.startsWith('@charset')) {
      return []; // Skip @charset rules
    }
    const selectorText = stylesheetRule.split('{')[0].trim();
    return selectorText.split(',').map(selector => selector.trim());
  }

  function processRules(stylesheetRules) {
    const filteredStylesheetRules = new Set();

    for (const stylesheetRule of stylesheetRules) {
      if (stylesheetRule.startsWith('@charset')) {
        continue;
      }

      let cssText = stylesheetRule;
      // Log URLs before processing to ensure they're not affecting CSS variables
      cssText = cssText.replace(/url\((['"]?)(data:image\/svg\+xml,[^'")+|[^'")+)(['"]?)\)/g, (match, quote1, filePath, quote2) => {
        if (filePath.startsWith('data:image/svg+xml,')) {
          return match;
        } else {
          return `url(${quote1}${filePath}${quote2})`;
        }
      });

      const selectors = extractSelectors(cssText);
      if (selectors.length > 0 && selectors.some(selector => !isGenericTagSelector(selector) && !isWpAdminBarSelector(selector))) {
        filteredStylesheetRules.add(cssText);
      }
    }

    return Array.from(filteredStylesheetRules);
  }

  // Start the process of gathering and sending CSS
  (async function () {
    const cssData = await gatherExternalStyles();
    console.log("CSS Data:", cssData);
    await sendCssToServer(cssData);
  })();

  // Send the gathered and minified CSS to the server
  async function sendCssToServer(cssData) {
    const nonce = window.dynamicallyLoadedFilesDataCss.cssNonce;
    const url = window.dynamicallyLoadedFilesDataCss.getUsedCssUrl;
    const pageContext = document.getElementById('used-css-info');
    const templateType = pageContext.getAttribute('data-template-or-slug');

    try {
      const response = await fetch(url, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'X-WP-Nonce': nonce
        },
        credentials: 'same-origin',
        body: JSON.stringify({
          template_type: templateType,
          css: cssData
        })
      });

      if (!response.ok) {
        throw new Error(`Network response was not ok: ${response.statusText}`);
      }

      const data = await response.json();
      console.log('CSS successfully sent to server:', data);
    } catch (error) {
      console.error('Error sending CSS data to server:', error);
    }

    console.log("CSS Data:", cssData);
  }

})();
