
/*
Rules are retrieved from the browser pool.
We get a list of classes that are used by elements on the page.
From all rules, we only keep the ones that match the classes.
*/

// The rules gathered from the browser are sometimes (very rarely) missing a value
// The solution would be to fetch CSS files from the server to process
// That means no cross origin, and using Fetch API


(function () {

  const essentialSelectors = ['*', '*::before', '*:before', '*::after', '*:after', 'html', 'body', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'ul', 'ol', 'li', 'a', 'header', 'footer', 'nav', 'section', 'article', 'aside', 'button', 'input', 'optgroup', 'select', 'textarea'];

  function isGenericTagSelector(selector) {
    return /^[a-z]+$/i.test(selector) && !essentialSelectors.includes(selector);
  }

  function isWpAdminBarSelector(selector) {
    return selector.includes('#wpadminbar');
  }

  function extractSelectors(rule) {
    const selectorGroups = rule.selectorText.split(',');
    return [...new Set(selectorGroups
      .map(selector => selector.trim())
      .flatMap(selector => selector.split(/(?<!\([^)]*)(?<!\[[^\]]*)(?<!\s)(?=[>+~ ])/))
      .filter(selector => {
        const result = selector.includes(':root') || !isGenericTagSelector(selector) && !isWpAdminBarSelector(selector) && selector.trim().length > 0;
        return result;
      }))];
  }

  function isAnySelectorMatching(selectors) {
    const uniqueSelectors = new Set(selectors); // Eliminate duplicates

    return Array.from(uniqueSelectors).some(selector => {
      if (selector.includes(':root')) {
        return true;
      }
      if (!selector.trim() || /^[\s>+~]+$/.test(selector)) {
        return false;
      }

      const isComplex = /[:\[\(]/.test(selector);
      if (isComplex) {
        const hasUnbalancedBrackets = (selector.match(/\[/g) || []).length !== (selector.match(/\]/g) || []).length;
        const hasUnbalancedParentheses = (selector.match(/\(/g) || []).length !== (selector.match(/\)/g) || []).length;
        if (hasUnbalancedBrackets || hasUnbalancedParentheses) {
          return false;
        }
      }

      try {
        return document.querySelectorAll(selector).length > 0;
      } catch (e) {
        return false;
      }
    });
  }

  // Every rule needs to have the file URL asociated with it, so that we can reconstruct shortened URLs for images, and fonts
  function processRules(rules, cssFilePath) {
    const filteredStyles = new Set();  // Using Set to handle unique entries automatically

    Array.from(rules).forEach(rule => {
      let cssText = rule.cssText;

      if (rule.type === CSSRule.STYLE_RULE || rule.type === CSSRule.FONT_FACE_RULE) {
        let selectors = rule.selectorText ? extractSelectors(rule) : '';
        if (!selectors || isAnySelectorMatching(selectors)) {
          // Log URLs before processing to ensure they're not affecting CSS variables
          cssText = cssText.replace(/url\((['"]?)(data:image\/svg\+xml,[^'")]+|[^'")]+)(['"]?)\)/g, (match, quote1, filePath, quote2) => {
            if (filePath.startsWith('data:image/svg+xml,')) {
              return match;
            } else {
              let fullPath = new URL(filePath, cssFilePath).href;
              let newPath = fullPath.indexOf('/wp-content/') !== -1 ? fullPath.substring(fullPath.indexOf('/wp-content/')) : fullPath;
              return `url(${quote1}${newPath}${quote2})`;
            }
          });

          // Specifically handle and log var() usage
          cssText = cssText.replace(/var\((--[^,]+), ([^\)]+)\)/g, (match, variableName, fallbackValue) => {
            return `var(${variableName}, ${fallbackValue})`;
          });

          filteredStyles.add(cssText);  // Add cssText to Set to avoid duplicates
        }
      } else if (rule.type === CSSRule.KEYFRAMES_RULE || rule.type === CSSRule.MEDIA_RULE || rule.type === CSSRule.SUPPORTS_RULE) {
        let subRules = processRules(rule.cssRules, cssFilePath);
        if (subRules.length) {
          let atRuleType = rule.type === CSSRule.KEYFRAMES_RULE ? '@keyframes' :
            rule.type === CSSRule.MEDIA_RULE ? '@media' :
              '@supports';
          filteredStyles.add(`${atRuleType} ${rule.conditionText} { ${subRules.join(' ')} }`);  // Add composite rule to Set
        }
      }

    });

    return Array.from(filteredStyles);  // Convert Set back to Array before returning
  }


  function gatherExternalStyles() {
    let allStyles = new Set();  // Using Set to automatically handle unique entries
    const excludeExternal = document.getElementById('superspeed-exclude-outside-links-used-css') !== null;

    const styleSheets = Array.from(document.styleSheets).filter(sheet => {
      if (sheet.ownerNode.id === 'dashicons-css' ||
        sheet.ownerNode.id === 'admin-bar-css' ||
        sheet.ownerNode.id === 'superspeed-wp-ss-tools'
      ) {
        return false;
      }
      return sheet.href && (!excludeExternal ? sheet.href.startsWith(window.location.origin) : true);
    });

    styleSheets.forEach(sheet => {
      try {
        if (sheet.cssRules) {
          const basePath = sheet.href.substring(0, sheet.href.lastIndexOf('/') + 1);
          let rulesText = processRules(sheet.cssRules, basePath);
          let minifiedCss = minifyCss(rulesText.join(' '));
          allStyles.add(minifiedCss);
        }
      } catch (e) {
        // console.warn("Error accessing stylesheet: ", sheet.href, "; Error: ", e.message);
      }
    });

    return Array.from(allStyles).join('');  // Convert Set back to Array and join into a single string
  }

  function minifyCss(cssText) {
    // Collapse all excess whitespace, and reformat CSS appropriately for minification
    let minified = cssText.replace(/\s+/g, ' ')
      .replace(/;\s*/g, ';')  // Normalize semicolon spacing
      .replace(/\s*{\s*/g, '{')
      .replace(/\s*}\s*/g, '}');

    // Remove the last semicolon in a style definition if it's directly before a }
    minified = minified.replace(/;(?=\s*})/g, '');

    return minified;
  }

  // Start the process of gathering and sending CSS
  const cssData = gatherExternalStyles();
  sendCssToServer(cssData);

  // Send the gathered and minified CSS to the server
  async function sendCssToServer(cssData) {
    const nonce = window.dynamicallyLoadedFilesDataCss.cssNonce;

    const url = window.dynamicallyLoadedFilesDataCss.getUsedCssUrl;
    const pageContext = document.getElementById('used-css-info');
    const templateType = pageContext.getAttribute('data-template-or-slug');

    try {
      const response = await fetch(url, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'X-WP-Nonce': nonce
        },
        credentials: 'same-origin',
        body: JSON.stringify({
          template_type: templateType,
          css: cssData
        })
      });

      if (!response.ok) {
        throw new Error(`Network response was not ok: ${response.statusText}`);
      }

      const data = await response.json();
      console.log('CSS successfully sent to server:', data);
    } catch (error) {
      console.error('Error sending CSS data to server:', error);
    }

  }

  // When all is working well, reload
  // window.location.reload();

  console.log("CSS Data:", cssData);

})();
