(function () {

  const DelayedScripts = {
    normal: [],
    defer: [],
    async: []
  };
  const jQueriesArray = [];

  var DOMLoaded = false;

  //add pageshow listener
  window.addEventListener("pageshow", (e) => {
    window.Persisted = e.persisted;
  });

  //add dom listener and trigger scripts
  function TriggerDOMListener() {

    //add dom listner if page is still loading
    if (document.readyState === 'loading') {
      document.addEventListener("DOMContentLoaded", TriggerDelayedScripts);
    } else {
      //trigger delayed script process
      TriggerDelayedScripts();
    }
  }

  TriggerDOMListener();

  //main delayed script process
  async function TriggerDelayedScripts() {

    //prep
    DelayEventListeners();
    DelayJQueryReady();
    ProcessDocumentWrite();
    SortDelayedScripts();
    PreloadDelayedScripts();

    //load scripts
    await LoadDelayedScripts(DelayedScripts.normal);
    await LoadDelayedScripts(DelayedScripts.defer);
    await LoadDelayedScripts(DelayedScripts.async);

    //trigger delayed DOM events
    await TriggerEventListeners();

    //load delayed styles
    document.querySelectorAll("link[data-delayedstyle]").forEach(function (e) {
      e.setAttribute("href", e.getAttribute("data-delayedstyle"));
    });

  }

  //delay original page event listeners
  function DelayEventListeners() {

    //create event listeners array
    let eventListeners = {};

    //delay dom event
    function delayDOMEvent(object, event) {

      //rewrites event name to trigger later
      function rewriteEventName(eventName) {
        return eventListeners[object].delayedEvents.indexOf(eventName) >= 0 ? "delay-" + eventName : eventName;
      }

      //make sure we haven't added this object yet
      if (!eventListeners[object]) {

        //setup object in eventlisteners array
        eventListeners[object] = {
          originalFunctions: {
            add: object.addEventListener,
            remove: object.removeEventListener
          },
          delayedEvents: []
        }

        //swap delayed events with originals
        object.addEventListener = function () {
          arguments[0] = rewriteEventName(arguments[0]);
          eventListeners[object].originalFunctions.add.apply(object, arguments);
        }
        object.removeEventListener = function () {
          arguments[0] = rewriteEventName(arguments[0]);
          eventListeners[object].originalFunctions.remove.apply(object, arguments);
        }
      }

      //add event to delayed events array for object
      eventListeners[object].delayedEvents.push(event);
    }

    //delay dom event trigger
    function delayDOMEventTrigger(object, event) {
      const originalEvent = object[event];
      Object.defineProperty(object, event, {
        get: !originalEvent ? function () { } : originalEvent,
        set: function (n) {
          object["delay" + event] = n;
        },
      });
    }

    //delay dom events
    delayDOMEvent(document, "DOMContentLoaded");
    delayDOMEvent(window, "DOMContentLoaded");
    delayDOMEvent(window, "load");
    delayDOMEvent(window, "pageshow");
    delayDOMEvent(document, "readystatechange");

    //delay dom event triggers
    delayDOMEventTrigger(document, "onreadystatechange");
    delayDOMEventTrigger(window, "onload");
    delayDOMEventTrigger(window, "onpageshow");
  }

  //delay jquery ready
  function DelayJQueryReady() {

    //store original jquery
    let originalJQuery = window.jQuery;

    //modify original jquery object
    Object.defineProperty(window, "jQuery", {

      //return original when accessed directly
      get() {
        return originalJQuery;
      },

      //modify value when jquery is set
      set(newJQuery) {

        //make sure it's valid and we haven't modified it already
        if (newJQuery && newJQuery.fn && !jQueriesArray.includes(newJQuery)) {

          //modify new jquery ready event
          newJQuery.fn.ready = newJQuery.fn.init.prototype.ready = function (originalJQuery) {

            //dom loaded, go ahead
            if (DOMLoaded) {
              originalJQuery.bind(document)(newJQuery);
            }

            //dom not loaded, so wait for listener
            else {
              document.addEventListener("delay-DOMContentLoaded", function () {
                originalJQuery.bind(document)(newJQuery);
              });
            }
          };

          //store on event
          const newJQueryOn = newJQuery.fn.on;

          //modify new jquery on event
          newJQuery.fn.on = newJQuery.fn.init.prototype.on = function () {

            if (this[0] === window) {

              //rewrite event name
              function rewriteEventName(eventName) {
                eventName = eventName.split(" ");
                eventName = eventName.map(function (name) {
                  if (name === "load" || name.indexOf("load.") === 0) {
                    return "delay-jquery-load";
                  } else {
                    return name;
                  }
                });
                eventName = eventName.join(" ");

                return eventName;
              }

              //rewrite event name/s
              if (typeof arguments[0] == "string" || arguments[0] instanceof String) {
                arguments[0] = rewriteEventName(arguments[0]);
              } else if (typeof arguments[0] == "object") {
                Object.keys(arguments[0]).forEach(function (argument) {
                  delete Object.assign(arguments[0], {
                    [rewriteEventName(argument)]: arguments[0][argument]
                  })[argument];
                });
              }
            }
            return newJQueryOn.apply(this, arguments), this;
          };

          //add modified jquery to storage array
          jQueriesArray.push(newJQuery);
        }

        //replace original jquery with modified version
        originalJQuery = newJQuery;
      }
    });
  }

  //print document write values directly after their parent script
  function ProcessDocumentWrite() {

    //create map to store scripts
    const map = new Map();

    //modify document.write functions
    document.write = document.writeln = function (value) {

      //prep
      var script = document.currentScript;
      var range = document.createRange();

      //make sure script isn't in map yet
      let mapScript = map.get(script);
      if (mapScript === void 0) {

        //add script's next sibling to map
        mapScript = script.nextSibling;
        map.set(script, mapScript);
      }

      //insert value before script's next sibling
      var fragment = document.createDocumentFragment();
      range.setStart(fragment, 0);
      fragment.appendChild(range.createContextualFragment(value));
      script.parentElement.insertBefore(fragment, mapScript);
    };
  }

  //find all delayed scripts and sort them by load order
  function SortDelayedScripts() {
    document.querySelectorAll('script[type="ss-delay-script"]').forEach(function (event) {
      if (event.hasAttribute("src")) {
        if (event.hasAttribute("defer") && event.defer !== false) {
          DelayedScripts.defer.push(event);
        } else if (event.hasAttribute("async") && event.async !== false) {
          DelayedScripts.async.push(event);
        } else {
          DelayedScripts.normal.push(event);
        }
      } else {
        DelayedScripts.normal.push(event);
      }
    });
  }

  //add block of preloads for delayed scripts that have src URLs
  function PreloadDelayedScripts() {
    var preloadFragment = document.createDocumentFragment();
    [...DelayedScripts.normal, ...DelayedScripts.defer, ...DelayedScripts.async].forEach(function (script) {
      var src = script.getAttribute("src");
      if (src) {
        var link = document.createElement("link");
        link.href = src;
        link.rel = "preload";
        link.as = "script";
        preloadFragment.appendChild(link);
      }
    });
    document.head.appendChild(preloadFragment);
  }

  //load array of delayed scripts one at a time
  async function LoadDelayedScripts(scripts) {

    //grab first script in array
    var script = scripts.shift();

    //replace script and move to the next one
    if (script) {
      await ReplaceScript(script);
      return LoadDelayedScripts(scripts);
    }

    //resolve when all scripts have been swapped
    return Promise.resolve();
  }

  //replace delayed script in document
  async function ReplaceScript(script) {

    //wait
    await NextFrame();

    //create new script and replace
    return new Promise(function (replaceScript) {

      //prep
      const newscript = document.createElement("script");

      //loop through script attributes
      [...script.attributes].forEach(function (attribute) {
        let attributeName = attribute.nodeName;

        if (attributeName !== "type") {

          //swap data-type if needed
          if (attributeName === "data-type") {
            attributeName = "type";
          }

          //add attribute to newscript
          newscript.setAttribute(attributeName, attribute.nodeValue);
        }
      });

      //src script
      if (script.hasAttribute("src")) {
        newscript.addEventListener("load", replaceScript);
        newscript.addEventListener("error", replaceScript);
      }

      //inline script
      else {
        newscript.text = script.text;
        replaceScript();
      }

      //replace original script with final
      script.parentNode.replaceChild(newscript, script);
    });
  }

  //trigger delayed event listeners after scripts have loaded
  async function TriggerEventListeners() {

    //set flag
    DOMLoaded = true;
    await NextFrame();

    //trigger events
    document.dispatchEvent(new Event("delay-DOMContentLoaded"));
    await NextFrame();
    window.dispatchEvent(new Event("delay-DOMContentLoaded"));
    await NextFrame();
    document.dispatchEvent(new Event("delay-readystatechange"));
    await NextFrame();
    if (document.delayonreadystatechange) {
      document.delayonreadystatechange();
    }
    await NextFrame();
    window.dispatchEvent(new Event("delay-load"));
    await NextFrame();
    if (window.delayonload) {
      window.delayonload();
    }
    await NextFrame();
    jQueriesArray.forEach(function (singleJQuery) {
      singleJQuery(window).trigger("delay-jquery-load")
    });
    const PageShowEvent = new Event("delay-pageshow");
    PageShowEvent.persisted = window.Persisted;
    window.dispatchEvent(PageShowEvent);
    await NextFrame();
    if (window.delayonpageshow) {
      window.delayonpageshow({
        persisted: window.Persisted
      });
    }
  }

  //wait for next frame before proceeding
  async function NextFrame() {
    return new Promise(function (e) {
      requestAnimationFrame(e);
    });
  }

  function RenameDOMAttribute(e, t, n) {
    if (e.hasAttribute && e.hasAttribute(t)) {
      e.setAttribute(n, e.getAttribute(t));
      e.removeAttribute(t);
    }
  }


})();