
/** Get info about LCP and display it **/


(function () {

  const screenWidth = window.innerWidth || document.documentElement.clientWidth || document.body.clientWidth;
  let lcpElement = null;
  let entry = null;
  let LcpPredetermined = false;
  let lastKnownPosition = null;
  let animationFrameId = null;

  if (screenWidth >= 1024) {
    lcpElement = document.querySelector('[is_lcp_desktop]');
  } else if (screenWidth >= 768) {
    lcpElement = document.querySelector('[is_lcp_tablet]');
  } else {
    lcpElement = document.querySelector('[is_lcp_mobile]');
  }

  if (!lcpElement) {

    // Debounce function remains unchanged
    function debounce(func, wait, immediate) {
      let timeout;
      return function () {
        const context = this, args = arguments;
        const later = function () {
          timeout = null;
          if (!immediate) func.apply(context, args);
        };
        const callNow = immediate && !timeout;
        clearTimeout(timeout);
        timeout = setTimeout(later, wait);
        if (callNow) func.apply(context, args);
      };
    }

    // Process LCP Entry with Debounce remains unchanged
    const debouncedProcessLCPEntry = debounce(processLCPEntry, 250);

    if ('PerformanceObserver' in window) {
      const observer = new PerformanceObserver((list) => {
        for (entry of list.getEntries()) {
          debouncedProcessLCPEntry(entry);
        }
      });
      observer.observe({ type: "largest-contentful-paint", buffered: true });

      // Delayed disconnection to ensure LCP events near the load time are captured
      function delayedDisconnectObserver() {
        setTimeout(() => {
          observer.disconnect();
          // console.log("Observer disconnected after ensuring all LCP entries are processed.");
        }, 0); // Makes the function sincronous
      }

      // Attach to load event or execute immediately if already loaded
      if (document.readyState === 'complete') {
        delayedDisconnectObserver();
      } else {
        window.addEventListener("load", delayedDisconnectObserver);
      }
    } else {
      console.warn("PerformanceObserver not supported in this browser.");
    }

  } else {
    LcpPredetermined = true;
    processLCPEntry(entry);
  }

  // Your existing processLCPEntry and related functions remain unchanged
  function processLCPEntry(entry) {

    if (LcpPredetermined) {
      if (lcpElement) {
        const rect = lcpElement.getBoundingClientRect();
        const x = rect.left + rect.width / 2;
        const y = rect.top + rect.height / 2;
        entry = { x, y };
      }
    } else {
      lcpElement = entry.element || document.elementFromPoint(entry.x, entry.y);
    }

    if (lcpElement) {
      // List of classes to exclude
      const excludeClasses = [
        'img-gray-placeholder',
        'ss-wrap-iframe-placeholder',
        'ss-iframe-placeholder',
        'ss-video-placeholder',
        'ss-allow-bg',
        'original-bg-loaded',
        'ss-bg-zero-opacity',
        'ss-fade-in-bg',
        'bg-lazy-loaded',
        'background-auto-set'
      ];

      const filteredClasses = Array.from(lcpElement.classList).filter(className => !excludeClasses.includes(className)).join('</span><span class="lcp-info-span">');
      const elementType = determineElementType(lcpElement);

      let htmlContent = '';

      if (elementType === "Text Block") {
        // Constructing the HTML string
        htmlContent = `
          <div id="lcp-info-container">
            <div class="lcp-info lcp-title-btn">
              <span class="lcp-title">Largest Contentful Paint (LCP)</span>
              <button class="lcp-close-button" onclick="document.querySelector('#lcp-info-container').remove();">+</button>
            </div>
            <div class="lcp-info">Type: ${elementType}</div>
          </div>
        `;
      } else {
        htmlContent = `
          <div id="lcp-info-container">
            <div class="lcp-info lcp-title-btn">
              <span class="lcp-title">Largest Contentful Paint (LCP)</span>
              <button class="lcp-close-button" onclick="document.querySelector('#lcp-info-container').remove();">+</button>
            </div>
            <div class="lcp-info">Type: ${elementType}</div>
            <div class="lcp-info-classes">Classes:<span class="lcp-info-span">${filteredClasses}</span></div>
          </div>
        `;
      }


      document.body.insertAdjacentHTML('beforeend', htmlContent);

      const infoContainer = document.querySelector("#lcp-info-container");
      positionInfoContainer(infoContainer, lcpElement);

      // Remove the CSS that hides the admin toolbar
      removeHelperStyles();
    }
  }


  function determineElementType(element) {
    if (element.tagName === "IMG") {
      return "Image";
    } else {
      const computedStyle = window.getComputedStyle(element);
      if (computedStyle.backgroundImage !== "none") {
        return "Background Image";
      } else {
        return "Text Block";
      }
    }
  }

  function positionInfoContainer(container, element) {
    let rect;

    if (LcpPredetermined) {
      rect = element.getBoundingClientRect();
    } else {
      rect = lastKnownPosition || element.getBoundingClientRect();
    }

    container.style.top = `${rect.top + window.scrollY}px`;
    container.style.left = `${rect.left + window.scrollX}px`;
    lastKnownPosition = rect;

    // Start observing for DOM changes
    observeDOMChanges(container, element);

    // Start the loop to periodically check the element's position
    startPeriodicCheck(container, element);
  }

  // Sets up MutationObserver to monitor for changes in the DOM
  // Modified observeDOMChanges to batch updates with requestAnimationFrame
  function observeDOMChanges(container, targetElement) {
    let scheduled = false;
    const observer = new MutationObserver((mutationsList, observer) => {
      if (!scheduled) {
        scheduled = true;
        requestAnimationFrame(() => {
          for (const mutation of mutationsList) {
            if (mutation.type === 'childList' || mutation.type === 'attributes') {
              const newRect = targetElement.getBoundingClientRect();
              container.style.top = `${newRect.top + window.scrollY}px`;
              container.style.left = `${newRect.left + window.scrollX}px`;
            }
          }
          scheduled = false;
        });
      }
    });

    const config = { attributes: true, childList: true, subtree: true };
    observer.observe(document.body, config);
  }


  // Modified startPeriodicCheck to optimize performance
  function startPeriodicCheck(container, targetElement) {
    function checkPosition() {
      const currentRect = targetElement.getBoundingClientRect();
      if (lastKnownPosition.top !== currentRect.top || lastKnownPosition.left !== currentRect.left) {
        requestAnimationFrame(() => {
          container.style.top = `${currentRect.top + window.scrollY}px`;
          container.style.left = `${currentRect.left + window.scrollX}px`;
          lastKnownPosition = currentRect;
        });
      }
      animationFrameId = requestAnimationFrame(checkPosition);
    }

    checkPosition();
  }


  // Remove the CSS that hides the admin toolbar
  function removeHelperStyles() {
    const styleElement = document.getElementById("superspeed-while-getting-lcp-info");
    if (styleElement) styleElement.remove();
  }

})();


