/*
-	Find all elements that have a background image, added through CSS, via the style attribute, inline or external file.
-	Send to the database the following info:
- * Each element’s custom attribute value
- * The URL of the <img> element
- * The URL of the background image for the screen size
- * LCP for the screen size
- * Visibility for the screen size
*/


(async function () {

  // Run the script only for an admin, author or editor
  const isEditor = document.querySelector('body.can-edit');
  if (!isEditor) {
    return;
  }

  const screenWidth = window.innerWidth || document.documentElement.clientWidth || document.body.clientWidth;
  const isInIframe = (window !== window.top);

  // Don't run the code on a small screen, unless it's inside an iframe
  if (screenWidth < 1023 && !isInIframe) {
    return;
  }

  // Do not scroll back to a point where the page was, before reloading
  if ('scrollRestoration' in history) {
    history.scrollRestoration = 'manual';
  }

  // Scroll to top
  window.scrollTo(0, 0);

  // Prevent scrolling by adding CSS
  const styleElement = document.createElement('style');
  styleElement.id = "while-getting-lcp";
  styleElement.textContent = `
      div#wpadminbar, html body.can-edit #page-info-wrap:has(#superspeed-page-info .page-info-item) {
          display: none !important;
      }
      html {
          margin-top: unset !important;
          height: 100% !important;
          overflow: hidden !important;
      }
      * {
      pointer-events: none !important;
      }
    `;
  document.head.appendChild(styleElement);
  document.body.appendChild(styleElement);

  // Scroll to top again
  window.scrollTo(0, 0);

  function debounce(fn, delay) {
    let timeoutID = null;
    return function () {
      clearTimeout(timeoutID);
      timeoutID = setTimeout(() => {
        fn.apply(this, arguments);
      }, delay);
    };
  }

  let screenSize;
  if (screenWidth >= 1024) {
    screenSize = 'Desktop';
  } else if (screenWidth >= 768) {
    screenSize = 'Tablet';
  } else {
    screenSize = 'Mobile';
  }

  const parentDocument = window.parent.document;
  const windowParentLocation = window.parent.location;
  const pageURL = window.location.href;

  // Inform the user that we are taking data, and ask for patience
  const loading = document.createElement('div');
  loading.id = "getting-lcp-loading";
  loading.textContent = `Getting images data for ${screenSize}...`; // Include screenSize in the message
  // Set all CSS styles using cssText
  loading.style.cssText = `
    position: fixed;
    top: 50%;
    left: 50%;
    transform: translate(-50%, -50%);
    color: white;
    font-size: clamp(13px, 2.5vw, 22px);
    padding: 0.5em 1em;
    border-radius: 0.5em;
    box-shadow: 10px 11px 50px black;
    opacity: 0.6;
    background: #003b87;
    z-index: 2147483647;
    font-family: system-ui;
`;
  // Append the loading element to the document's body
  document.body.appendChild(loading);

  /*
  function removeLoadingNotification() {
    // Access the loading element by its ID or other identifier in the parent document
    const loading = parentDocument.getElementById('getting-lcp-loading');
    // Check if the loading element exists and remove it if it does
    if (loading) {
      loading.remove();
    }
  }
  */

  let largestPaintEntry = 0;
  let largestSize = 0; // Initialize to track the largest entry size
  let lastEntryTimeout;
  let observer; // Define observer in a higher scope

  function processLCPEntry(entry) {
    if (entry.size > largestSize) {
      largestSize = entry.size;
      largestPaintEntry = entry;
    }

    // Clear any previously set timeout
    clearTimeout(lastEntryTimeout);

    // Set a new timeout to check if this is the last entry
    lastEntryTimeout = setTimeout(() => {
      if (observer) { // Ensure observer is defined before calling disconnect
        observer.disconnect(); // Safe to call disconnect
        observer = null; // Nullify observer to prevent future use after disconnection
        // console.log("Observer disconnected, all LCP entries processed.");
        collectAndSendElementDataDebounced();
      }
    }, 250); // The timeout should be slightly longer than the debounce interval
  }

  const debouncedProcessLCPEntry = debounce(processLCPEntry, 200);
  const collectAndSendElementDataDebounced = debounce(collectAndSendElementData, 50);

  if ('PerformanceObserver' in window) {
    // console.log("PerformanceObserver is supported in this browser.");
    observer = new PerformanceObserver((list) => {
      // console.log("new PerformanceObserver");
      for (const entry of list.getEntries()) {
        // console.log('PerformanceObserver entry: ', entry);
        debouncedProcessLCPEntry(entry);
      }
    });

    observer.observe({ type: 'largest-contentful-paint', buffered: true });

    // If the document has already loaded, check immediately
    if (document.readyState === 'complete') {
      clearTimeout(lastEntryTimeout);
      lastEntryTimeout = setTimeout(() => {
        if (observer) { // Check if observer is still defined
          observer.disconnect();
          observer = null; // Prevent further use
          // console.log("Observer disconnected, potentially missed LCP entries due to late check.");
        }
        collectAndSendElementDataDebounced();
      }, 250);
    } else {
      window.addEventListener("load", () => {
        clearTimeout(lastEntryTimeout);
        lastEntryTimeout = setTimeout(() => {
          if (observer) { // Check if observer is still defined
            observer.disconnect();
            observer = null; // Prevent further use
            // console.log("Observer disconnected, potentially missed LCP entries due to late check.");
          }
          collectAndSendElementDataDebounced();
        }, 250);
      });
    }
  } else {
    // console.warn("PerformanceObserver not supported in this browser.");
  }


  function isVisible(el) {
    if (!el) return false;
    const rect = el.getBoundingClientRect();
    const style = window.getComputedStyle(el);
    return rect.width > 0 && rect.height > 0 && style.display !== 'none' && style.visibility !== 'hidden' && (!el.parentElement || isVisible(el.parentElement));
  }

  function createDataScriptElement(data) {
    const scriptElement = document.createElement('script');
    scriptElement.type = 'application/json';
    scriptElement.classList.add('keep-images-data');
    scriptElement.textContent = JSON.stringify(data);
    parentDocument.body.appendChild(scriptElement);
  }

  function createEmptyScriptElementFlag() {
    if (isInIframe) {
      const scriptElement = document.createElement('script');
      if (screenWidth > 470) {
        scriptElement.id = 'keep-images-data-tablet';
      } else {
        scriptElement.id = 'keep-images-data-mobile';
      }
      parentDocument.body.appendChild(scriptElement);
    }
  }


  // Array of excluded classes
  const excludedClasses = [
    'img-gray-placeholder',
    'ss-wrap-iframe-placeholder',
    'ss-iframe-placeholder',
    'ss-video-placeholder',
    'ss-allow-bg',
    'original-bg-loaded',
    'ss-bg-zero-opacity',
    'ss-fade-in-bg',
    'bg-lazy-loaded',
    'background-auto-set'
  ];

  // Helper function to get the selector for the element
  function getElementSelector(el) {
    // Get the element's ID, if it exists
    const id = el.id ? `#${el.id}` : '';

    // Get all classes of the element, excluding the specified ones
    const classes = Array.from(el.classList)
      .filter(cls => !excludedClasses.includes(cls))
      .map(cls => `.${cls}`)
      .join('');

    // Construct the selector with ID and classes
    return `${id}${classes}`;
  }

  // Create a URL based selecotr for images
  function getImageSelector(el) {
    if (el.tagName.toLowerCase() === 'img' && el.src) {
      // Extract the base part of the filename, assuming the format is consistent as per the given examples
      const baseNameMatch = el.src.match(/\/([^\/]+?)(-\d+x\d+)?\.\w+$/);
      if (baseNameMatch) {
        const baseName = baseNameMatch[1]; // This captures the primary name of the image without dimensions
        return `[src*='/${baseName}']`; // Omit the file extension for a more generic selector
      }
    }
    return ''; // Return empty if no suitable selector could be formulated
  }

  // Create a URL based selecotr for background images
  function getBackgroundImageSelector(el) {
    // Extract URL from style attribute if it exists
    if (window.getComputedStyle(el).backgroundImage) {
      const style = el.style.backgroundImage || el.style.background;
      if (style) {
        const urlMatch = style.match(/url\(["']?(.+?)["']?\)/);
        if (urlMatch && urlMatch[1]) {
          // Extract the base part of the filename from the URL
          const baseNameMatch = urlMatch[1].match(/\/([^\/]+?)(-\d+x\d+)?(\.\w+)$/);
          if (baseNameMatch) {
            const baseName = baseNameMatch[1]; // This captures the primary name of the image without dimensions
            return `[style*='/${baseName}']`; // Use the base name in a style attribute selector
          }
        }
      }
    }
    return ''; // Return empty if no suitable selector could be formulated
  }

  // const extractUrl = (bgImage) => bgImage.match(/url\("?(.+?)"?\)/)?.[1] || null;

  function extractBackgroundImageData(backgroundImage) {
    const matches = backgroundImage.match(/url\((['"]?)(.*?)\1\)/i);
    if (matches && matches[2]) {
      return {
        url: matches[2],
        // Checking if the background image is mixed in with an overlay color or gradient
        overlay: backgroundImage.includes('linear-gradient') ? backgroundImage : null,
      };
    }
    return { url: null, overlay: null };
  }

  async function collectAndSendElementData() {
    const elements = document.querySelectorAll('img, div, section, article, header, footer, aside, h1, h2, p, span');


    elements.forEach(el => {
      if (document.getElementById('wpadminbar')?.contains(el) || !isVisible(el)) return;
    
      const isImage = el.tagName.toLowerCase() === 'img';
      const backgroundImageData = extractBackgroundImageData(window.getComputedStyle(el).backgroundImage);
      const hasBgImage = !!backgroundImageData.url;
    
      // Check if the image source or background image URL is base64 or SVG
      if ((isImage && (el.src.includes('data:image') || el.src.includes('.svg'))) ||
        (hasBgImage && (backgroundImageData.url.includes('data:image') || backgroundImageData.url.includes('.svg')))) {
        return; // Skip this element and continue with the next in the loop
      }
    
      const isLcpElement = largestPaintEntry && el === largestPaintEntry.element;
    
      let createSelector = getElementSelector(el) + getImageSelector(el) + getBackgroundImageSelector(el);
      if (!createSelector) {
        return;
      }
    
      // If the LCP element is an image or has a background image, or is an image, or is a background image
      if (isLcpElement && isImage || isLcpElement && hasBgImage || isImage || hasBgImage) {
        const data = {
          selector: createSelector,
          isImage: isImage ? 1 : 0,
          url: isImage ? el.src : backgroundImageData.url || 'null',
          overlay: backgroundImageData.overlay || 0,
          IsLcp: isLcpElement ? 1 : 0,
          VisibleOn: isVisible(el) ? screenSize : 0,
        };
        createDataScriptElement(data);
      }
    });

    /*
    elements.forEach(el => {
      if (document.getElementById('wpadminbar')?.contains(el) || !isVisible(el)) return;

      const isImage = el.tagName.toLowerCase() === 'img';
      const bgImageUrl = extractUrl(window.getComputedStyle(el).backgroundImage);
      const hasBgImage = !!bgImageUrl;

      // Check if the image source or background image URL is base64 or SVG
      if ((isImage && (el.src.includes('data:image') || el.src.includes('.svg'))) ||
        (hasBgImage && (bgImageUrl.includes('data:image') || bgImageUrl.includes('.svg')))) {
        return; // Skip this element and continue with the next in the loop
      }

      const isLcpElement = largestPaintEntry && el === largestPaintEntry.element;

      let createSelector = getElementSelector(el) + getImageSelector(el) + getBackgroundImageSelector(el)
      if (!createSelector) {
        return;
      }

      // If the LCP element is an image or has a background image, or is an image, or is a background image
      if (isLcpElement && isImage || isLcpElement && hasBgImage || isImage || hasBgImage) {
        const data = {
          selector: createSelector,
          isImage: isImage ? 1 : 0,
          url: isImage ? el.src : bgImageUrl || 'null',
          IsLcp: isLcpElement ? 1 : 0,
          VisibleOn: isVisible(el) ? screenSize : 0,
        };
        createDataScriptElement(data);
      }
    });
*/
    createEmptyScriptElementFlag();

    if (!isInIframe && !document.getElementById('iframeForMobile') && !document.getElementById('iframeForTablet')) {

      createIframeForMobile();
      createIframeForTablet();
    }

    if (isInIframe && parentDocument.getElementById('keep-images-data-tablet') && parentDocument.getElementById('keep-images-data-mobile')) {

      const collectedData = collectDataFromScriptElements();
      if (collectedData.length) {
        const normalizeUrl = (url) => {
          return url.toLowerCase().replace(/^www\./, '').split('#')[0].split('?')[0];
        };

        const pageURL = normalizeUrl(windowParentLocation.href);
        const data = { pageUrl: pageURL, elements: collectedData };

        await sendToServer(data);
      }
    }
  }


  function collectDataFromScriptElements() {

    let dataElements;
    if (!isInIframe) {
      dataElements = document.getElementsByClassName('keep-images-data');
    } else {
      dataElements = parentDocument.getElementsByClassName('keep-images-data');
    }
    const collectedData = Array.from(dataElements).map(el => JSON.parse(el.textContent));
    return collectedData;
  }

  // Send data to server, now with checks in place
  async function sendToServer(data) {
    const nonce = window.dynamicallyLoadedFilesDataImages.imagesNonce;
    const url = window.dynamicallyLoadedFilesDataImages.getImagesDataUrl;

    // console.log(`Sending images data to server...\nURL: ${url}\nNonce: ${nonce}\nData:`, JSON.stringify(data, null, 2));

    try {
      const response = await fetch(url, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'X-WP-Nonce': nonce,
        },
        credentials: 'same-origin',
        body: JSON.stringify(data),
      });

      if (!response.ok) {
        throw new Error(`Network response was not ok: ${response.statusText}`);
      }

      // const responseBody = await response.json(); // Assumes the server always responds with JSON on success
      // console.log('Server response:', JSON.stringify(responseBody, null, 2)); // Log the JSON response in a readable format

      // console.log('Images data successfully sent to server');
    } catch (error) {
      // console.error('Error sending data to server:', error);
    }

    windowParentLocation.reload();
    // removeLoadingNotification();
    // removeIframesFromParent();
  }

  function createIframeForTablet() {
    const iframe = document.createElement('iframe');
    iframe.id = 'iframeForTablet';
    iframe.src = pageURL;
    iframe.style.cssText = 'height: 640px; width: 800px; border: none; position: fixed; top: 0; left: 0; z-index: 2147483647;';
    document.body.appendChild(iframe);
  }

  function createIframeForMobile() {
    const iframe = document.createElement('iframe');
    iframe.id = 'iframeForMobile';
    iframe.src = pageURL;
    iframe.style.cssText = 'width: 470px; height: 100%; border: none; position: fixed; top: 0; right: 0; z-index: 2147483647;';
    document.body.appendChild(iframe);
  }

  /*
  async function removeIframesFromParent() {

    const styleElement = parentDocument.getElementById("while-getting-lcp");
    if (styleElement) styleElement.remove();

    const iframeIDs = ['iframeForTablet', 'iframeForMobile'];
    iframeIDs.forEach(iframeID => {
      const iframe = parentDocument.getElementById(iframeID);
      if (iframe) iframe.parentNode.removeChild(iframe);
    });
  }
*/

})();