
/*
Rules are retrieved from the browser pool.
We get a list of classes that are used by elements on the page.
From all rules, we only keep the ones that match the classes.
*/

// The rules gathered from the browser are sometimes (very rarely) missing a value
// The solution would be to fetch CSS files from the server to process
// That means no cross origin, and using Fetch API


(function () {

  const screenWidth = window.innerWidth || document.documentElement.clientWidth || document.body.clientWidth;

  // Don't run the code on a small screen
  if (screenWidth < 900) {
    return;
  }

  const essentialSelectors = ['*', 'html', 'body', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6',
    'p', 'a', 'header', 'footer', 'nav', 'section', 'article', 'aside', 'button',
    'input', 'optgroup', 'select', 'textarea', 'label', 'table', 'thead', 'tbody',
    'tr', 'th', 'td', 'form', 'hr', 'strong', 'em', 'b', 'i', 'pre', 'code', 'ul',
    'ol', 'li', 'dl', 'dt', 'dd'];

  function isGenericTagSelector(selector) {
    return /^[a-z]+$/i.test(selector) && !essentialSelectors.includes(selector);
  }

  function isWpAdminBarSelector(selector) {
    return selector.includes('#wpadminbar');
  }


  function extractSelectors(rule) {
    const selectorGroups = rule.selectorText.split(/,(?![^(]*\)|\[.*?\])/);
    const extractedSelectors = [...new Set(selectorGroups
      .map(selector => selector.trim())
      .flatMap(selector => {
        const baseSelectors = selector.split(/(?<!\([^)]*)(?<!\[[^\]]*)(?<!\s)(?=[>+~ ])/);
        return baseSelectors.flatMap(baseSelector => {
          const hasFocus = baseSelector.includes(':focus');
          const hasHover = baseSelector.includes(':hover');
          const hasPlaceholder = baseSelector.includes('::placeholder');
          const baseWithoutPseudoClasses = hasPlaceholder ? '::placeholder' : baseSelector.replace(/:(?:focus|hover)/g, '');

          const selectors = [baseWithoutPseudoClasses];
          if (hasHover) selectors.push(`${baseWithoutPseudoClasses}:hover`);
          if (hasFocus) selectors.push(`${baseWithoutPseudoClasses}:focus`);

          return selectors;
        });
      })
      .filter(selector => {
        const result = selector.includes(':root') || !isGenericTagSelector(selector) && !isWpAdminBarSelector(selector) && selector.trim().length > 0;
        return result;
      }))];
    return extractedSelectors;
  }


  function isAnySelectorMatching(selectors) {
    const uniqueSelectors = new Set(selectors); // Eliminate duplicates

    return Array.from(uniqueSelectors).some(selector => {
      if (selector.includes(':root')) {
        return true;
      }
      if (!selector.trim() || /^[\s>+~]+$/.test(selector)) {
        return false;
      }

      const isComplex = /[:\[\(]/.test(selector);
      if (isComplex) {
        const hasUnbalancedBrackets = (selector.match(/\[/g) || []).length !== (selector.match(/\]/g) || []).length;
        const hasUnbalancedParentheses = (selector.match(/\(/g) || []).length !== (selector.match(/\)/g) || []).length;
        if (hasUnbalancedBrackets || hasUnbalancedParentheses) {
          return false;
        }
      }


      const isAttributeSelector = /\[.*?\]/.test(selector);
      const hasPseudoElement = /::(?:before|after|placeholder)/.test(selector);
      const isVendorSpecificPseudoElement = /::(?:-webkit-|-moz-|-ms-|-o-)?(?:scrollbar|selection)/.test(selector);
      const hasFocusPseudoClass = /:focus/.test(selector);
      const hasHoverPseudoClass = /:hover/.test(selector);
      const isPlaceholderSelector = selector === '::placeholder';

      if (isVendorSpecificPseudoElement || isPlaceholderSelector) {
        return true;
      } else if (isAttributeSelector || hasPseudoElement || hasFocusPseudoClass || hasHoverPseudoClass) {
        const baseSelector = selector.replace(/\[.*?\]|::(?:before|after)|:(?:focus|hover)/g, '');
        try {
          if (isAttributeSelector) {
            const attributeName = selector.match(/\[(.*?)\]/)[1];
            return document.querySelectorAll(`[${attributeName}]`).length > 0;
          } else {
            return document.querySelectorAll(baseSelector).length > 0;
          }
        } catch (e) {
          return false;
        }
      } else {
        try {
          return document.querySelectorAll(selector).length > 0;
        } catch (e) {
          return false;
        }
      }
    });
  }



  function processRules(rules, cssFilePath, mediaQuery = '') {
    const filteredStyles = [];

    Array.from(rules).forEach(rule => {
      let cssText = rule.cssText;

      if (rule.type === CSSRule.STYLE_RULE || rule.type === CSSRule.FONT_FACE_RULE) {
        let selectors = rule.selectorText ? extractSelectors(rule) : '';
        if (!selectors || isAnySelectorMatching(selectors)) {
          // Log URLs before processing to ensure they're not affecting CSS variables
          cssText = cssText.replace(/url\((['"]?)(data:image\/svg\+xml,[^'")]+|[^'")]+)(['"]?)\)/g, (match, quote1, filePath, quote2) => {
            if (filePath.startsWith('data:image/svg+xml,')) {
              return match;
            } else {
              let fullPath = new URL(filePath, cssFilePath).href;
              let newPath = fullPath.indexOf('/wp-content/') !== -1 ? fullPath.substring(fullPath.indexOf('/wp-content/')) : fullPath;
              return `url(${quote1}${newPath}${quote2})`;
            }
          });

          // Specifically handle and log var() usage
          cssText = cssText.replace(/var\((--[^,]+), ([^\)]+)\)/g, (match, variableName, fallbackValue) => {
            return `var(${variableName}, ${fallbackValue})`;
          });

          if (mediaQuery) {
            filteredStyles.push(`@media ${mediaQuery} { ${cssText} }`);
          } else {
            filteredStyles.push(cssText);
          }
        }
      } else if (rule.type === CSSRule.KEYFRAMES_RULE || rule.type === CSSRule.MEDIA_RULE || rule.type === CSSRule.SUPPORTS_RULE) {
        let subRules;
        let subMediaQuery = mediaQuery;

        if (rule.type === CSSRule.MEDIA_RULE) {
          subMediaQuery = mediaQuery ? `${mediaQuery} and ${rule.conditionText}` : rule.conditionText;
          subRules = processRules(rule.cssRules, cssFilePath, subMediaQuery);
        } else {
          subRules = processRules(rule.cssRules, cssFilePath, mediaQuery);
        }

        if (subRules.length) {
          let atRuleType = rule.type === CSSRule.KEYFRAMES_RULE ? '@keyframes' :
            rule.type === CSSRule.MEDIA_RULE ? '@media' :
              '@supports';
          filteredStyles.push(`${atRuleType} ${rule.conditionText} { ${subRules.join(' ')} }`);
        }
      }
    });

    return filteredStyles;
  }


  function gatherExternalStyles() {
    let allStyles = []; // Using an array to maintain the order of styles

    const styleSheets = Array.from(document.styleSheets).filter(sheet => {
      if (sheet.ownerNode.id === 'dashicons-css' ||
        sheet.ownerNode.id === 'admin-bar-css' ||
        sheet.ownerNode.id === 'superspeed-wp-ss-tools'
      ) {
        return false;
      }
      return sheet.href;
    });

    styleSheets.forEach(sheet => {
      try {
        if (sheet.cssRules) {
          const basePath = sheet.href.substring(0, sheet.href.lastIndexOf('/') + 1);
          const mediaQuery = sheet.media.mediaText;
          let rulesText = processRules(sheet.cssRules, basePath, mediaQuery);
          let minifiedCss = minifyCss(rulesText.join(' '));
          allStyles.push(minifiedCss); // Push the minified CSS to the array
        }
      } catch (e) {
        // console.warn("Error accessing stylesheet: ", sheet.href, "; Error: ", e.message);
      }
    });

    return allStyles.join(''); // Join the array of styles into a single string
  }


  function minifyCss(cssText) {
    // Remove all CSS comments
    let minified = cssText.replace(/\/\*[\s\S]*?\*\//g, '');

    // Collapse all excess whitespace, and reformat CSS appropriately for minification
    minified = minified.replace(/\s+/g, ' ')
      .replace(/;\s*/g, ';')  // Normalize semicolon spacing
      .replace(/\s*{\s*/g, '{')
      .replace(/\s*}\s*/g, '}')
      .replace(/\s*:\s*/g, ':') // Remove space after colon
      .replace(/,\s*/g, ','); // Remove space after comma

    // Remove spaces before and after the > child selector
    minified = minified.replace(/\s*>\s*/g, '>');

    // Remove leading 0 before decimal point for all numeric values
    minified = minified.replace(/\b0\.\d+/g, match => match.replace('0.', '.'));

    // Remove space after colon in transition property
    minified = minified.replace(/transition:\s*/g, 'transition:');

    // Remove 0 before s unit in transition property's duration value
    minified = minified.replace(/\b(\d*\.?\d+)0s\b/g, '$1s');

    // Remove the last semicolon in a style definition if it's directly before a }
    minified = minified.replace(/;(?=\s*})/g, '');

    // minified = cssText;
    return minified;
  }



  // Start the process of gathering and sending CSS
  const cssData = gatherExternalStyles();
  sendCssToServer(cssData);

  // Send the gathered and minified CSS to the server
  async function sendCssToServer(cssData) {
    const nonce = window.dynamicallyLoadedFilesDataCss.cssNonce;

    const url = window.dynamicallyLoadedFilesDataCss.getUsedCssUrl;
    const pageContext = document.getElementById('used-css-info');
    const templateType = pageContext.getAttribute('data-template-or-slug');

    try {
      const response = await fetch(url, {
        method: 'POST',
        headers: {
          'Content-Type': 'application/json',
          'X-WP-Nonce': nonce
        },
        credentials: 'same-origin',
        body: JSON.stringify({
          template_type: templateType,
          css: cssData
        })
      });

      if (!response.ok) {
        throw new Error(`Network response was not ok: ${response.statusText}`);
      }

      const data = await response.json();
      console.log('CSS successfully sent to server:', data);
    } catch (error) {
      console.error('Error sending CSS data to server:', error);
    }

  }

  console.log("CSS Data:", cssData);

})();
