/*** CUSTOM SRCSET AND LAZY LOAD FOR BACKGROUND IMAGES***/



(function () {

  /*** Auto-set Background Image Size ***/

  const SSBackgroundImages = document.querySelectorAll('[bg-srcset-all], [bg-srcset-desktop], [bg-srcset-tablet], [bg-srcset-mobile]');

  // Check if there are any elements found with the specified attributes.
  if (SSBackgroundImages.length > 0) {

    // Function to parse pixel density settings from JSON element
    let settings = {}; // Initialize settings object

    function parsePixelDensitySettings() {
      const settingsElement = document.getElementById('superspeed-fixed-pixel-density-bg-img');
      if (settingsElement && settingsElement.textContent) {
        try {
          settings = JSON.parse(settingsElement.textContent);
        } catch (e) {
          console.error('Error parsing pixel density settings:', e);
          settings = {
            custom_or_native_desktop: "native",
            pixel_density_desktop: 1,
            custom_or_native_tablet: "native",
            pixel_density_tablet: 1,
            custom_or_native_mobile: "native",
            pixel_density_mobile: 1,
            image_info: false
          };
        }
      }
    }

    parsePixelDensitySettings();

    // Enhanced Pixel Density Determination
    function getPixelDensity() {
      const currentWidth = window.innerWidth;
      if (currentWidth > 1024) {
        return settings.custom_or_native_desktop === "custom" ? settings.pixel_density_desktop : window.devicePixelRatio || 1;
      } else if (currentWidth > 480 && currentWidth <= 1024) {
        return settings.custom_or_native_tablet === "custom" ? settings.pixel_density_tablet : window.devicePixelRatio || 1;
      } else {
        return settings.custom_or_native_mobile === "custom" ? settings.pixel_density_mobile : window.devicePixelRatio || 1;
      }
    }

    const canEdit = document.querySelector('body.can-edit');

    if (canEdit) {
      // Function to display pixel density information
      function displayPixelDensityInfo() {
        if (!settings.image_info) return;

        const nativePixelDensity = window.devicePixelRatio || 1;
        const pixelDensityDisplay = document.createElement('div');
        pixelDensityDisplay.classList.add("pixel-density-bg-display-wrap");

        const superspeedPageInfo = document.getElementById('superspeed-page-info');

        let existingNativePixelDensityInfo = document.querySelector('#native-pixel-info');
        if (!existingNativePixelDensityInfo) {
          existingNativePixelDensityInfo = document.createElement('div');
          existingNativePixelDensityInfo.id = 'native-pixel-info';
          existingNativePixelDensityInfo.textContent = `Native Pixel Density Of This Screen: ${nativePixelDensity}`;
          pixelDensityDisplay.appendChild(existingNativePixelDensityInfo); // Append immediately to avoid detached DOM trees
        } else {
          existingNativePixelDensityInfo.textContent = `Native Pixel Density Of This Screen: ${nativePixelDensity}`;
        }

        // Dynamically construct settings list to improve readability and maintainability
        const settingsList = document.createElement('ul');
        settingsList.classList.add("page-info-item");
        settingsList.innerHTML = `<span class="pixel-density-title">Settings For Background Images</span>
        <li class="pixel-desity-screen">Pixel Density Desktop: ${settings.custom_or_native_desktop === "custom" ? `Custom - ${settings.pixel_density_desktop}` : "Native"}</li>
        <li class="pixel-desity-screen">Pixel Density Tablet: ${settings.custom_or_native_tablet === "custom" ? `Custom - ${settings.pixel_density_tablet}` : "Native"}</li>
        <li class="pixel-desity-screen">Pixel Density Mobile: ${settings.custom_or_native_mobile === "custom" ? `Custom - ${settings.pixel_density_mobile}` : "Native"}</li>`;

        // Append the constructed settings list to pixelDensityDisplay, then append to superspeed-page-info
        pixelDensityDisplay.appendChild(settingsList);
        superspeedPageInfo.appendChild(pixelDensityDisplay);
      }

      displayPixelDensityInfo();
    }

    // Function to get and display pixel density
    function getAndDisplayPixelDensity() {
      const pixelDensity = getPixelDensity();
      return pixelDensity;
    }


    // Function to fetch and display image size
    function fetchImageSize(url, callback, timeout = 5000) { // timeout in milliseconds
      const controller = new AbortController();
      const signal = controller.signal;

      // Set a timeout to abort the fetch
      const timeoutId = setTimeout(() => controller.abort(), timeout);

      fetch(url, { signal })
        .then(response => {
          clearTimeout(timeoutId); // Clear the timeout on successful response
          const size = response.headers.get('Content-Length');
          if (size) {
            const sizeInKB = (size / 1024).toFixed(0); // Convert bytes to kilobytes
            callback(sizeInKB + ' KB');
          } else {
            callback('Size not available');
          }
        })
        .catch(e => {
          if (e.name === 'AbortError') {
            callback('Fetch timeout'); // Handle fetch timeout
          } else {
            callback('Error fetching size');
          }
          console.error('Error fetching image size:', e);
        });
    }


    // Caching Mechanisms
    let sourcesCache = {};

    // Adjusted Function to Use Source Cache
    function chooseImageFromSrcset(srcset, pixelDensity, longestDimension) {
      let sources = sourcesCache[srcset];
      if (!sources) {
        sources = srcset.split(',').map(s => {
          const [url, minWidth] = s.trim().split('|');
          return { url, minWidth: parseInt(minWidth, 10) };
        });
        sources.sort((a, b) => b.minWidth - a.minWidth);
        sourcesCache[srcset] = sources;
      }

      const adjustedDimension = longestDimension * pixelDensity || 1;
      let selectedSource = sources.find(source => adjustedDimension >= source.minWidth);
      if (!selectedSource) {
        selectedSource = sources[sources.length - 1];
      }
      return selectedSource.url;
    }

    if (canEdit) {
      let infoElementsCache = {};

      // Revised Function to Create and Update Info Elements
      function createAndUpdateInfoElements(element, url, longestDimension) {
        if (!settings.image_info) return; // Exit if image_info is false
        // Generate a unique identifier for the element
        let elementId = element.getAttribute('element-id');
        if (!elementId) {
          elementId = 'element-' + Date.now() + '-' + Math.random().toString(36).substring(2, 9);
          element.setAttribute('element-id', elementId);
        }

        let infoContainer = infoElementsCache[elementId];

        // Create the info container if it does not exist
        if (!infoContainer) {
          infoContainer = document.createElement("div");
          infoContainer.classList.add("bg-info-container");

          // Create a button to hide the container
          const hideButton = document.createElement("button");
          hideButton.textContent = "+";
          hideButton.classList.add("hide-img-bg-info-container");
          hideButton.onclick = function () {
            infoContainer.style.display = 'none'; // Directly hide the container
          };

          // Append the button to the info container
          infoContainer.appendChild(hideButton);

          // Insert the info container before the element in the DOM
          element.parentNode.insertBefore(infoContainer, element);

          // Cache the newly created container
          infoElementsCache[elementId] = infoContainer;
        }

        // Update or create URL and Longest Dimension info
        let urlInfo = infoContainer.querySelector('.auto-img-url');
        let dimensionInfo = infoContainer.querySelector('.auto-img-size');

        if (!urlInfo) {
          urlInfo = document.createElement("li");
          urlInfo.classList.add("auto-img-url");
          infoContainer.appendChild(urlInfo);
        }
        urlInfo.textContent = `URL: ${url}`;

        if (!dimensionInfo) {
          dimensionInfo = document.createElement("li");
          dimensionInfo.classList.add("auto-img-size");
          infoContainer.appendChild(dimensionInfo);
        }
        dimensionInfo.textContent = `Longest Dimension: ${longestDimension}px`;

        // Fetch and display image size, update if already exists
        fetchImageSize(url, (size) => {
          let sizeInfo = infoContainer.querySelector('.auto-img-size-kb');
          if (!sizeInfo) {
            sizeInfo = document.createElement("li");
            sizeInfo.classList.add("auto-img-size-kb");
            infoContainer.appendChild(sizeInfo);
          }
          sizeInfo.textContent = `Size: ${size}`;
        });
      }
    }

    function getScreenSpecificOverlay(element) {
      if (element.getAttribute('overlay')) {
        return element.getAttribute('overlay'); // All screen sizes
      } else {
        const currentWidth = window.innerWidth;
        if (currentWidth >= 1024) { // Desktop
          return element.getAttribute('overlay_desktop');
        } else if (currentWidth >= 768) { // Tablet
          return element.getAttribute('overlay_tablet');
        } else { // Mobile
          return element.getAttribute('overlay_mobile');
        }
      }
    }

    // Function to update background image
    function updateBackgroundImage(element, pixelDensity) {
      if (!canEdit) {
        return;
      }

      const width = window.innerWidth; // Use viewport width to determine screen size
      let srcsetAttribute;

      // Determine which attribute to use based on screen size
      if (width <= 767) {
        srcsetAttribute = element.getAttribute('bg-srcset-mobile') || element.getAttribute('bg-srcset-all');
      } else if (width >= 768 && width <= 1023) {
        srcsetAttribute = element.getAttribute('bg-srcset-tablet') || element.getAttribute('bg-srcset-all');
      } else {
        srcsetAttribute = element.getAttribute('bg-srcset-desktop') || element.getAttribute('bg-srcset-all');
      }

      if (!srcsetAttribute) return; // Exit if no appropriate srcset is found

      const longestDimension = Math.max(element.offsetWidth, element.offsetHeight);
      const chosenImage = chooseImageFromSrcset(srcsetAttribute, pixelDensity, longestDimension);
      const overlayValue = getScreenSpecificOverlay(element);
      let backgroundImageStyle = '';

      if (overlayValue) {
        // Replace the image URL in the overlayValue string with the chosenImage
        const overlayValueUpdated = overlayValue.replace(/url\("?([^"]+)"?\)/, `url("${chosenImage}")`);
        backgroundImageStyle = `background-image: ${overlayValueUpdated};`;
      } else {
        backgroundImageStyle = `background-image: url('${chosenImage}');`;
      }

      // Check if the new chosen image is different from the current background image
      const currentStyle = element.getAttribute('style') || '';
      if (currentStyle.includes(backgroundImageStyle)) {
        return;
      }

      // Update the style attribute only if the chosen image is different
      element.setAttribute('style', currentStyle + (currentStyle ? ' ; ' : '') + backgroundImageStyle);
      element.classList.add('background-auto-set');
      createAndUpdateInfoElements(element, chosenImage, longestDimension);
    }



    // Global variable to track the largest screen width
    let maxWidthReached = 0;

    // Function to update all background images
    function updateAllBackgroundImages() {
      const currentWidth = window.innerWidth;
      const pixelDensity = getAndDisplayPixelDensity(); // Get and display pixel density

      // Update images only if the screen width is increasing
      if (currentWidth > maxWidthReached) {
        maxWidthReached = currentWidth;
        const elements = document.querySelectorAll('[bg-srcset-mobile], [bg-srcset-tablet], [bg-srcset-desktop], [bg-srcset-all]');
        elements.forEach(element => updateBackgroundImage(element, pixelDensity));
      }
    }

    // Throttle function
    const throttle = (func, limit) => {
      let inThrottle;
      return function () {
        const args = arguments;
        const context = this;
        if (!inThrottle) {
          func.apply(context, args);
          inThrottle = true;
          setTimeout(() => (inThrottle = false), limit);
        }
      };
    };

    // Initial update and setup for resize event
    updateAllBackgroundImages();
    const throttledUpdate = throttle(updateAllBackgroundImages, 250);
    window.addEventListener("resize", throttledUpdate);

    // Cleanup on page unload
    window.addEventListener("beforeunload", () => {
      window.removeEventListener("resize", throttledUpdate);
    });

  }










  /*** Lazy Load Background Images ***/

  // Check for the presence of '#superspeed-lazy-bg-fadein' before any attribute manipulation
  const lazyLoadStyleExists = document.getElementById('superspeed-lazy-bg-fadein') !== null;

  // Attribute manipulation based on lazy load style presence
  if (lazyLoadStyleExists) {
    const elements = document.querySelectorAll('[has_background_image]');
    elements.forEach(element => {
      if (element.children.length === 0) {
        element.setAttribute('has_background_image', 'no-children');
      }
    });
  }

  // Selecting targets that may need lazy loading or visibility changes
  const targetSelectors = 'div, section, header, footer, article, aside, main, nav, span, h1, h2, p, li, ul, ol, table, tr, td, form';
  // Select elements not already processed
  let targets = Array.from(document.querySelectorAll(targetSelectors))
    .filter(target => !target.classList.contains('ss-allow-bg'));

  const options = {
    root: null,
    rootMargin: '100px', // Adjusted for better anticipation of element visibility
    threshold: 0.0 // Adjusted to detect elements as soon as they start entering the viewport
  };

  const observer = new IntersectionObserver((entries, observer) => {
    entries.forEach(entry => {
      if (entry.isIntersecting) {
        entry.target.classList.add('ss-allow-bg');

        // Dynamically check for computed background images
        const computedStyle = window.getComputedStyle(entry.target);
        if (computedStyle.backgroundImage.includes('url(')) {
          entry.target.classList.add('bg-lazy-loaded');
        }

        observer.unobserve(entry.target);
      }
    });


    // Disconnect the observer if no more targets to observe
    if (!targets.find(target => !target.classList.contains('ss-allow-bg'))) {
      observer.disconnect();
    }
  }, options);

  targets.forEach(target => observer.observe(target));

})();
