/*** CUSTOM SRCSET AND LAZY LOAD FOR IMAGES***/

(function () {

  /* Custom srcset - Auto-set optimal thumbnail */

  const SSimages = document.querySelectorAll('img[ss-srcset]:not([src*=".svg"], .avatar-26, .avatar-64)');
  if (SSimages.length > 0) {

    // Function to parse pixel density settings from JSON element
    let settings = {}; // Initialize settings object

    const settingsElement = document.getElementById('superspeed-fixed-pixel-density-img');

    function parsePixelDensitySettings() {

      if (settingsElement && settingsElement.textContent) {
        try {
          settings = JSON.parse(settingsElement.textContent);
        } catch (e) {
          console.error('Error parsing pixel density settings:', e);
          settings = {
            custom_or_native_desktop: "native",
            pixel_density_desktop: 1,
            custom_or_native_tablet: "native",
            pixel_density_tablet: 1,
            custom_or_native_mobile: "native",
            pixel_density_mobile: 1,
            image_info: false
          };
        }
      }
    }

    parsePixelDensitySettings();

    const currentWidth = window.innerWidth;
    let pixelDensity;

    // Determine Pixel Density
    if (settingsElement) {
      // Using fixed pixel density values from settings
      if (currentWidth > 1024) { // desktop
        pixelDensity = settings.custom_or_native_desktop === "custom" ? settings.pixel_density_desktop : window.devicePixelRatio || 1;
      } else if (currentWidth > 480 && currentWidth <= 1024) { // tablet
        pixelDensity = settings.custom_or_native_tablet === "custom" ? settings.pixel_density_tablet : window.devicePixelRatio || 1;
      } else { // mobile
        pixelDensity = settings.custom_or_native_mobile === "custom" ? settings.pixel_density_mobile : window.devicePixelRatio || 1;
      }
    } else {
      // Default to device pixel ratio if settings are not available
      pixelDensity = window.devicePixelRatio || 1;
    }

    const canEdit = document.querySelector('body.can-edit');

    if (canEdit) {
      if (settings.image_info) {
        const nativePixelDensity = window.devicePixelRatio || 1;
        const pixelDensityDisplay = document.createElement('div');
        pixelDensityDisplay.classList.add("pixel-density-img-display-wrap");

        const superspeedPageInfo = document.getElementById('superspeed-page-info');

        let existingNativePixelDensityInfo = superspeedPageInfo.querySelector('#native-pixel-info');
        if (!existingNativePixelDensityInfo) {
          existingNativePixelDensityInfo = document.createElement('div');
          existingNativePixelDensityInfo.id = 'native-pixel-info';
          existingNativePixelDensityInfo.textContent = `Native Pixel Density Of This Screen: ${nativePixelDensity}`;
          // Append newly created element if it didn't exist before
          pixelDensityDisplay.appendChild(existingNativePixelDensityInfo);
        } else {
          // If it exists, just update its content in case the pixel density has changed (though unlikely between checks)
          existingNativePixelDensityInfo.textContent = `Native Pixel Density Of This Screen: ${nativePixelDensity}`;
        }

        // Constructing the settings list dynamically
        const settingsList = document.createElement('ul');
        settingsList.classList.add("page-info-item");
        settingsList.innerHTML = `<span class="pixel-density-title">Settings For Images</span>
        <li class="pixel-desity-screen">Pixel Density Desktop: ${settings.custom_or_native_desktop === "custom" ? `Custom - ${settings.pixel_density_desktop}` : "Native"}</li>
        <li class="pixel-desity-screen">Pixel Density Tablet: ${settings.custom_or_native_tablet === "custom" ? `Custom - ${settings.pixel_density_tablet}` : "Native"}</li>
        <li class="pixel-desity-screen">Pixel Density Mobile: ${settings.custom_or_native_mobile === "custom" ? `Custom - ${settings.pixel_density_mobile}` : "Native"}</li>`;

        // Append settings list to the pixelDensityDisplay
        pixelDensityDisplay.appendChild(settingsList);

        // Finally, append the pixelDensityDisplay to superspeed-page-info
        superspeedPageInfo.appendChild(pixelDensityDisplay);
      }
    }

    // Get image file size
    if (settings.image_info) {
      function fetchImageSize(url, callback) {
        fetch(url)
          .then(response => {
            const size = response.headers.get('Content-Length');
            if (size) {
              const sizeInKB = (size / 1024).toFixed(0); // Convert bytes to kilobytes. Round off the decimals
              callback(sizeInKB + ' KB');
            } else {
              callback('Size not available');
            }
          })
      }
    }

    // Store the maximum window width reached, so that we don't copute for downsizing
    let maxWidthReached = {};
    const infoElementsCache = {}; // Cache for info elements
    const sourcesCache = {}; // Cache for sources

    const updateImageSourceAfterLoad = (img, checkWidth = true) => {
      const uniqueId = img.getAttribute("data-unique-id");

      if (img.complete && img.naturalHeight !== 0) {
        updateImageSource(img, checkWidth);
        // Update maxWidthReached after the update
        maxWidthReached[uniqueId] = Math.max(maxWidthReached[uniqueId], window.innerWidth);
      } else {
        img.addEventListener("load", () => {
          updateImageSource(img, checkWidth);
          // Update maxWidthReached after the update
          maxWidthReached[uniqueId] = Math.max(maxWidthReached[uniqueId], window.innerWidth);
        }, { once: true });
      }
    };


    const updateImageSource = (img, checkWidth = true) => {
      const currentWidth = window.innerWidth;
      const selector = img.getAttribute("data-unique-id");

      // Only update if the current window width is larger than the maximum width reached for this image
      if (!checkWidth || currentWidth > maxWidthReached[selector]) {

        let sources = sourcesCache[selector];

        if (!sources) {
          const ssSrcset = img.getAttribute("ss-srcset");
          sources = ssSrcset.split(",").map(src => {
            const [url, minWidth] = src.trim().split("|");
            const adjustedWidth = parseInt(minWidth, 10);
            return { src: url, minWidth: adjustedWidth };
          });

          // Sort sources by minWidth in descending order
          sources.sort((a, b) => b.minWidth - a.minWidth);

          // Cache the sorted sources array
          sourcesCache[selector] = sources;
        }

        const greaterSide = Math.max(img.offsetWidth, img.offsetHeight);
        const adjustedForPXDensity = greaterSide * pixelDensity;

        // Find the source that is greater than the current size, starting from the largest
        const selectedSource = sources.find(source => adjustedForPXDensity >= source.minWidth) || sources[sources.length - 1];

        // Load the new image in a temporary Image object
        if (img.src !== selectedSource.src) {
          const tempImg = new Image();
          tempImg.onload = () => {
            // Once the new image is loaded, update the original image's src
            img.src = selectedSource.src;
          };
          tempImg.src = selectedSource.src;
        }

        // Running the function outside the IF statement, to provide info for images that are very small and not auto-set
        updateInfoElements(img, selectedSource, greaterSide);

        // Update the maxWidthReached for this image
        maxWidthReached[selector] = Math.max(maxWidthReached[selector], currentWidth);
      }
    };


    const updateInfoElements = (img, selectedSource, greaterSide) => {
      if (canEdit) {
        const selector = img.getAttribute("data-unique-id");
        if (settings.image_info) {
          // Fetch and display image size, then update info elements
          fetchImageSize(selectedSource.src, size => {
            const infoElements = infoElementsCache[selector] || {};
            // Update or create the size element
            if (!infoElements['info-size-kb']) {
              const infoSizeKB = document.createElement('div');
              infoSizeKB.classList.add('auto-img-size-kb');
              img.parentNode.insertBefore(infoSizeKB, infoElements["info-url"]); // Insert before the URL element
              infoElements['info-size-kb'] = infoSizeKB;
            }
            infoElements['info-size-kb'].textContent = `Size: ${size} KB`;
            infoElements["info-url"].textContent = `URL: ${selectedSource.src}`;
            infoElements["info-size"].textContent = `Longest Dimension: ${greaterSide}px`;

            // Cache the updated info elements
            infoElementsCache[selector] = infoElements;
          });
        }
      }
    };

    // Initialize all images and info elements on page load
    let imageInstanceCounter = 0;

    SSimages.forEach((img) => {
      // Set a unique selector. Using 8 characters for the random string
      let selector = 'ss-' + Math.random().toString(36).substring(2, 10);
      img.setAttribute("data-unique-id", selector);

      // Generate a unique identifier for this image instance
      const uniqueId = `${selector}-instance-${imageInstanceCounter++}`;
      img.setAttribute("data-unique-id", uniqueId);

      // Initialize maxWidthReached for this specific instance
      if (!maxWidthReached[uniqueId]) {
        maxWidthReached[uniqueId] = 0;
      }

      const infoContainer = document.querySelector(`[data-info-for="${uniqueId}"]`);

      if (infoContainer) {
        const infoURL = infoContainer.querySelector('.auto-img-url');
        const infoSize = infoContainer.querySelector('.auto-img-size');
        const infoSizeKB = infoContainer.querySelector('.auto-img-size-kb');

        if (infoURL) infoURL.textContent = `URL: ${source.src}`;
        if (infoSize) infoSize.textContent = `Longest Dimension: ${greaterSide}px`;

        if (infoSizeKB && settings.image_info) {
          fetchImageSize(source.src, size => {
            infoSizeKB.textContent = `Size: ${size}`;
          });
        }
      }

      const createInfoElements = (img) => {
        if (canEdit) {
          const selector = img.getAttribute("data-unique-id");
          const infoContainer = document.createElement("div");
          infoContainer.classList.add("img-info-container");

          const infoURL = document.createElement("li");
          infoURL.classList.add("auto-img-url");

          const infoSize = document.createElement("li");
          infoSize.classList.add("auto-img-size");

          const infoSizeKB = document.createElement("li");
          infoSizeKB.classList.add("auto-img-size-kb");

          // Create a button to hide the container
          const hideButton = document.createElement("button");
          hideButton.textContent = "+";
          hideButton.classList.add("hide-img-bg-info-container");
          hideButton.onclick = function () {
            infoContainer.style.display = 'none'; // Directly hide the container
          };

          // Append the hide button to the info container
          infoContainer.appendChild(hideButton);

          // Append the info elements to the container
          infoContainer.appendChild(infoURL);
          infoContainer.appendChild(infoSize);
          infoContainer.appendChild(infoSizeKB);

          // Insert the container before the img in the DOM
          img.parentNode.insertBefore(infoContainer, img);

          // Cache the elements
          infoElementsCache[selector] = {
            "info-url": infoURL,
            "info-size": infoSize,
            "info-size-kb": infoSizeKB,
            "info-container": infoContainer  // Adding the container itself to the cache
          };
        }
      }

      if (settings.image_info && canEdit) {
        createInfoElements(img, uniqueId);
      }
      updateImageSourceAfterLoad(img, false);
    });

    // Throttle function to limit the rate at which a function can fire
    const throttle = (func, limit) => {
      let inThrottle;
      return function () {
        const args = arguments;
        const context = this;
        if (!inThrottle) {
          func.apply(context, args);
          inThrottle = true;
          setTimeout(() => (inThrottle = false), limit);
        }
      };
    };

    // Update all images on window resize with throttling
    const throttledUpdate = throttle(() => {
      SSimages.forEach((img) => updateImageSource(img, true));
    }, 250);

    window.addEventListener("resize", throttledUpdate);

    // Function to remove all event listeners when the page is unloaded or navigated away from
    const cleanup = () => {
      window.removeEventListener("resize", throttledUpdate);
    };

    window.addEventListener("beforeunload", cleanup);

  }


  /****************************************** */



  /*** LAZY LOAD IMAGES ***/

  const LLimages = document.querySelectorAll('[lazy_loaded_img]');

  if (LLimages) {

    // Optimize Intersection Observer: Use one observer for all images
    const observer = new IntersectionObserver((entries) => {
      entries.forEach((entry) => {
        const img = entry.target;

        // If the image is within the viewport and not already faded-in
        if (entry.isIntersecting && !img.classList.contains("fade-in")) {
          // Add 'fade-in' class for visual effect
          img.classList.add("fade-in");

          // Check if the image has the 'data-src' attribute for lazy loading
          if (img.hasAttribute('data-src')) {
            // Set the 'src' attribute using the 'data-src' value
            img.src = img.getAttribute('data-src');
            // Remove the 'data-src' attribute after setting the 'src'
            img.removeAttribute('data-src');
          }

          // Cleanup Observers
          observer.unobserve(img);
        }
      });
    }, {
      // Throttle Intersection Observer Callbacks
      threshold: 0.1,  // This means 10% of the image is visible
    });


    // Combine Loops
    LLimages.forEach((imageElement) => {
      // Check if the image has the 'ss-img-placeholder' class and remove the placeholder if it's the next sibling
      const nextSibling = imageElement.nextElementSibling;

      if (nextSibling && nextSibling.classList.contains("ss-img-placeholder")) {
        nextSibling.remove();
      }

      // Add the class 'ss-image-loaded' to each image
      imageElement.classList.add("ss-image-loaded");

      // Use Intersection Observer to check if the image is within the viewport
      observer.observe(imageElement);
    });

    addLazyLoadMarkingStyle();
  }

  // If the gray placeholder is used, the CSS is not applied
  // That's why are are adding it with JS, once the image is on the screen
  function addLazyLoadMarkingStyle() {
    const grayPlaceholder = document.querySelectorAll('[img_placeholder="gray"]');
    const markImagesStyle = document.getElementById('superspeed-mark-lazy-images');
    if (grayPlaceholder && markImagesStyle) {
      const lazyLoadMarkingStyle = document.createElement('style');
      lazyLoadMarkingStyle.id = "lazyLoadMarkingStyle";
      lazyLoadMarkingStyle.textContent = `
          /* Added from lazy-load-images.js, because styles were not applied regularly */
          .can-edit img[lazy_loaded_img] {
            outline: dashed 0.25rem blue !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }
        `;
      document.body.appendChild(lazyLoadMarkingStyle);
    }
  }


})();
