
/* List Script and Style files on the frontend */

(function () {
  const element = document.querySelector("#wp-admin-bar-list-js-css .ab-item");
  element.addEventListener('click', listJsCssFiles);

  async function listJsCssFiles() {
    let parentDiv = document.createElement('div');
    parentDiv.id = 'absolute-lists-grid';
    parentDiv.setAttribute("aria-label", "Parent div for scripts and styles lists");
    document.body.appendChild(parentDiv);

    const closeButton = document.createElement('button');
    closeButton.id = 'ss-js-css-close';
    closeButton.textContent = 'Close';
    closeButton.addEventListener('click', removeParentDiv);
    parentDiv.appendChild(closeButton);

    parentDiv.addEventListener('click', handleButtonClick);

    function removeParentDiv() {
      document.body.removeChild(parentDiv);
      closeButton.removeEventListener('click', removeParentDiv);
    }

    async function handleButtonClick(event) {
      const { target } = event;
      if (target.matches('#ss-js-css-copy')) {
        await copyToClipboard(target.parentElement.parentElement.querySelector('.ss-files-list').innerText, target.nextElementSibling);
      }
    }

    function createAssetList(assetType) {
      const div = document.createElement('div');
      div.setAttribute('data-type', assetType);
      div.classList.add('absolute-list');
      div.classList.add(assetType); // Add classes separately to avoid the error
      div.setAttribute("aria-label", `${assetType} list`);
      div.innerHTML = `
        <div class="title ${assetType}">List of ${assetType.charAt(0).toUpperCase() + assetType.slice(1)}</div>
        <div class="ss-btn-msg-cont">
          <button id="ss-js-css-copy">Copy List</button>
          <span id="ss-js-css-message"></span>
        </div>
        <ul class="ss-files-list"></ul>`;
      parentDiv.appendChild(div);

      updateList(div.querySelector('.ss-files-list'), assetType);
    }



    function updateList(ul, assetType) {
      ul.innerHTML = '';  // Clear existing list items
      const attribute = assetType === 'scripts' ? 'src' : 'href';
      let assets;

      // Select the correct assets based on the type
      if (assetType === 'scripts') {
        // Include both external scripts and inline scripts with an ID
        assets = Array.from(document.querySelectorAll(`script[src]:not([data-dynamic='true']), script[id]:not([src], [id="script"])`));
      } else {
        // Include only external stylesheets, exclude inline styles and non-stylesheet links
        assets = Array.from(document.querySelectorAll(`link[href][rel='stylesheet']:not([data-dynamic='true'], [id="style"])`));
      }

      assets.forEach(asset => {
        const assetURL = asset[attribute];
        const assetID = asset.id || '';

        // Exclude assets with 'superspeed' in URL or ID
        if ((assetURL && assetURL.includes('superspeed')) ||
          (assetURL && assetURL.includes('dashicons')) ||
          (assetURL && assetURL.includes('css/admin-bar.min')) ||
          assetID.includes('superspeed')) {
          return;
        }


        // Handle external assets
        if (asset[attribute] && !asset[attribute].startsWith('data:')) {
          try {
            const url = new URL(asset[attribute]);
            const path = url.pathname.substring(1);
            const hostname = url.hostname;
            const searchParams = url.search;

            const li = document.createElement('li');
            li.className = 'ss-file-item external-file'; // Class for external files

            if (hostname === location.hostname) {
              // Local file
              li.textContent = `${path}`;
            } else {
              // Third-party file
              li.textContent = `${hostname}${url.pathname}${searchParams}`;
            }

            ul.appendChild(li);
          } catch (error) {
            console.error('Failed to parse URL: ', asset[attribute]);
          }
        } else if (assetType === 'scripts' && asset.id) {
          // Handling inline scripts with IDs (only for scripts)
          const li = document.createElement('li');
          li.className = 'ss-file-item inline-file'; // Class for inline files
          li.textContent = asset.id;
          ul.appendChild(li);
        }
      });
    }

    createAssetList('scripts');
    createAssetList('styles');

    // Add as the next element after 'parentDiv'
    const parentDivSpacer = document.createElement('div');
    parentDivSpacer.id = 'absolute-lists-grid-spacer';
    parentDiv.appendChild(parentDivSpacer);

    async function copyToClipboard(text, copiedSpan) {
      try {
        await navigator.clipboard.writeText(text);
        copiedSpan.textContent = 'List Copied to Clipboard';
        setTimeout(() => copiedSpan.textContent = '', 3000);
      } catch (err) {
        console.error('Failed to copy: ', err);
        copiedSpan.textContent = 'Failed to copy text. Please try again.';
        setTimeout(() => copiedSpan.textContent = '', 3000);
      }
    }

  }
})();

