/** For background images that are not lazy loaded and not using auto-set thumbnails, 
 * we first set the smallest image using PHP
 * and then we return the original source using JS
 **/

(function () {
  let cachedWidth = window.innerWidth; // Cache window width

  function getScreenSpecificImageUrl(element) {
    if (element.getAttribute('bg_url')) {
      return element.getAttribute('bg_url'); // All screen sizes
    } else {
      const width = cachedWidth; // Use cached window width
      if (width >= 1024) { // Desktop
        return element.getAttribute('bg_url_desktop');
      } else if (width >= 768) { // Tablet
        return element.getAttribute('bg_url_tablet');
      } else { // Mobile
        return element.getAttribute('bg_url_mobile');
      }
    }
  }

  function getScreenSpecificOverlay(element) {
    if (element.getAttribute('overlay')) {
      return element.getAttribute('overlay'); // All screen sizes
    } else {
      const width = cachedWidth; // Use cached window width
      if (width >= 1024) { // Desktop
        return element.getAttribute('overlay_desktop');
      } else if (width >= 768) { // Tablet
        return element.getAttribute('overlay_tablet');
      } else { // Mobile
        return element.getAttribute('overlay_mobile');
      }
    }
  }

  function updateBackgroundImage(element) {
    if (['bg-srcset-all', 'bg-srcset-desktop', 'bg-srcset-tablet', 'bg-srcset-mobile'].some(attr => element.hasAttribute(attr))) {
      return;
    }

    const overlayValue = getScreenSpecificOverlay(element);
    const imageUrl = getScreenSpecificImageUrl(element);

    if (overlayValue) {
      const img = new Image();
      img.onload = function () {
        const backgroundOverlayStyle = `; background-image: ${overlayValue};`;
        if (element.hasAttribute('style')) {
          element.setAttribute('style', element.getAttribute('style') + " " + backgroundOverlayStyle);
        } else {
          element.setAttribute('style', backgroundOverlayStyle);
        }
        if (!element.classList.contains('original-bg-loaded')) {
          element.classList.add('original-bg-loaded');
        }
        img.onload = img.onerror = null; // Clean up to prevent memory leaks
      };
      img.onerror = function () {
        console.error(`Failed to load image: ${overlayValue}`);
        img.onload = img.onerror = null; // Clean up to prevent memory leaks
      };
      img.src = imageUrl;
    }

    else if (imageUrl) {
      const img = new Image();
      img.onload = function () {
        const backgroundImageStyle = `; background-image: url('${imageUrl}');`;
        if (element.hasAttribute('style')) {
          element.setAttribute('style', element.getAttribute('style') + " " + backgroundImageStyle);
        } else {
          element.setAttribute('style', backgroundImageStyle);
        }
        if (!element.classList.contains('original-bg-loaded')) {
          element.classList.add('original-bg-loaded'); // Ensure class is only added if not present
        }
        img.onload = img.onerror = null; // Clean up to prevent memory leaks
      };
      img.onerror = function () {
        console.error(`Failed to load image: ${imageUrl}`);
        img.onload = img.onerror = null; // Clean up to prevent memory leaks
      };
      img.src = imageUrl;
    }
  }

  function debounce(func, wait) {
    let timeout;
    return function executedFunction(...args) {
      const later = () => {
        clearTimeout(timeout);
        func.apply(this, args);
      };
      clearTimeout(timeout);
      timeout = setTimeout(later, wait);
    };
  }

  // Using debounce for window resize event
  const debouncedResize = debounce(() => {
    cachedWidth = window.innerWidth; // Update cached width on resize
    const elements = document.querySelectorAll("[bg_url_desktop], [bg_url_tablet], [bg_url_mobile], [bg_url]");
    elements.forEach(updateBackgroundImage);
  }, 250); // Adjust the wait time as needed

  window.addEventListener('resize', debouncedResize);

  if ('IntersectionObserver' in window) {
    const observer = new IntersectionObserver((entries, observer) => {
      entries.forEach((entry) => {
        if (entry.isIntersecting) {
          updateBackgroundImage(entry.target);
          observer.unobserve(entry.target); // Disconnect observer for this element
        }
      });
    }, {
      threshold: [0.0, 0.25, 0.5, 0.75, 1.0], // Trigger at various intersection points
      rootMargin: '100px' // Start loading images 100px before they enter the viewport
    });

    requestAnimationFrame(() => {
      const elements = document.querySelectorAll("[bg_url_desktop], [bg_url_tablet], [bg_url_mobile], [bg_url]");
      elements.forEach((element) => {
        observer.observe(element);
      });
    });
  } else {
    console.error("IntersectionObserver is not supported in this browser.");
  }
})();


