/** For images that are not lazy loaded and are not using auto-set thumbnails, 
 * we first set the smallest image using PHP
 * and then we return the original source using JS
 * if the image has the native 'srcset', we sore it at 'data-srcset', and place back when the image is in view
 **/

(function () {
  // Function to handle image visibility changes
  function handleImageVisibility(entries, observer) {
    entries.forEach(function (entry) {
      if (entry.isIntersecting) {
        // Image is in or near the viewport
        const image = entry.target;
        // Handle ss-original-src-img attribute
        const originalSrc = image.getAttribute('ss-original-src-img');
        if (originalSrc && image.src !== originalSrc) {
          loadImage(image, originalSrc, observer);
        }
        // Handle data-srcset attribute if ss-original-src-img is not present
        else if (!originalSrc && image.hasAttribute('data-srcset')) {
          const srcsetData = image.getAttribute('data-srcset');
          image.setAttribute('srcset', srcsetData);
          image.removeAttribute('data-srcset');
          observer.unobserve(image); // Stop observing once attribute is set
        }
      }
    });
  }

  // Function to load image and set src attribute
  function loadImage(image, src, observer) {
    // Create a new Image object
    const tempImage = new Image();

    // Set the onload event handler for the new image
    tempImage.onload = function () {
      // Once the new image is fully loaded, swap the src attribute
      image.src = src; // This directly sets the src attribute
      observer.unobserve(image); // Stop observing once loaded
    };

    // Set the src attribute of the new image to trigger loading
    tempImage.src = src;
  }

  // Use requestAnimationFrame for better performance
  requestAnimationFrame(() => {
    // Select all img elements with 'ss-original-src-img' or 'data-srcset' attributes
    const images = document.querySelectorAll('[ss-original-src-img]:not([ss-srcset]), [data-srcset]:not([ss-srcset])');

    // Create an Intersection Observer
    const observer = new IntersectionObserver(handleImageVisibility, {
      root: null, // Use the viewport as the root
      threshold: 0.01, // Trigger when at least 1% of the image is in the viewport
    });

    // Observe each image
    images.forEach(image => {
      observer.observe(image);
    });
  });
})();

