(async function () {
  "use strict";

  const getStoredData = (key, session = false) => {
    const storage = session ? sessionStorage : localStorage;
    return JSON.parse(storage.getItem(key)) || undefined;
  };

  const setStoredData = (key, value, session = false) => {
    const storage = session ? sessionStorage : localStorage;
    storage.setItem(key, JSON.stringify(value));
  };

  const configElement = document.querySelector("#SSpreFetchConfig");
  let CONFIG;
  try {
    CONFIG = JSON.parse(configElement.getAttribute("ss-fetch-config"));
  } catch (e) {
    console.error("Error parsing config JSON", e);
    return;
  }

  const isMobile = window.innerWidth <= CONFIG.COMMON.MOBILE_WIDTH_THRESHOLD;
  const deviceSpecificConfig = isMobile ? CONFIG.MOBILE : CONFIG.DESKTOP;
  const commonConfig = CONFIG.COMMON;
  const testFileUrl = CONFIG.COMMON.TEST_FILE_URL;
  const testMode = CONFIG.COMMON.ENABLE_TEST_MODE;
  const RTT_THRESHOLD = isMobile
    ? deviceSpecificConfig.RTT_THRESHOLD
    : deviceSpecificConfig.RTT_THRESHOLD;

  let updateLogElement;

  // Collect values to display on the frontend
  let networkCurrentValue = '';
  let networRequiredValue = '';
  let serverStatusCurrentValue = '';
  let serverStatusRequiredValue = '';

  // Function to create and update the log element on the screen
  if (testMode) {
    // Clear specific local and session storage items if we are in test mode
    localStorage.removeItem("isEnabledOnMobile");
    localStorage.removeItem("isSlowDevice");
    localStorage.removeItem("requiredMemory");
    localStorage.removeItem("requiredCores");
    sessionStorage.removeItem("isSlowNetwork");
    sessionStorage.removeItem("serverUnderLoad");

    updateLogElement = (message) => {
      // Add the message to an accordion at the top of the screen if it exists
      const pageInfoWrap2 = document.getElementById('superspeed-page-info');
      if (pageInfoWrap2) {
        const messageTitle = document.createElement("div");
        messageTitle.classList.add("preload-links-title", "page-info-item");
        messageTitle.textContent = 'Preload Links Device Info';
        pageInfoWrap2.appendChild(messageTitle);

        const messageElement = document.createElement("div");
        messageElement.classList.add("preload-links-error");
        messageElement.textContent = message;
        pageInfoWrap2.appendChild(messageElement);
      }
    };
  }


  // The updateAndExit function
  const updateAndExit = (
    key,
    value,
    requiredValue,
    currentValue,
    reason,
    session = false
  ) => {
    setStoredData(key, value, session);

    if (testMode) {
      let logMessage;
      if (reason === "mobile not enabled") {
        logMessage = `Exiting pre-fetch due to: ${reason}.`;
      } else {
        logMessage = `Exiting pre-fetch due to: ${reason}. Current value: ${currentValue}, Required or Maximum value: ${requiredValue}`;
      }

      console.log(logMessage);
      updateLogElement(logMessage); // Update the log element on the screen
    }

    return;
  };

  // Check if enabled on mobile
  let isEnabledOnMobile = getStoredData("isEnabledOnMobile");
  if (isEnabledOnMobile === false)
    return updateAndExit(
      "isEnabledOnMobile",
      false,
      true,
      isMobile,
      "mobile not enabled"
    );

  if (typeof isEnabledOnMobile === "undefined") {
    if (isMobile && !CONFIG.COMMON.ENABLE_ON_MOBILE) {
      updateAndExit(
        "isEnabledOnMobile",
        false,
        true,
        isMobile,
        "mobile not enabled"
      );
      return;
    }
    setStoredData("isEnabledOnMobile", true);
  }

  // Check Device Speed
  // Get stored data
  let isSlowDevice = getStoredData("isSlowDevice");

  // Declare variables to be used later
  let requiredCores, requiredMemory;

  // Retrieve in-memory device properties
  // Using high default values, in case hardware can't be detected
  const cpuCores = window.navigator.hardwareConcurrency || 6;
  const deviceMemory = window.navigator.deviceMemory || 4;

  // Early exit if session storage already flagged this as a slow device
  if (isSlowDevice && !testMode) {
    // Get stored data
    requiredCores = getStoredData("requiredCores");
    requiredMemory = getStoredData("requiredMemory");

    updateAndExit(
      "isSlowDevice",
      true,
      `${requiredCores} cores, ${requiredMemory} GB memory`,
      `${cpuCores} cores, ${deviceMemory} GB memory`,
      "device is slow"
    );
    return;
  }

  if (typeof isSlowDevice === "undefined") {
    // Set required cores and memory based on device type and testing conditions
    if (isMobile) {
      if (testMode) {
        requiredCores = 1;
        requiredMemory = 1;
      } else {
        requiredCores = 4;
        requiredMemory = 2;
      }
    } else {
      // Laptop or Desktop
      requiredCores = 2;
      requiredMemory = 2;
    }

    // Set stored data
    setStoredData("requiredCores", requiredCores);
    setStoredData("requiredMemory", requiredMemory);

    // Check if the device is slow based on these requirements
    isSlowDevice = cpuCores <= requiredCores || deviceMemory < requiredMemory;

    if (isSlowDevice) {
      // Update session storage
      updateAndExit(
        "isSlowDevice",
        true,
        `${requiredCores} cores, ${requiredMemory} GB memory`,
        `${cpuCores} cores, ${deviceMemory} GB memory`,
        "device is slow"
      );
      setStoredData("isSlowDevice", true);
      return;
    } else {
      // Cache the result for future reference
      setStoredData("isSlowDevice", false);
    }
  }

  // Check network speed
  // Check for network API support
  if ("connection" in window.navigator) {
    const effectiveType = window.navigator.connection.effectiveType;

    // Make sure the getStoredData function is accessible here
    let isSlowNetwork = getStoredData("isSlowNetwork", true);

    // Declare variable to be used later
    let requiredEffectiveType;

    if (isSlowNetwork && !testMode) {
      // Get stored data
      requiredEffectiveType = getStoredData("requiredEffectiveType", true);

      updateAndExit(
        "isSlowNetwork",
        true,
        requiredEffectiveType,
        effectiveType,
        "network is slow",
        true
      );
      return;
    }

    if (typeof isSlowNetwork === "undefined") {
      requiredEffectiveType = "3g or higher"; // Used only for info message
      // Modify the array to include what you consider slow networks
      isSlowNetwork = ["slow-2g", "2g" /* '3g', '4g', '5g' */].includes(
        effectiveType
      );

      if (isSlowNetwork) {
        // Set stored data
        setStoredData("requiredEffectiveType", requiredEffectiveType, true);

        updateAndExit(
          "isSlowNetwork",
          true,
          requiredEffectiveType,
          effectiveType,
          "network is slow",
          true
        );
        return;
      }

      // Collect values to display on the frontend
      networkCurrentValue = effectiveType;
      networRequiredValue = requiredEffectiveType;

      setStoredData("isSlowNetwork", false, true);
    }
  } else {
    console.log(
      "Preload links JS could not identify network speed, because Network Information API is not supported."
    );
  }

  // Check for server load
  let serverUnderLoad = getStoredData("serverUnderLoad", true);
  const rttThreshold = isMobile
    ? deviceSpecificConfig.RTT_THRESHOLD
    : deviceSpecificConfig.RTT_THRESHOLD;

  if (serverUnderLoad && !testMode)
    return updateAndExit(
      "serverUnderLoad",
      true,
      rttThreshold,
      rttThreshold,
      "server under load",
      true
    );

  if (typeof serverUnderLoad === "undefined") {
    try {
      const beforeFetch = Date.now();
      const response = await fetch(testFileUrl); // This was calling the home page
      const afterFetch = Date.now();
      const rtt = afterFetch - beforeFetch;
      if (!response.ok) {
        console.log(`Server health check failed with status ${response.status}`);
      }
      serverUnderLoad = rtt > rttThreshold;

      if (serverUnderLoad) {
        updateAndExit(
          "serverUnderLoad",
          true,
          rttThreshold,
          rtt,
          "server under load",
          true
        );
        return;
      }
      setStoredData("serverUnderLoad", false, true);

      // Collect values to display on the frontend
      serverStatusCurrentValue = rtt;
      serverStatusRequiredValue = rttThreshold;

    } catch (error) {
      console.error("Fetch error while checking server load:", error);
      return;
    }
  }

  // Constants and Variables
  const FETCH_TIMEOUT = isMobile
    ? deviceSpecificConfig.MOBILE_FETCH_TIMEOUT
    : deviceSpecificConfig.DESKTOP_FETCH_TIMEOUT;
  const RETRY_DELAY_MS = commonConfig.RETRY_DELAY_MS;
  const MAX_RETRIES = commonConfig.MAX_RETRIES;
  const MOBILE_DELAY_MS = deviceSpecificConfig.MOBILE_DELAY_MS;
  const MOBILE_DELAY_BETWEEN_LINKS_MS =
    deviceSpecificConfig.MOBILE_DELAY_BETWEEN_LINKS_MS;

  // Named Constants
  const LOADED = "loaded";
  const IDLE = "idle";
  const CURRENT_PAGE = "current-page";

  // State management
  const state = {
    cache: new Map(),
    controllers: new Map(),
    fetchInProgress: new Set(),
    fetchedLinks: new Set(),
    viewportLinks: new Set(),
  };

  /********************** */

  // Add information to the frontend, above the header
  if (testMode) {

    // Define updateLogElement for debugging
    updateLogElement = (message) => {
      let logWrapper = document.getElementById("logWrapper");
      if (!logWrapper) {
        // Create wrapper div if it does not exist
        logWrapper = document.createElement("div");
        logWrapper.id = "logWrapper";
        document.body.appendChild(logWrapper);
      }
      // Append new message as a paragraph
      const msgElement = document.createElement("p");
      msgElement.textContent = message;
      logWrapper.appendChild(msgElement);
    };

    // Update Page Info function specific for test mode
    function updatePageInfo() {
      const deviceType = isMobile ? 'Mobile' : 'Desktop/Laptop';
      const cpuCores = navigator.hardwareConcurrency || 'Unknown';
      const memory = navigator.deviceMemory || 'Unknown';
      const serverLoadStatus = getStoredData("serverUnderLoad", true) ? 'Under load' : 'Normal';
      const isSlowNetwork = getStoredData("isSlowNetwork", true) ? 'Yes' : 'No';
      const isSlowDevice = getStoredData("isSlowDevice") ? 'Yes' : 'No';

      const infoHtml = `
        <div class="preload-links-title page-info-item">Preload Links Device Info</div>
        <ul class="preload-links-list">
            <li class="preload-links-item">Device Type: ${deviceType}</li>
            <li class="preload-links-item">Is Device Slow: <b>${isSlowDevice}</b>. CPU Cores: ${cpuCores}, Device Memory: ${memory} GB</li>
            <li class="preload-links-item">Is Network Slow: <b>${isSlowNetwork}</b>. Current value: ${networkCurrentValue}, Required value: ${networRequiredValue}.</li>
            <li class="preload-links-item">Server Response: <b>${serverLoadStatus}</b>. Current value: ${serverStatusCurrentValue}ms, Maximum value: ${serverStatusRequiredValue}ms.</li>
        </ul>
    `;

      const pageInfoWrap = document.getElementById('superspeed-page-info');
      if (pageInfoWrap) {
        pageInfoWrap.insertAdjacentHTML('beforeend', infoHtml);
      } else {
        console.error("Element with ID 'superspeed-page-info' not found.");
      }
    }

    // Call updatePageInfo to display the insights in test mode
    updatePageInfo();
  }
  /************************ */

  // Utility Functions
  const setLinkAttribute = (link, value) =>
    link.setAttribute("link-preload", value);
  const createController = () => new AbortController();

  // Mark all similar links as loaded
  const markAllSimilarLinksAsLoaded = (href) => {
    const links = document.querySelectorAll(`a[href="${href}"]`);
    links.forEach((link) => setLinkAttribute(link, LOADED));
  };

  // Fetch and Preload resources
  const fetchResource = (href, element, delay = 0, isMobile = false) => {
    if (state.fetchedLinks.has(href)) return;

    // Only set to 'waiting' if not on a mobile device, because of the delays
    if (isMobile) {
      setLinkAttribute(element, "waiting");
    }

    const controller = createController();
    state.controllers.set(element, controller);

    setTimeout(() => {
      if (isMobile) {
        setLinkAttribute(element, "loading");
      }
      performFetch(href, controller, element);
    }, delay);
  };

  // Async Function to perform fetch
  async function performFetch(href, controller, linkElement, retries = 0) {
    const finalizeFetch = (doc) => {
      preloadAssets(doc, linkElement);
      setLinkAttribute(linkElement, LOADED);
      state.fetchInProgress.delete(href);
      markAllSimilarLinksAsLoaded(href);
    };

    // Check the cache before fetching
    if (state.cache.has(href)) {
      const cachedData = state.cache.get(href);
      const doc = new DOMParser().parseFromString(cachedData, "text/html");
      finalizeFetch(doc);
      return;
    }

    // Set timeout based on device-specific configuration
    const timeoutPromise = new Promise((_, reject) => {
      setTimeout(() => {
        controller.abort();
        reject(new Error("Fetch timed out"));
      }, FETCH_TIMEOUT);
    });

    try {
      let fetchPromise;

      if (window.fetch) {
        // Using Fetch API if available
        fetchPromise = fetch(href, {
          signal: controller.signal,
          headers: {
            'X-Requested-With': 'XMLHttpRequest'  // Set the AJAX header
          }
        }).then((response) => {
          if (!response.ok)
            throw new Error(`Request failed with status ${response.status}`);
          return response.text();
        });
      } else {
        // Fallback to XMLHttpRequest
        fetchPromise = new Promise((resolve, reject) => {
          var xhr = new XMLHttpRequest();
          xhr.open("GET", href, true);
          xhr.onreadystatechange = function () {
            if (xhr.readyState == 4) {
              if (xhr.status >= 200 && xhr.status < 300) {
                resolve(xhr.responseText);
              } else {
                reject(new Error(`Request failed with status ${xhr.status}`));
              }
            }
          };
          xhr.send();
        });
      }

      const data = await Promise.race([fetchPromise, timeoutPromise]);

      // Cache the successfully fetched data
      state.cache.set(href, data);
      state.fetchedLinks.add(href);

      const doc = new DOMParser().parseFromString(data, "text/html");
      finalizeFetch(doc);
    } catch (error) {
      state.fetchInProgress.delete(href);

      if (retries < MAX_RETRIES) {
        setTimeout(() => {
          performFetch(href, controller, linkElement, retries + 1, Date.now());
        }, RETRY_DELAY_MS);
      } else {
        //console.error(`Max retries reached for ${href}. Fetch failed.`);
      }
    }
  }

  // Preload assets based on the configuration
  function preloadAssets(doc, linkElement) {

    // Current page origin
    const pageOrigin = new URL(window.location.href).origin;

    // Define asset types and their corresponding selectors or custom logic functions
    const preloadableTypes = [
      {
        type: "css",
        selector: 'link[rel="stylesheet"]',
      },
      {
        type: "js",
        selector: "script[src]",
      },
      {
        type: "image",
        selector: 'link[rel="preload"][as="image"]',
      },
    ];

    // Loop through each asset type to check if it should be preloaded
    for (const { type, selector } of preloadableTypes) {
      const assets = doc.querySelectorAll(selector);

      // Preload each asset
      assets.forEach((asset) => {
        // Determine the URL of the asset (href for link tags, src for script tags)
        let assetUrl = asset.href || asset.src;

        // Ensure the asset URL is defined before attempting to fetch it
        if (typeof assetUrl === "undefined") {
          return;
        }

        // Create a URL object from the asset's URL to access its origin
        const assetOrigin = new URL(assetUrl, document.baseURI).origin;

        // Only proceed with fetching the resource if it's from the same origin
        if (assetOrigin === pageOrigin) {
          fetchResource(assetUrl, linkElement);
        } else {
          // console.warn(`Skipping cross-origin asset: ${assetUrl}`);
        }
      });
    }
  }

  // Desktop Event Handlers
  let mouseOverHandler;
  let mouseOutHandler;

  if (!isMobile) {
    mouseOverHandler = (event) => {
      const link = event.target.closest("a");
      if (!link) return;

      const href = new URL(link.getAttribute("href"), window.location.href)
        .href;
      if (
        [LOADED, CURRENT_PAGE].includes(link.getAttribute("link-preload")) ||
        state.fetchedLinks.has(href) ||
        state.fetchInProgress.has(href)
      )
        return;

      setLinkAttribute(link, "loading");
      state.fetchInProgress.add(href);

      const controller = createController();
      state.controllers.set(link, controller);
      performFetch(href, controller, link);
    };

    mouseOutHandler = (event) => {
      const link = event.target.closest("a");
      if (!link) return;

      if (state.controllers.has(link)) {
        state.controllers.get(link).abort();
        state.controllers.delete(link);

        const linkState = link.getAttribute("link-preload");
        if (![LOADED, CURRENT_PAGE].includes(linkState)) {
          setLinkAttribute(link, "aborted");
        }
      }
    };
  }

  // Helper function to check if two URLs are the same, considering the path and optionally the query
  function areUrlsSame(url1, url2, considerQuery = false) {
    // Remove trailing slashes for comparison
    const path1 = url1.pathname.replace(/\/$/, "");
    const path2 = url2.pathname.replace(/\/$/, "");

    if (
      url1.protocol === url2.protocol &&
      url1.hostname === url2.hostname &&
      path1 === path2
    ) {
      if (considerQuery) {
        return url1.search === url2.search;
      }
      return true;
    }
    return false;
  }

  // Initialization
  function init() {
    // Initialize the current page's URL
    const currentPageUrl = new URL(window.location.href);

    // Get all anchor tags on the page
    const links = document.querySelectorAll("a");

    // Initialize attributes for each link based on whether it points to the current page
    links.forEach((link) => {
      const linkUrl = new URL(link.getAttribute("href"), window.location.href);
      const isCurrentPage = areUrlsSame(currentPageUrl, linkUrl, false);
      const hasNoQuery = !linkUrl.search;

      if (
        isCurrentPage &&
        (hasNoQuery || areUrlsSame(currentPageUrl, linkUrl, true))
      ) {
        setLinkAttribute(link, CURRENT_PAGE);
      } else {
        setLinkAttribute(link, IDLE);
      }
    });

    // Initialize Intersection Observer if on a mobile device
    if (isMobile) {
      const observerOptions = {
        root: null,
        rootMargin: "0px",
        threshold: 0.1,
      };

      const observer = new IntersectionObserver(
        handleIntersectionMobile,
        observerOptions
      );
      links.forEach((link) => observer.observe(link));
    }
  }

  let fetchQueue = [];
  let isFetching = false;

  // For mobile devices
  function handleIntersectionMobile(entries) {
    //console.log("Handling intersection for mobile...");

    // Sort the entries based on their position in the DOM
    const sortedEntries = entries.sort((a, b) => {
      return a.target.compareDocumentPosition(b.target) &
        Node.DOCUMENT_POSITION_PRECEDING
        ? 1
        : -1;
    });

    sortedEntries.forEach((entry) => {
      const linkElement = entry.target;
      const href = new URL(
        linkElement.getAttribute("href"),
        window.location.href
      ).href;

      // Skip fetching for links that point to the current page or include '#'
      if (
        [CURRENT_PAGE, "#"].includes(
          linkElement.getAttribute("link-preload")
        ) ||
        href.includes("#")
      ) {
        //console.log(`Skipping link: ${href}`);
        return;
      }

      if (entry.isIntersecting) {
        //console.log(`Link is intersecting: ${href}`);
        state.viewportLinks.add(href);

        if (!state.fetchInProgress.has(href)) {
          //console.log(`Adding link to fetchQueue: ${href}`);
          fetchQueue.push({
            href,
            linkElement,
          });
        }
      } else {
        //console.log(`Link is not intersecting: ${href}`);
        state.viewportLinks.delete(href);
        if (state.controllers.has(linkElement)) {
          //console.log(`Aborting fetch for link: ${href}`);
          state.controllers.get(linkElement).abort();
          state.controllers.delete(linkElement);
        }
      }
    });

    if (!isFetching) {
      //console.log("Starting to process fetch queue...");
      processFetchQueue();
    }
  }

  async function processFetchQueue() {
    isFetching = true;
    //console.log("Processing fetch queue...");

    while (fetchQueue.length > 0) {
      const { href, linkElement } = fetchQueue.shift();

      if (!href) {
        //console.log("Undefined href, skipping...");
        continue;
      }

      if (!state.viewportLinks.has(href)) {
        //console.log(`Link ${href} is no longer in the viewport, skipping...`);
        continue;
      }

      if (
        linkElement.getAttribute("link-preload") === CURRENT_PAGE ||
        href.includes("#")
      ) {
        //console.log(`Link ${href} should not be loaded, skipping without delay.`);
        continue;
      }

      //console.log(`Fetching resource: ${href} after initial delay of ${MOBILE_DELAY_MS}ms`);
      fetchResource(href, linkElement, MOBILE_DELAY_MS, true); // Pass true for isMobile

      //console.log(`Delaying for ${MOBILE_DELAY_BETWEEN_LINKS_MS}ms before next fetch`);
      await new Promise((resolve) =>
        setTimeout(() => {
          //console.log("Resuming fetch queue processing...");
          resolve();
        }, MOBILE_DELAY_BETWEEN_LINKS_MS)
      );
    }

    //console.log("Finished processing fetch queue.");
    isFetching = false;
  }

  // Function to abort all ongoing fetches
  function abortAllFetches() {
    for (const controller of state.controllers.values()) {
      controller.abort();
    }
    state.controllers.clear();
    state.fetchInProgress.clear();
  }

  // Main logic
  init();
  if (!isMobile) {
    document.addEventListener("mouseover", mouseOverHandler);
    document.addEventListener("mouseout", mouseOutHandler);
  }
  window.addEventListener("beforeunload", abortAllFetches);



})();
