<?php

namespace SuperSpeed;

use SuperSpeed\Utility;

class Media
{

  // Lazy load Images cache the result of checking which page we are on
  private $buffering_started_lazyload_images = false;

  // Add custom IDs to elements
  // private $buffering_started_add_custom_ids = false;

  // Set custom attributes based on gathered images data
  private $buffering_started_setting_attributes_from_gathered_image_data = false;

  // Background Images: SRCSET and Preload
  private $buffering_started_process_bg_img = false;

  // Lazy load iFrames, cache the result of checking if function should run
  private $buffering_started_lazyload_iframes = false;

  // Lazy load iFrames, cache the result of checking if function should run
  private $buffering_started_lazyload_videos = false;

  // Do we have image data and are images on the page
  private $imageDataAvailable = false;

  // Do we have image data and are background images on the page
  private $backgroundImageDataAvailable = false;


  public function __construct()
  {
    // Exit if we could not get the current URL. Skipping static resource because it is not a page.
    $current_url = Utility::get_current_page_url();
    if (!$current_url) {
      return;
    }

    // Call the helper function early to store the data - Do we have image info, and are using it on this page
    $this->determineImageDataAvailability();

    // Remove image sizes
    add_filter('intermediate_image_sizes', array($this, 'remove_image_thumbnails'));

    // Define additional image sizes / thumbnails
    add_action('after_setup_theme', array($this, 'add_custom_image_sizes'));

    // If the user has enabled auto-setting LCP images
    // And if we should be getting image data on this page
    if (Utility::get_images_info_on_this_page()) {

      // Use JS Ajax to store image data in the database
      add_action('wp_ajax_store_images_data', array($this, 'handle_get_images_data'));
      // add_action('wp_ajax_nopriv_store_images_data', array($this, 'handle_get_images_data')); // For non-logged-in users

      // Add the url of the post or page, so that it can be used to remove imges data
      add_action('wp', array($this, 'add_meta_with_post_url'));
      // When the post is edited, remove images data
      add_action('save_post', array($this, 'if_post_is_edited_remove_images_data'), 10, 3);

      // Dynamically load JS files of frontend tools for getting Images Data
      add_action('rest_api_init', [$this, 'register_rest_routes']);
      // Loading get-images-data.js after get-used-css.js
      add_action('wp_footer', [$this, 'enqueue_scripts_dynamically']);

      // Remove image data for specific page from the frontend
      add_action('wp_ajax_refresh_image_data', array($this, 'handle_refresh_image_data'));

      // Delete all image data from the frontend
      add_action('wp_ajax_delete_all_image_data', array($this, 'handle_delete_all_image_data'));
    }

    // If it's a frontend HTML page
    if (!Utility::is_page_editor_or_wp_backend()) {

      // Run these CSS rules at all times
      add_action('wp_head', array($this, 'css_to_run_always'), -1 * PHP_INT_MAX);

      // Images: lazy load, preload, auto-set thumbnails, auto-set LCP
      add_action('template_redirect', array($this, 'buffer_start_process_images'), 9000);
      add_action('shutdown', array($this, 'buffer_end_process_images'), PHP_INT_MAX);

      // Add an HTML element that will hold the settings for auto-setting image thumbnails
      add_action('wp_head', array($this, 'auto_set_image_thumbs_settings'));

      // CSS for Images options: placeholder, mark preloaded img, fade in
      add_action('wp_head', array($this, 'css_for_image_lazy_and_preload'));

      // CSS for Background images options
      add_action('wp_head', array($this, 'css_for_loading_background_images'));

      // Set attributes from gathered image data
      add_action('template_redirect', array($this, 'buffer_start_for_setting_attributes_from_gathered_image_data'), PHP_INT_MAX);
      add_action('shutdown', array($this, 'buffer_end_for_setting_attributes_from_gathered_image_data'), PHP_INT_MAX);

      // CSS for marking the LCP
      add_action('wp_head', array($this, 'css_for_marking_the_lcp'));

      // Background images: lazy load, preload, auto-set thumbnails, auto-set LCP
      add_action('template_redirect', array($this, 'buffer_start_for_background_images'), 9000);
      add_action('shutdown', array($this, 'buffer_end_for_background_images'), PHP_INT_MAX);

      // Add an HTML element that will hold the settings for auto-setting image backgropund thumbnails
      add_action('wp_head', array($this, 'auto_set_background_image_thumbs_settings'));

      // Lazy Load iFrames
      add_action('template_redirect', array($this, 'buffer_start_lazyload_iframes'), 9999);
      add_action('shutdown', array($this, 'buffer_end_lazyload_iframes'), PHP_INT_MAX);

      // CSS for iFrames options: placeholder, mark excluded
      add_action('wp_head', array($this, 'css_for_lazy_loaded_iframe'));

      // Lazy Load Videos
      add_action('template_redirect', array($this, 'buffer_start_lazyload_videos'), 9999);
      add_action('shutdown', array($this, 'buffer_end_lazyload_videos'), PHP_INT_MAX);

      // CSS for Videos options: placeholder, mark excluded, mark preloaded thumbnail
      add_action('wp_head', array($this, 'css_for_lazy_loaded_videos'));

      // Adds a dynamic class to featured images based on the type of post (including CPTs)
      add_filter('post_thumbnail_html', array($this, 'add_class_to_featured_images'), 10, 5);
      // Attach initialization to the 'plugins_loaded' action to ensure WooCommerce is loaded
      add_action('plugins_loaded', array($this, 'initialize'));
    }
  }

  public function initialize()
  {
    // Check if WooCommerce is active, then add WooCommerce-specific hooks
    if (class_exists('WooCommerce')) {
      add_filter('woocommerce_single_product_image_thumbnail_html', array($this, 'add_class_to_main_woo_product_image'));
    }
  }


  // Run these CSS rules at all times
  public function css_to_run_always()
  {
?>
    <style id="always-on-superspeed" type="text/css">
      *:where([has_background_image]) {
        background-size: cover;
      }

      html body [is_lcp_mobile],
      html body [is_lcp_tablet],
      html body [is_lcp_mobile] {
        background-size: cover;
      }

      img {
        max-width: 100%;
        max-height: 100%;
        height: auto;
        width: 100%;
      }
    </style>
  <?php
  }

  // Define additional image sizes / thumbnails
  function add_custom_image_sizes()
  {

    // Add new image sizes
    add_image_size('x_small', 256, 256, false);
    add_image_size('small', 384, 384, false);
    add_image_size('x_medium', 512, 512, false);
    add_image_size('xx_medium', 768, 768, false);
    // Large: 1024x1024
    add_image_size('x_large', 1536, 1536, false);
    add_image_size('xx_large', 2048, 2048, false);
    add_image_size('xxx_large', 4096, 4096, false);
  }

  // Remove image sizes
  function remove_image_thumbnails($sizes)
  {
    // Define the sizes to keep
    $keep_sizes = array(
      // Native
      'thumbnail' => __('Thumbnail'),
      'medium' => __('Medium'),
      'large' => __('Large'),

      // Custom
      'x_small' => __('X-Small'),
      'small' => __('Small'),
      'x_medium' => __('X-Medium'),
      'xx_medium' => __('XX-Medium'),
      'x_large' => __('X-Large'),
      'xx_large' => __('XX-Large'),
      'xxx_large' => __('XXX-Large'),
    );

    // Create a new array to store the filtered sizes
    $filtered_sizes = array();

    // Add any new image sizes you want to include
    $filtered_sizes['new_size_1'] = __('thumbnail');
    $filtered_sizes['new_size_2'] = __('medium');
    $filtered_sizes['new_size_3'] = __('large');

    $filtered_sizes['new_size_4'] = __('xxx_large');
    $filtered_sizes['new_size_5'] = __('xx_large');
    $filtered_sizes['new_size_6'] = __('x_large');
    $filtered_sizes['new_size_8'] = __('xx_medium');
    $filtered_sizes['new_size_9'] = __('x_medium');
    $filtered_sizes['new_size_10'] = __('small');
    $filtered_sizes['new_size_11'] = __('x_small');


    // Loop through the $sizes array and keep only the specified sizes
    foreach ($sizes as $size => $label) {
      if (isset($keep_sizes[$size])) {
        $filtered_sizes[$size] = $label;
      }
    }

    return $filtered_sizes;
  }


  // Is used in many functions
  // Helper function to check if an element (or its ancestors) has a specific class
  private function hasParentWithClass($element, $classList)
  {
    // Exit if we could not get the current URL. Skipping static resource because it is not a page.
    $current_url = Utility::get_current_page_url();
    if (!$current_url) {
      return;
    }

    $parent = $element->parentNode;
    while ($parent && $parent->nodeType == XML_ELEMENT_NODE) {
      if ($parent->hasAttribute('class')) {
        $parent_classes = explode(' ', $parent->getAttribute('class'));

        // Using array_intersect to check for common classes
        if (array_intersect($classList, $parent_classes)) {
          return true;
        }
      }
      $parent = $parent->parentNode;
    }
    return false;
  }


  // Helper functions for preloading images
  private function add_preload_link($img, $preloadImages, $originalSrc)
  {
    // Get the original 'src' attribute and URL-decode it
    $originalSrc = urldecode($img->getAttribute('src'));

    // Ensure the source is an absolute URL
    if (substr($originalSrc, 0, 1) === '/') {
      $originalSrc = home_url($originalSrc);
    }

    // Remove the 'loading' attribute to prevent default browser lazy loading
    $img->removeAttribute('loading');

    // We preload the smallest thumbnail, and later set a higher reolution image
    $smallestThumb = self::get_smallest_thumbnail($originalSrc);
    $img->setAttribute('src', $smallestThumb);

    // If we have data from the database
    if (!$this->imageDataAvailable) {
      $img->setAttribute('img_preloaded', '');

      // Add a preload link for the chosen image. Will be loaded for all screen sizes
      $preloadImages .= "\n" . '<link rel="preload" as="image" fetchpriority="high" href="' . $smallestThumb . '"></link>';
    } else {
      $visibleOnDesktop = $img->hasAttribute('visible_on_desktop');
      $visibleOnTablet = $img->hasAttribute('visible_on_tablet');
      $visibleOnMobile = $img->hasAttribute('visible_on_mobile');

      // If the image is on every screen size
      if ($visibleOnDesktop && $visibleOnTablet && $visibleOnMobile) {

        // Add a preload link for the chosen image
        $preloadImages .= "\n" . '<link rel="preload" as="image" fetchpriority="high" href="' . $smallestThumb . '"></link>';
        $img->setAttribute('img_preloaded_dekstop', '');
        $img->setAttribute('img_preloaded_tablet', '');
        $img->setAttribute('img_preloaded_mobile', '');
      } else {

        // Preload the image only on the screen size where it is visible
        if ($visibleOnDesktop) {
          $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(min-width: 800.1px)'>";
          $img->setAttribute('img_preloaded_dekstop', '');
        }
        if ($visibleOnTablet) {
          $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(min-width: 400.1px) and (max-width: 800px)'>";
          $img->setAttribute('img_preloaded_tablet', '');
        }
        if ($visibleOnMobile) {
          $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(max-width: 400px)'>";
          $img->setAttribute('img_preloaded_mobile', '');
        }
      }
    }

    // If we are not auto-setting thumbnails and not using the default WP 'srcset', restore the original size with JS on the 'load' event 
    if (!$img->getAttribute('ss-srcset') && !$img->getAttribute('srcset')) {
      $img->setAttribute('ss-original-src-img', $originalSrc);
    }

    // If we are not auto setting thumbnails, delay the default 'srcset' to load on the 'load' event 
    if (!$img->getAttribute('ss-srcset') && $img->getAttribute('srcset')) {
      $originalSrcset = $img->getAttribute('srcset');
      $img->removeAttribute('srcset');
      $img->setAttribute('data-srcset', $originalSrcset);
    }

    global $ss_global_variable_for_preloading_images;
    $ss_global_variable_for_preloading_images .= $preloadImages;
  }


  // Helper function to get the smallest thumbnail
  public static function get_smallest_thumbnail($imageUrl)
  {
    // URL-decode it
    $imageUrl = urldecode($imageUrl);

    $imageId = attachment_url_to_postid($imageUrl);
    // Check if the image ID is not found
    if ($imageId === 0) {
      $fullSizeSrc = preg_replace('/-\d+x\d+(?=\.\w+$)/', '', $imageUrl);
      $imageId = attachment_url_to_postid($fullSizeSrc);

      // If we still didn't get the image ID, return the original URL
      if ($imageId === 0) {
        return $imageUrl;
      }
    }

    // Retrieve full image and x_small image URLs
    $imageFull = wp_get_attachment_image_src($imageId, 'full')[0];
    $smallestThumb = wp_get_attachment_image_src($imageId, 'x_small')[0];

    if ($smallestThumb === $imageFull) {
      $smallestThumb = wp_get_attachment_image_src($imageId, 'medium')[0];
    }

    return $smallestThumb;
  }


  /*** IMAGES: LAZY LOAD, PRELOAD, AUTO-SET THUMBNAILS ***/

  public function buffer_start_process_images()
  {

    // Main enable and auto-preload LCP are off, exit
    // For performance reasons, we are using one HTML parsing for many functions
    if (get_option('lazy_load_images') !== '1' && get_option('auto_preload_lcp') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      return;
    }

    // If the user has enabled auto-setting Used CSS
    // If the CSS file is missing
    // If the page is new or edited
    // And if we should be getting Used CSS on this page
    if (Utility::getting_used_css_data()) {
      return;
    }

    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    ob_start(array($this, "process_images"));
    $this->buffering_started_lazyload_images = true;
  }


  public function process_images($html)
  {
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    $images = $dom->getElementsByTagName('img');


    // Do not preload the WordPress admin avatar
    $always_exclude_classes = ['avatar-26', 'avatar-64'];

    $auto_preload_lcp = get_option('auto_preload_lcp') === '1';
    if ($auto_preload_lcp) {
      $exclude_home_lcp = get_option('lcp_prelod_exclude_home_cbx') === '1';
      $pages_exclude_lcp = get_option('lcp_prelod_exclude_pages_cbx') === '1';
      if ($pages_exclude_lcp) {
        $pages_exclude_list_lcp = get_option('lcp_prelod_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_pages_textarea'))) : [];
      }

      $exclude_posts = get_option('lcp_prelod_exclude_posts') === '1';
      $exclude_products = get_option('lcp_prelod_exclude_products') === '1';
      $exclude_cpts = get_option('lcp_prelod_exclude_cpts') === '1';
      if ($exclude_cpts) {
        $exclude_cpts_list = get_option('lcp_prelod_exclude_cpts_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_pages_textarea'))) : [];
      }

      $lcp_prelod_exclude_selector = get_option('lcp_prelod_exclude_selector') === '1';
      if ($lcp_prelod_exclude_selector) {
        $lcp_prelod_exclude_selector_textarea = get_option('lcp_prelod_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_selector_textarea'))) : [];
      }
      $lcp_prelod_exclude_parent_selector = get_option('lcp_prelod_exclude_parent_selector') === '1';
      if ($lcp_prelod_exclude_parent_selector) {
        $lcp_prelod_exclude_parent_selector_textarea = get_option('lcp_prelod_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_parent_selector_textarea'))) : [];
      }
    }

    $images_main_enable = get_option('lazy_load_images') === '1';


    $lazy_load_images_checkbox = get_option('lazy_load_images_checkbox') === '1';
    if ($lazy_load_images_checkbox) {
      $lazy_image_exclude_pages_cbx = get_option('lazy_image_exclude_pages_cbx') === '1';
      $lazy_image_exclude_home = get_option('lazy_image_exclude_home') === '1';
      $lazy_image_exclude_pages_textarea = get_option('lazy_image_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lazy_image_exclude_selector_textarea'))) : [];
      $lazy_image_exclude_selector = get_option('lazy_image_exclude_selector') === '1';
      $lazy_image_exclude_selector_textarea = get_option('lazy_image_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_image_exclude_selector_textarea'))) : [];
      $lazy_image_exclude_parent_selector = get_option('lazy_image_exclude_parent_selector') === '1';
      $lazy_image_exclude_parent_selector_textarea = get_option('lazy_image_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_image_exclude_parent_selector_textarea'))) : [];

      $imagesPlaceholderRadio = get_option('imagesPlaceholderRadio');
      $placeholder_gray = $imagesPlaceholderRadio === 'lazy_image_placeholder_gray';
      $placeholder_low_rez = $imagesPlaceholderRadio === 'lazy_image_placeholder_low_rez';
      $placeholder_none = $imagesPlaceholderRadio === 'lazy_image_placeholder_none';

      $lazy_img_placeholder_exclude_selector = get_option('lazy_img_placeholder_exclude_selector') === '1';
      $lazy_img_placeholder_exclude_selector_textarea = get_option('lazy_img_placeholder_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_placeholder_exclude_selector_textarea'))) : [];
      $lazy_img_placeholder_exclude_parent_selector = get_option('lazy_img_placeholder_exclude_parent_selector') === '1';
      $lazy_img_placeholder_exclude_parent_selector_textarea = get_option('lazy_img_placeholder_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_placeholder_exclude_parent_selector_textarea'))) : [];

      $img_load_smallest_thumbnail_exclude_selector = get_option('img_load_smallest_thumbnail_exclude_selector') === '1';
      $img_load_smallest_thumbnail_exclude_selector_textarea = get_option('img_load_smallest_thumbnail_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('img_load_smallest_thumbnail_exclude_selector_textarea'))) : [];
      $img_load_smallest_thumbnail_exclude_parent_selector = get_option('img_load_smallest_thumbnail_exclude_parent_selector') === '1';
      $img_load_smallest_thumbnail_exclude_parent_selector_textarea = get_option('img_load_smallest_thumbnail_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('img_load_smallest_thumbnail_exclude_parent_selector_textarea'))) : [];
    }

    $image_auto_set_thumbnails = get_option('image_auto_set_thumbnails') === '1';
    if ($image_auto_set_thumbnails) {
      $image_auto_set_exclude_selector = get_option('image_auto_set_exclude_selector') === '1';
      $image_auto_set_exclude_selector_textarea = get_option('image_auto_set_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('image_auto_set_exclude_selector_textarea'))) : [];
      $image_auto_set_exclude_parent_selector = get_option('image_auto_set_exclude_parent_selector') === '1';
      $image_auto_set_exclude_parent_selector_textarea = get_option('image_auto_set_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('image_auto_set_exclude_parent_selector_textarea'))) : [];
    }

    $lazy_img_preload_first_enable = get_option('lazy_img_preload_first_enable') === '1';
    if ($lazy_img_preload_first_enable) {
      $preloadCount = floatval(get_option('lazy_img_preload_first_x_select', 1)); // Number of images, from the top, that should be preloaded
      $exclude_selector = get_option('img_first_x_exlude_selector') === '1';
      $exclude_parent_selector = get_option('img_first_x_exlude_parent_selector') === '1';
      $exclude_selector_list = get_option('img_first_x_exlude_selector_textarea') ? array_map('trim', explode("\n", get_option('img_first_x_exlude_selector_textarea'))) : [];
      $exclude_parent_selector_list = get_option('img_first_x_exlude_selector_parent_textarea') ? array_map('trim', explode("\n", get_option('img_first_x_exlude_selector_parent_textarea'))) : [];
      $exclude_home = get_option('lazy_img_exclude_home_cbx') === '1';
      $include_home = get_option('lazy_img_include_home_cbx') === '1';
      $Radio_img_first_x_select_pages = get_option('img_first_x_select_pages_radio');
      $all_pages = $Radio_img_first_x_select_pages === 'img_first_x_all_pages';
      $exclude_pages = $Radio_img_first_x_select_pages === 'img_first_x_exclude_pages';
      $include_pages = $Radio_img_first_x_select_pages === 'img_first_x_include_pages';
      $exclude_list = get_option('img_first_x_exlude_pages_textarea') ? array_map('trim', explode("\n", get_option('img_first_x_exlude_pages_textarea'))) : [];
      $include_list = get_option('img_first_x_include_pages_textarea') ? array_map('trim', explode("\n", get_option('img_first_x_include_pages_textarea'))) : [];
    }

    $lazy_img_preload_selector_enable = get_option('lazy_img_preload_selector_enable') === '1';
    if ($lazy_img_preload_selector_enable) {
      $preload_featured_img_blog = get_option('preload_featured_img_blog') === '1';
      $preload_featured_img_product = get_option('preload_featured_img_product') === '1';
      $preload_selector = get_option('lazy_img_preload_selector') === '1';
      if ($preload_selector) {
        $preload_selector_list = get_option('lazy_img_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_selector_textarea'))) : [];
      }
      $preload_parent_selector = get_option('lazy_img_preload_parent_selector') === '1';
      if ($preload_parent_selector) {
        $preload_parent_selector_list = get_option('lazy_img_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_parent_selector_textarea'))) : [];
      }
    }

    $preloadImages = null;

    /// SET WIDTH AND HEIGHT FOR IMAGES, AND LOAD THEM ASYNCHRONOUSLY ///

    foreach ($images as $img) {

      // Load images asynchronously
      $img->setAttribute('decoding', 'async');

      // Skip if width and height are already set
      if ($img->hasAttribute('width') && $img->hasAttribute('height')) {
        continue;
      }

      // Get the original 'src' attribute and URL-decode it
      $originalSrc = urldecode($img->getAttribute('src'));

      // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
      if (
        empty($originalSrc) ||
        $originalSrc === 'null' ||
        strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg' ||
        strpos($originalSrc, 'base64') !== false
      ) {
        continue;
      }


      // Ensure the source is an absolute URL
      if (substr($originalSrc, 0, 1) === '/') {
        $originalSrc = home_url($originalSrc);
      }

      // Check if the URL is valid
      if (filter_var($originalSrc, FILTER_VALIDATE_URL) === false) {
        continue;
      }

      // Initialize CURL to check the image's existence without downloading it
      $ch = curl_init($originalSrc);
      curl_setopt($ch, CURLOPT_NOBODY, true);
      curl_exec($ch);
      $retCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
      curl_close($ch);

      // Check if the image exists
      if ($retCode != 200) {
        // Image does not exist or is inaccessible
        continue;
      }

      // Now it's safe to get the image size
      $dimensions = @getimagesize($originalSrc);
      if ($dimensions) {
        $imageWidth = $dimensions[0];
        $imageHeight = $dimensions[1];

        // Set the width and height attributes
        $img->setAttribute('width', $imageWidth);
        $img->setAttribute('height', $imageHeight);
      }
    }


    /// AUTO-SET THUMBNAILS FOR IMAGES ///

    if ($image_auto_set_thumbnails && $images_main_enable) {

      foreach ($images as $img) {

        if (!$img->hasAttribute('src')) {
          // error_log('auto-set image does not have the src attribute');
          continue;
        }

        // Get the original 'src' attribute and URL-decode it
        $originalSrc = urldecode($img->getAttribute('src'));

        // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
        if (
          empty($originalSrc)
          || $originalSrc === 'null'
          || strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg'
          || strpos($originalSrc, 'base64') !== false
        ) {
          continue;
        }

        // Get all classes from the image
        $classes = explode(' ', $img->getAttribute('class'));

        // Exclude by class
        if ($image_auto_set_exclude_selector && array_intersect($image_auto_set_exclude_selector_textarea, $classes)) {
          continue;
        }

        // Exclude by parent class
        if ($image_auto_set_exclude_parent_selector && $this->hasParentWithClass($img, $image_auto_set_exclude_parent_selector_textarea)) {
          continue;
        }

        // If the URL starts with a slash, add the domain
        if (substr($originalSrc, 0, 1) === '/') {
          $originalSrc = home_url($originalSrc);
        }

        // Attempt to get the image ID from the URL
        $imageId = attachment_url_to_postid($originalSrc);

        // If the image ID wasn't found, adjust the URL and try again
        if ($imageId === 0) {
          $fullSizeSrc = preg_replace('/-\d+x\d+(?=\.\w+$)/', '', $originalSrc);
          $imageId = attachment_url_to_postid($fullSizeSrc);
        }
        // error_log('$imageId: ' . $imageId);

        // If we still didn't get the image ID, do nothing, and exit loop
        if ($imageId === 0) {
          $img->setAttribute('auto_set_failed', 'no-id');
          continue;
        }

        // Map each thumbnail top show above the image size in pixels
        $sizesToLabels = [
          'full' =>       '8192',
          'xxx_large' =>  '4096', // 4096x4096
          'xx_large' =>   '2048', // 2048x2048
          'x_large' =>    '1536', // 1536x1536
          'large' =>      '1024', // 1024x1024
          'xx_medium' =>  '768',  // 768x768
          'x_medium' =>   '512',  // 512x512
          'small' =>      '256',  // 384x384
          'x_small' =>    '0'     // 256x256
        ];

        // Retrieve full image and x_small image URLs
        $imageFull = wp_get_attachment_image_src($imageId, 'full')[0];
        $x_small = wp_get_attachment_image_src($imageId, 'x_small')[0];

        // error_log('$imageFull' . $imageFull);
        // error_log('$x_small' . $x_small);

        $srcsetParts = [];
        foreach ($sizesToLabels as $size => $label) {
          $imageSrc = wp_get_attachment_image_src($imageId, $size);

          // Check if the current image size URL matches the x_small URL, fall back to default sizes if necessary
          if ($x_small === $imageFull && isset($imageSrc[0])) {
            // error_log('Set fallback thumbs');
            switch ($size) {
              case 'xxx_large':
                $imageSrc = wp_get_attachment_image_src($imageId, 'full');
                break;
              case 'xx_large':
              case 'x_large':
                $imageSrc = wp_get_attachment_image_src($imageId, 'large');
                break;
              case 'xx_medium':
              case 'x_medium':
                $imageSrc = wp_get_attachment_image_src($imageId, 'medium_large');
                break;
              case 'small':
              case 'x_small':
                $imageSrc = wp_get_attachment_image_src($imageId, 'medium');
                break;
            }
          }

          // Add to srcset string if URL is available.
          if ($imageSrc && isset($imageSrc[0])) {
            // error_log('Thumbail: ' . $imageSrc[0]);
            $srcsetParts[] = $imageSrc[0] . '|' . $label;
          }
        }

        $ssSrcsetString = implode(',', $srcsetParts);
        // error_log('SourceSet: ' . print_r($srcsetParts));
        // Replacing this with a custom attribute
        $img->removeAttribute('srcset');
        $img->removeAttribute('sizes');

        // Set the ss-srcset attribute on the image
        $img->setAttribute('ss-srcset', $ssSrcsetString);

        // If images are not lazy-loaded, set the smallest thumbnail for the source, and an attribute to resore it from for excluded images
        if (!$lazy_load_images_checkbox) {
          $smallestThumb = self::get_smallest_thumbnail($originalSrc);
          $img->setAttribute('src', $smallestThumb);
        }
      }
    }


    /// PRELOAD FIRST FEW IMAGES ///

    if ($lazy_img_preload_first_enable && $images_main_enable) {

      // Flag for preloading first few images on this page
      $preload_first_few_on_this_page = false;

      if (!$all_pages) {
        $is_home = Utility::is_homepage();
      }

      // Run on all pages
      if ($all_pages) {
        $preload_first_few_on_this_page = true;
      }

      // Run on all except these excluded pages
      elseif ($exclude_pages) {
        if ($is_home && $exclude_home) {
          // Don't run on the homepage
          $preload_first_few_on_this_page = false;
        } else {
          // Don't run on pages from the list
          if (!empty($exclude_list)) {
            $current_URL = Utility::get_current_page_url();
            foreach ($exclude_list as $string) {
              $normalizedString = Utility::normalizeUrlString($string);
              if (stripos($current_URL, $normalizedString) !== false) {
                $preload_first_few_on_this_page = false;
                break;
              }
            }
          }
        }
      }

      // Run only on specified pages
      elseif ($include_pages) {
        if ($is_home && $include_home) {
          // Run on Home page
          $preload_first_few_on_this_page = true;
        } else {
          if (!empty($include_list)) {
            $current_URL = Utility::get_current_page_url();
            foreach ($include_list as $string) {
              $normalizedString = Utility::normalizeUrlString($string);
              if (stripos($current_URL, $normalizedString) !== false) {
                $preload_first_few_on_this_page = true;
                break;
              }
            }
          }
        }
      }


      // Preloading first few images on this page
      if ($preload_first_few_on_this_page && $images_main_enable) {

        // Iterate over each image element
        $counter = 0;

        foreach ($images as $img) {

          // Get the original 'src' attribute and URL-decode it
          $originalSrc = urldecode($img->getAttribute('src'));

          // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
          if (
            empty($originalSrc) ||
            $originalSrc === 'null' ||
            strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg' ||
            strpos($originalSrc, 'base64') !== false
          ) {
            continue;
          }


          $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

          // Exclude images with the "always_exclude_classes"
          if (array_intersect($always_exclude_classes, $classes)) {
            continue;
          }

          // Exclude images by selector
          if ($exclude_selector && array_intersect($exclude_selector_list, $classes)) {
            continue;
          }

          // Exclude images by parent selector
          if ($exclude_parent_selector && $this->hasParentWithClass($img, $exclude_parent_selector_list)) {
            continue;
          }

          // Count the images
          if ($counter < $preloadCount) {
            $this->add_preload_link($img, $preloadImages, $originalSrc);
            $counter++;
            continue;
          }
        }
      }
    }

    /// PRELOAD BY SELECTOR ///

    // We preload only the first occurance of an image for each class
    // We preload only the first child image of a parent
    if ($lazy_img_preload_selector_enable && $images_main_enable) {
      // A list to keep track of already processed classes
      $processed_classes = [];
      $processed_parent_classes = [];

      foreach ($images as $img) {

        // Get the original 'src' attribute and URL-decode it
        $originalSrc = $img->getAttribute('src');

        // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
        if (
          empty($originalSrc) ||
          $originalSrc === 'null' ||
          strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg' ||
          strpos($originalSrc, 'base64') !== false
        ) {
          continue;
        }

        $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

        if ($preload_featured_img_blog) {
          $preload_selector_list[] = 'featured_image_post';
        }
        if ($preload_featured_img_product) {
          $preload_selector_list[] = 'featured_image_woo_product';
        }

        // PRELOAD BY SELECTOR
        if (
          $preload_selector && !empty($preload_selector_list) ||
          $preload_featured_img_blog ||
          $preload_featured_img_product
        ) {
          foreach ($preload_selector_list as $selector) {
            if (in_array($selector, $classes) && !in_array($selector, $processed_classes)) {
              $this->add_preload_link($img, $preloadImages, $originalSrc);
              $processed_classes[] = $selector;
              break; // break out of the selector loop
            }
          }
        }

        // PRELOAD BY PARENT SELECTOR
        if ($preload_parent_selector && !empty($preload_parent_selector_list)) {
          foreach ($preload_parent_selector_list as $selector) {
            if ($this->hasParentWithClass($img, [$selector]) && !in_array($selector, $processed_parent_classes)) {
              $this->add_preload_link($img, $preloadImages, $originalSrc);
              $processed_parent_classes[] = $selector;
              break; // break out of the selector loop if a match is found
            }
          }
        }
      }
    }

    /// AUTO-PRELOAD LCP ///

    if ($auto_preload_lcp && $this->imageDataAvailable && (get_option('lcp_enable') === '1')) {

      // Set flag for excluding preloading on these pages
      $preload_lcp_on_this_page = true;

      if ($exclude_home_lcp) {
        $is_home_lcp = Utility::is_homepage();
        if ($is_home_lcp) {
          $preload_lcp_on_this_page = false;
        }
      }

      // Run on all except these excluded pages
      elseif ($pages_exclude_lcp) {
        // Don't run on pages from the list
        if (!empty($pages_exclude_list_lcp)) {
          $current_Url_lcp = Utility::get_current_page_url();
          foreach ($pages_exclude_list_lcp as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_Url_lcp, $normalizedString) !== false) {
              $preload_lcp_on_this_page = false;
            }
          }
        }
      }

      // Do not run on single blog posts
      elseif ($exclude_posts && is_single() && get_post_type() === 'post') {
        $preload_lcp_on_this_page = false;
      }

      // Do not run on single product pages
      elseif ($exclude_products && is_singular('product')) {
        $preload_lcp_on_this_page = false;
      }

      // Excluding by CPT ingle post
      // We don't need a foreach loop because the is_singular() function can accept an array of post type names directly
      elseif ($exclude_cpts && !empty($exclude_cpt_list) && is_singular($exclude_cpt_list)) {
        $preload_lcp_on_this_page = false;
      }


      // Preloading LCP images on this page
      if ($preload_lcp_on_this_page) {

        // There can be a maxinum of 3 LCP images, one for each scren size
        $counterLcp = 0;
        $preloadCountLcp = 3;

        foreach ($images as $img) {

          $isLcpDesktop = $img->hasAttribute('is_lcp_desktop');
          $isLcpTablet = $img->hasAttribute('is_lcp_tablet');
          $isLcpMobile = $img->hasAttribute('is_lcp_mobile');

          if (!$isLcpDesktop && !$isLcpTablet && !$isLcpMobile) {
            continue;
          }

          // Get the original 'src' attribute and URL-decode it
          $originalSrc = urldecode($img->getAttribute('src'));

          // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
          if (
            empty($originalSrc) ||
            $originalSrc === 'null' ||
            strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg' ||
            strpos($originalSrc, 'base64') !== false
          ) {
            continue;
          }

          $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

          // Exclude images with the "always_exclude_classes"
          if (array_intersect($always_exclude_classes, $classes)) {
            continue;
          }

          // Exclude images by selector
          if ($lcp_prelod_exclude_selector && array_intersect($lcp_prelod_exclude_selector_textarea, $classes)) {
            continue;
          }

          // Exclude images by parent selector
          if ($lcp_prelod_exclude_parent_selector && $this->hasParentWithClass($img, $lcp_prelod_exclude_parent_selector_textarea)) {
            continue;
          }

          // Count the images
          if ($counterLcp < $preloadCountLcp) {

            // Get the original 'src' attribute and URL-decode it
            $originalSrc = urldecode($img->getAttribute('src'));

            // Ensure the source is an absolute URL
            if (substr($originalSrc, 0, 1) === '/') {
              $originalSrc = home_url($originalSrc);
            }

            // Remove the 'loading' attribute to prevent default browser lazy loading
            $img->removeAttribute('loading');

            // We preload the smallest thumbnail, and later set a higher reolution image
            $smallestThumb = self::get_smallest_thumbnail($originalSrc);
            $img->setAttribute('src', $smallestThumb);

            // If the image is on every screen size
            if ($isLcpDesktop && $isLcpTablet && $isLcpMobile) {

              // Add a preload link for the chosen image
              $preloadImages .= "\n" . '<link rel="preload" as="image" fetchpriority="high" href="' . $smallestThumb . '"></link>';
              $img->setAttribute('img_preloaded_dekstop', '');
              $img->setAttribute('img_preloaded_tablet', '');
              $img->setAttribute('img_preloaded_mobile', '');
            } else {

              // Preload the image only on the screen size where it is visible
              if ($isLcpDesktop) {
                $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(min-width: 800.1px)'>";
                $img->setAttribute('img_preloaded_dekstop', '');
              }
              if ($isLcpTablet) {
                $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(min-width: 400.1px) and (max-width: 800px)'>";
                $img->setAttribute('img_preloaded_tablet', '');
              }
              if ($isLcpMobile) {
                $preloadImages .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $smallestThumb . "' media='(max-width: 400px)'>";
                $img->setAttribute('img_preloaded_mobile', '');
              }
            }

            // If we are not auto-setting thumbnails and not using the default WP 'srcset', restore the original size with JS
            if (!$img->getAttribute('ss-srcset') && !$img->getAttribute('srcset')) {
              $img->setAttribute('ss-original-src-img', $originalSrc);
            }

            // If we are not auto setting thumbnails, delay the default 'srcset' to load on the 'load' event 
            if (!$img->getAttribute('ss-srcset') && $img->getAttribute('srcset')) {
              $originalSrcset = $img->getAttribute('srcset');
              $img->removeAttribute('srcset');
              $img->setAttribute('data-srcset', $originalSrcset);
            }

            // If the same image is the LCP for each screen, stop searching afer you find it
            if ($isLcpDesktop && $isLcpTablet &&  $isLcpMobile) {
              break;
            } else {
              $counterLcp++;
              continue;
            }
          }
        }
        global $ss_global_variable_for_preloading_images;
        $ss_global_variable_for_preloading_images .= $preloadImages;
      }
    }


    /// LAZY LOAD IMAGES ///
    if ($lazy_load_images_checkbox && $images_main_enable) {

      // Set flag
      $lazy_images_on_this_page = true;

      // Exclude on Home page if the option is selected
      if ($lazy_image_exclude_home && Utility::is_homepage()) {
        $lazy_images_on_this_page = false;
      }

      // Run on all except these excluded pages
      elseif ($lazy_image_exclude_pages_cbx && !empty($lazy_image_exclude_pages_textarea)) {
        // Don't run on pages from the list
        $current_Url_lazy_img = Utility::get_current_page_url();
        foreach ($lazy_image_exclude_pages_cbx as $string) {
          $normalizedString = Utility::normalizeUrlString($string);
          if (stripos($current_Url_lazy_img, $normalizedString) !== false) {
            $lazy_images_on_this_page = false;
          }
        }
      }

      // Lazy load images on this page
      if ($lazy_images_on_this_page) {

        foreach ($images as $img) {

          if ($this->imageDataAvailable) {
            if (
              $img->hasAttribute('img_preloaded_dekstop') ||
              $img->hasAttribute('img_preloaded_tablet') ||
              $img->hasAttribute('img_preloaded_mobile')
            ) {
              continue;
            }
          } else {
            if ($img->hasAttribute('img_preloaded')) {
              continue;
            }
          }

          // Get the original 'src' attribute and URL-decode it
          $originalSrc = urldecode($img->getAttribute('src'));

          // Check if 'src' attribute is not empty, if the image is an SVG, or if it's a base64 encoded image
          if (
            empty($originalSrc) ||
            $originalSrc === 'null' ||
            strtolower(pathinfo($originalSrc, PATHINFO_EXTENSION)) === 'svg' ||
            strpos($originalSrc, 'base64') !== false
          ) {
            continue;
          }

          $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

          // Exclude by class
          if ($lazy_image_exclude_selector && array_intersect($lazy_image_exclude_selector_textarea, $classes)) {
            $img->removeAttribute('loading');
            continue;
          }

          // Exclude by parent class
          if ($lazy_image_exclude_parent_selector && $this->hasParentWithClass($img, $lazy_image_exclude_parent_selector_textarea)) {
            $img->removeAttribute('loading');
            continue;
          }

          // Native browser lazy loading
          $img->setAttribute('loading', 'lazy');
          $img->removeAttribute('fetchpriority');

          // Custom marker for CSS and JS
          $img->setAttribute('lazy_loaded_img', '');
        }
      }
    }


    /// GRAY & NONE PLACEHOLDER FOR LAZY LOADED IMAGES ///

    // If the option to add the gray placeholder is enabled, wrap the images and add the gray placeholder div
    // We do the same for No placeholder, but with a transparent placeholder
    if ($lazy_load_images_checkbox && $images_main_enable) {
      if ($placeholder_gray || $placeholder_none) {

        // error_log('$placeholder_gray: ' . $placeholder_gray);
        // Collects images into an array to escape errors made by DOM manipulation
        $images = iterator_to_array($dom->getElementsByTagName('img'));

        foreach ($images as $img) {

          // Exclude images that are not lazy loaded
          if (!$img->hasAttribute('lazy_loaded_img')) {
            continue;
          }

          $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

          // The placeholder Gray has exclusions, while placeholde Non does not
          if ($placeholder_gray) {

            // Exclude by class
            if ($lazy_img_placeholder_exclude_selector && array_intersect($lazy_img_placeholder_exclude_selector_textarea, $classes)) {
              continue;
            }

            // Exclude by parent class
            if ($lazy_img_placeholder_exclude_parent_selector && $this->hasParentWithClass($img, $lazy_img_placeholder_exclude_parent_selector_textarea)) {
              continue;
            }
          }

          // Set an identifier
          if ($placeholder_gray) {
            $img->setAttribute('img_placeholder', 'gray');
          } else {
            $img->setAttribute('img_placeholder', 'none');
          }

          // Check if $classes is an array and not empty
          $classString = '';
          if (is_array($classes) && !empty($classes)) {
            $classString = implode(' ', $classes);
          }

          // Add our class, with space before it, in case there are existing classes
          $classString .= ' ss-img-placeholder';

          $placeholderWidth = $img->getAttribute('width');
          $placeholderHeight = $img->getAttribute('height');
          $placeholderStyle = $img->getAttribute('style');

          // Gray SVG for the gray placeholder, and nothing for placeholder - None
          if ($placeholder_gray) {
            $placeholderSrc = 'data:image/svg+xml;utf8,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%3E%3Crect%20width%3D%2210000%22%20height%3D%2210000%22%20fill%3D%22gray%22%3E%3Canimate%20attributeName%3D%22fill%22%20values%3D%22lightgray%3B%20darkgray%3B%20lightgray%22%20dur%3D%222s%22%20repeatCount%3D%22indefinite%22%20%2F%3E%3C%2Frect%3E%3C%2Fsvg%3E';
          } else {
            $placeholderSrc = 'data:image/svg+xml;utf8,<svg xmlns="http://www.w3.org/2000/svg"><rect width="10000" height="10000" fill="transparent"/></svg>';
          }

          // Create the cloned image placeholder
          $clonedImagePlaceholder = $dom->createElement('img');

          // Set attributes
          $clonedImagePlaceholder->setAttribute('class', $classString);
          $clonedImagePlaceholder->setAttribute('width', $placeholderWidth);
          $clonedImagePlaceholder->setAttribute('height', $placeholderHeight);
          $clonedImagePlaceholder->setAttribute('style', $placeholderStyle);
          $clonedImagePlaceholder->setAttribute('src', $placeholderSrc);

          // Check if the original image has a next sibling
          if ($img->nextSibling) {

            // Place the cloned image placeholder after the original image
            $img->parentNode->insertBefore($clonedImagePlaceholder, $img->nextSibling);
          } else {

            // If the original image is the last child, append the cloned image
            $img->parentNode->appendChild($clonedImagePlaceholder);
          }
        }
      }
    }


    /// LOW RESOLUTION PLACEHOLDER FOR LAZY LOADED IMAGES ///

    // If the option to add the gray placeholder is enabled, wrap the images and add the gray placeholder div
    if ($lazy_load_images_checkbox && $images_main_enable) {

      if ($placeholder_low_rez) {
        // error_log('$placeholder_low_rez: ' . $placeholder_low_rez);

        foreach ($images as $img) {

          // Exclude images that are not lazy loaded
          if (!$img->hasAttribute('lazy_loaded_img')) {
            continue;
          }

          $classes = explode(' ', $img->getAttribute('class')); // Get all classes from the image.

          // Exclude by class
          if ($img_load_smallest_thumbnail_exclude_selector && array_intersect($img_load_smallest_thumbnail_exclude_selector_textarea, $classes)) {
            continue;
          }

          // Exclude by parent class
          if ($img_load_smallest_thumbnail_exclude_parent_selector && $this->hasParentWithClass($img, $img_load_smallest_thumbnail_exclude_parent_selector_textarea)) {
            continue;
          }

          // Get the original 'src' attribute and URL-decode it
          $originalSrc = urldecode($img->getAttribute('src'));

          // If the URL starts with a slash, add the domain
          if (substr($originalSrc, 0, 1) === '/') {
            $originalSrc = home_url($originalSrc);
          }

          // Attempt to get the image ID from the URL
          $imageId = attachment_url_to_postid($originalSrc);

          // If the image ID wasn't found, adjust the URL and try again
          if ($imageId === 0) {
            $fullSizeSrc = preg_replace('/-\d+x\d+(?=\.\w+$)/', '', $originalSrc);
            $imageId = attachment_url_to_postid($fullSizeSrc);
          }

          // If we still didn't get the image ID, do nothing, and exit loop
          if ($imageId === 0) {
            $img->setAttribute('low_rez_failed', 'no-id');
            continue;
          }

          $smallestThumb = self::get_smallest_thumbnail($originalSrc);
          // Set low resolution placeholder
          $img->setAttribute('src', $smallestThumb);

          // If we are not auto-setting thumbnails, save the original URL to revert with JS
          if (!$img->hasAttribute('ss-srcset')) {
            $dataSrc = $img->getAttribute('src');
            $img->setAttribute('data-src', $dataSrc);
          }

          // Set an identifier
          $img->setAttribute('img_placeholder', 'low-rez');

          // Do not lazy load, since we have the low resolution image on the screen
          // A higher resolution image will be set with JS
          $img->removeAttribute('loading');
        }
      }
    }



    // Return the complete HTML
    return $dom->saveHTML();
  }

  // Already performed the check if we are on the right page, and stored it in the variable '$this->buffering_started_critical_css'
  public function buffer_end_process_images()
  {
    if ($this->buffering_started_lazyload_images) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_lazyload_images = false;
    }
  }


  // Add an HTML element that will hold the settings for auto-setting image thumbnails
  public function auto_set_image_thumbs_settings()
  {

    // Main Enable
    if (get_option('lazy_load_images') !== '1') {
      return;
    }

    // Enable option
    if (get_option('image_auto_set_thumbnails') !== '1') {
      return;
    }

    $image_info = get_option('image_auto_set_thumbs_info') === '1';
    $image_custom_pixel_density_desktop = get_option('image_custom_pixel_density_desktop') === '1';
    $image_custom_pixel_density_tablet = get_option('image_custom_pixel_density_tablet') === '1';
    $image_custom_pixel_density_mobile = get_option('image_custom_pixel_density_mobile') === '1';

    $image_info_value = 'false';
    if ($image_info) {
      $image_info_value = 'true';
    }

    $custom_or_native_desktop = 'native';
    if ($image_custom_pixel_density_desktop) {
      $custom_or_native_desktop = 'custom';
    }

    $custom_or_native_tablet = 'native';
    if ($image_custom_pixel_density_tablet) {
      $custom_or_native_tablet = 'custom';
    }

    $custom_or_native_mobile = 'native';
    if ($image_custom_pixel_density_mobile) {
      $custom_or_native_mobile = 'custom';
    }

    $pixel_density_desktop = floatval(get_option('image_auto_set_fixed_pixels_desktop'));
    $pixel_density_tablet = floatval(get_option('image_auto_set_fixed_pixels_tablet'));
    $pixel_density_mobile = floatval(get_option('image_auto_set_fixed_pixels_mobile'));


    echo '<script type="application/json" id="superspeed-fixed-pixel-density-img">
    {
        "custom_or_native_desktop": "' . $custom_or_native_desktop . '",
        "pixel_density_desktop": ' . $pixel_density_desktop . ',
        "custom_or_native_tablet": "' . $custom_or_native_tablet . '",
        "pixel_density_tablet": ' . $pixel_density_tablet . ',
        "custom_or_native_mobile": "' . $custom_or_native_mobile . '",
        "pixel_density_mobile": ' . $pixel_density_mobile . ',
        "image_info": ' . $image_info_value . '
    }
    </script>
 ';
  }


  // CSS for Images options: placeholder, mark preloaded img, fade in
  public function css_for_image_lazy_and_preload()
  {

    // Main Enable
    if (get_option('lazy_load_images') !== '1') {
      return;
    }

    $lazy_load_images_checkbox = get_option('lazy_load_images_checkbox') === '1';
    $imagesPlaceholderRadio = get_option('imagesPlaceholderRadio');
    $placeholder_gray = $imagesPlaceholderRadio === 'lazy_image_placeholder_gray';
    $placeholder_none = $imagesPlaceholderRadio === 'lazy_image_placeholder_none';

    // Move lazy loaded image outside the viewport
    if (($placeholder_gray && $lazy_load_images_checkbox) || ($placeholder_none && $lazy_load_images_checkbox)) {
      echo '<style id="superspeed-move-lazy-images" type="text/css">
    /* We have added a wrap and child to be in place of the image
    The image is moved far away to the left */
    img[lazy_loaded_img]:not(.ss-image-loaded) {
        position: absolute !important;
        translate: -1000vw !important;
        width: revert-layer;
      }
      </style>';
    }

    // Visually mark lazy loaded images
    if (get_option('lazyloaded_img_visually_mark') === '1' && $lazy_load_images_checkbox) {
      echo '<style  id="superspeed-mark-lazy-images" type="delayed-style">
        .can-edit img[lazy_loaded_img] {
            outline: dashed 0.25rem blue !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
        }
      </style>';
    }

    // Visually mark preloaded images
    if (get_option('preloaded_img_visually_mark') === '1') {
      if ($this->imageDataAvailable) {
        echo '<style id="superspeed-mark-preloaded-images" type="delayed-style">
          @media screen and (min-width: 993px) {
            .can-edit img[img_preloaded_dekstop] {
              outline: dashed 0.25rem red !important;
              border: solid 0.25rem white !important;
              outline-offset: -0.25rem !important;
            }
          }

          @media screen and (min-width: 768px) and (max-width: 992px) {
            .can-edit img[img_preloaded_tablet] {
              outline: dashed 0.25rem red !important;
              border: solid 0.25rem white !important;
              outline-offset: -0.25rem !important;
            }
          }

          @media screen and (max-width: 767px) {
            .can-edit img[img_preloaded_mobile] {
              outline: dashed 0.25rem red !important;
              border: solid 0.25rem white !important;
              outline-offset: -0.25rem !important;
            }
          }

            </style>';
      } else {
        echo '<style id="superspeed-mark-preloaded-images" type="delayed-style">
          .can-edit img[img_preloaded] {
              outline: dashed 0.25rem red !important;
              border: solid 0.25rem white !important;
              outline-offset: -0.25rem !important;
          }
          </style>';
      }
    }

    // Fade-in images
    if (get_option('lazy_img_fadein') === '1' && $lazy_load_images_checkbox) {

      $fadeInDuration = floatval(get_option('lazy_img_fadein_duration', 1));

      // Always initialize these to avoid undefined variable errors
      $selectors_string_fadein = '';
      $parent_selectors_string_fadein = '';

      // Exclude fade in images by class
      if (get_option('lazy_img_fadein_exclude_selector') === '1') {
        $lazy_img_fadein_exclude_selector_textarea = get_option('lazy_img_fadein_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_fadein_exclude_selector_textarea'))) : [];

        $formatted_selectors = [];
        foreach ($lazy_img_fadein_exclude_selector_textarea as $selector) {
          $formatted_selectors[] = '.' . $selector;
        }

        if (!empty($formatted_selectors)) {
          $selectors_string_fadein = ', ' . implode(', ', $formatted_selectors);
        }
      }

      // Exclude fade in images by parent class
      if (get_option('lazy_img_fadein_exclude_parent_selector') === '1') {
        $lazy_img_fadein_exclude_parent_selector_textarea = get_option('lazy_img_fadein_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_img_fadein_exclude_parent_selector_textarea'))) : [];

        $formatted_parent_selectors = [];
        foreach ($lazy_img_fadein_exclude_parent_selector_textarea as $selector) {
          $formatted_parent_selectors[] = '.' . $selector . ' *';
        }

        if (!empty($formatted_parent_selectors)) {
          $parent_selectors_string_fadein = ', ' . implode(', ', $formatted_parent_selectors);
        }
      }

      if ($this->imageDataAvailable) {

        echo '<style id="superspeed-fade-in-lazy-loaded-images" type="text/css">
        @media screen and (min-width: 993px) {
        img[lazy_loaded_img]:not(.fade-in, [img_preloaded], [is_lcp_desktop], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          opacity: 0;
        }
        img[lazy_loaded_img].fade-in:not([img_preloaded], [is_lcp_desktop], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          animation-name: SSFadeInImage;
          animation-duration: ' . $fadeInDuration . 's;
          animation-timing-function: ease;
        }
      }

      @media screen and (min-width: 768px) and (max-width: 992px) {
        img[lazy_loaded_img]:not(.fade-in, [img_preloaded], [is_lcp_tablet], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          opacity: 0;
        }
        img[lazy_loaded_img].fade-in:not([img_preloaded], [is_lcp_tablet], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          animation-name: SSFadeInImage;
          animation-duration: ' . $fadeInDuration . 's;
          animation-timing-function: ease;
        }
      }

      @media screen and (max-width: 767px) {
        img[lazy_loaded_img]:not(.fade-in, [img_preloaded], [is_lcp_mobile], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          opacity: 0;
        }
        img[lazy_loaded_img].fade-in:not([img_preloaded], [is_lcp_mobile], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          animation-name: SSFadeInImage;
          animation-duration: ' . $fadeInDuration . 's;
          animation-timing-function: ease;
        }
      }

      /* The animation has only *from*, so that the element can have a custom opacity */
      @keyframes SSFadeInImage {
        from {
          opacity: 0;
        }
      }
    </style>';
      } else {
        echo '<style id="superspeed-fade-in-lazy-loaded-images" type="text/css">
  
        img[lazy_loaded_img]:not(.fade-in, [img_preloaded], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          opacity: 0;
        }
      
        img[lazy_loaded_img].fade-in:not([img_preloaded], [img_placeholder="gray"], [img_placeholder="low-rez"]' . $selectors_string_fadein . $parent_selectors_string_fadein . ') {
          animation-name: SSFadeInImage;
          animation-duration: ' . $fadeInDuration . 's;
          animation-timing-function: ease;
        }
      
        /* The animation has only *from*, so that the element can have a custom opacity */
        @keyframes SSFadeInImage {
          from {
            opacity: 0;
          }
        }
  
    </style>';
      }
    }
  }



  /*** CSS FOR LAZY LOADING AND PRELOADING BACKGROUD IMAGES ***/

  // Helper function for Background images CSS options
  public function css_for_loading_background_images()
  {

    // Main enable
    if (get_option('lazy_load_background') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      return;
    }

    // If the user has enabled auto-setting Used CSS
    // If the CSS file is missing
    // If the page is new or edited
    // And if we should be getting Used CSS on this page
    if (Utility::getting_used_css_data()) {
      return;
    }

    $exclude_home = get_option('lazy_background_exclude_home') === '1';
    $pages_exclude = get_option('lazy_background_exclude_pages_cbx') === '1';
    $pages_exclude_list = get_option('lazy_background_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_pages_textarea'))) : [];

    if ($exclude_home && Utility::is_homepage()) {
      return;
    }

    // Run on all except these excluded pages
    elseif ($pages_exclude) {
      // Don't run on pages from the list
      if (!empty($pages_exclude_list)) {
        $current_Url = Utility::get_current_page_url();
        foreach ($pages_exclude_list as $string) {
          $normalizedString = Utility::normalizeUrlString($string);
          if (stripos($current_Url, $normalizedString) !== false) {
            return;
          }
        }
      }
    }

    $lazy_background_checkbox = get_option('lazy_background_checkbox') === '1';
    if ($lazy_background_checkbox) {
      $lazy_background_exclude_selector = get_option('lazy_background_exclude_selector') === '1';
      $lazy_background_exclude_selector_textarea = get_option('lazy_background_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_selector_textarea'))) : [];
      $lazy_background_exclude_parent_selector = get_option('lazy_background_exclude_parent_selector') === '1';
      $exclude_bg_preload_parent_selector_list = get_option('lazy_background_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_parent_selector_textarea'))) : [];
      $lazy_bg_fadein = get_option('lazy_bg_fadein') === '1';
    }

    $styles = '';

    if ($lazy_background_checkbox) {

      // Delay JS script wil check for this element, before loading the JS file lazy-load-backgrouds.js
      echo '<style id="superspeed-lazy-load-backgrounds" type="text/css"></style>';

      if ($lazy_bg_fadein) {
        echo '<style id="superspeed-lazy-bg-fadein" type="text/css"></style>';
      }

      // Always initialize these to avoid undefined variable errors
      $selectors_string = '';
      $parent_selectors_string = '';

      // Exclude background images by class
      if ($lazy_background_exclude_selector && !empty($lazy_background_exclude_selector_textarea)) {

        $formatted_selectors = [];
        foreach ($lazy_background_exclude_selector_textarea as $selector) {
          $formatted_selectors[] = '.' . $selector;
        }

        if (!empty($formatted_selectors)) {
          $selectors_string = ', ' . implode(', ', $formatted_selectors);
        }
      }

      // Exclude background images by parent class
      if ($lazy_background_exclude_parent_selector && !empty($exclude_bg_preload_parent_selector_list)) {

        $formatted_parent_selectors = [];
        foreach ($exclude_bg_preload_parent_selector_list as $selector) {
          $formatted_parent_selectors[] = '.' . $selector . ' *';
        }

        if (!empty($formatted_parent_selectors)) {
          $parent_selectors_string = ', ' . implode(', ', $formatted_parent_selectors);
        }
      }

      /* Grey pseudo element for <img>: .ss-img-placeholder */
      /* Preloaded CSS Background: [bg_preloaded_dekstop], [bg_preloaded_tablet], [bg_preloaded_mobile] */
      /* LCP elements: [is_lcp_desktop], [is_lcp_tablet], [is_lcp_mobile] */

      // If we don't have image data, delay all background images, and exclude by class or parent class
      if ($lazy_background_checkbox) {
        if ($this->backgroundImageDataAvailable) {
          $styles .= '
            *:not(.ss-allow-bg ' . $selectors_string . $parent_selectors_string . ') {
              background-image: unset !important;
            }
          ';
        } else {
          $styles .= '
          *:not(.ss-allow-bg) {
            background-image: unset !important;
          }
        ';
        }

        // Fade in the background image
        if ($lazy_bg_fadein) {
          $fadeInDuration = floatval(get_option('lazy_bg_fadein_duration', 1));

          /*** FADE IN ***/
          // We can only fade in background images of we have image data
          // JS checks which *ss-allow-bg* element has a background image and adds the class 'ss-bg-zero-opacity'
          // We need to set the opcity to zero, so that it doesn't flicker
          // Zero opacity is removed, once the image is loaded

          if ($this->backgroundImageDataAvailable) {

            // Separating the screen widths to exclude from fade-in only the LCP for that screen width, instead of all preloaded elements
            $styles .= '
              @media screen and (min-width: 993px) {
                [has_background_image="no-children"]:not(.bg-lazy-loaded, [bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_desktop], [bg_preloaded_dekstop] ' . $selectors_string . $parent_selectors_string . ') {
                  opacity: 0;
                }
                [has_background_image="no-children"].ss-allow-bg:not([bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_desktop], [bg_preloaded_dekstop] ' . $selectors_string . $parent_selectors_string . ') {
                  animation-name: ss-fade-in-bg;
                  animation-duration: ' . $fadeInDuration . 's;
                  animation-timing-function: ease-out;
                  animation-fill-mode: forwards;
                }
              }

              @media screen and (min-width: 768px) and (max-width: 992px) {
                [has_background_image="no-children"]:not(.bg-lazy-loaded, [bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_tablet], [bg_preloaded_tablet]  ' . $selectors_string . $parent_selectors_string . ') {
                  opacity: 0;
                }
                [has_background_image="no-children"].ss-allow-bg:not([bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_tablet], [bg_preloaded_tablet]  ' . $selectors_string . $parent_selectors_string . ') {
                  animation-name: ss-fade-in-bg;
                  animation-duration: ' . $fadeInDuration . 's;
                  animation-timing-function: ease-out;
                  animation-fill-mode: forwards;
                }
              }

              @media screen and (max-width: 767px) {
                [has_background_image="no-children"]:not(.bg-lazy-loaded, [bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_mobile], [bg_preloaded_mobile] ' . $selectors_string . $parent_selectors_string . ') {
                  opacity: 0;
                }
                [has_background_image="no-children"].ss-allow-bg:not([bg_placeholder="gray"], [bg_placeholder="low-rez"], .ss-img-placeholder, [is_lcp_mobile], [bg_preloaded_mobile] ' . $selectors_string . $parent_selectors_string . ') {
                  animation-name: ss-fade-in-bg;
                  animation-duration: ' . $fadeInDuration . 's;
                  animation-timing-function: ease-out;
                  animation-fill-mode: forwards;
                }
              }

                /* The animation has only *from*, so that the element can have a custom opacity */
                @keyframes ss-fade-in-bg {
                  from {
                    opacity: 0;
                  }
                }
                ';
          }
        }

        // Visually mark lazy loaded background images
        if (get_option('preloaded_bg_lazyloaded_mark') === '1') {
          if ($this->backgroundImageDataAvailable) {
            // Excluding video placeholders
            // Preloaded Background: [bg_preloaded_dekstop], [bg_preloaded_tablet], [bg_preloaded_mobile]
            // Setting smallest thumbnail, because we are not lazy loading [ss-original-src-bg]
            // If we have data info, exclude markng the LCP for the screensize, if it has been preloaded
            $styles .= '
              @media screen and (min-width: 993px) {
                .can-edit [bg_lazy_loaded_desktop]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_dekstop] ' . $selectors_string . $parent_selectors_string . '),
                .can-edit [bg_lazy_loaded]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_dekstop] ' . $selectors_string . $parent_selectors_string . ') {
                  outline: dashed 0.25rem blue !important;
                  border: solid 0.25rem white !important;
                  outline-offset: -0.25rem !important;
                }
              }

              @media screen and (min-width: 768px) and (max-width: 992px) {
                .can-edit [bg_lazy_loaded_tablet]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_tablet] ' . $selectors_string . $parent_selectors_string . '),
                .can-edit [bg_lazy_loaded]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_tablet] ' . $selectors_string . $parent_selectors_string . ') {
                  outline: dashed 0.25rem blue !important;
                  border: solid 0.25rem white !important;
                  outline-offset: -0.25rem !important;
                }
              }
              
              @media screen and (max-width: 767px) {
                .can-edit [bg_lazy_loaded_mobile]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_mobile] ' . $selectors_string . $parent_selectors_string . '),
                .can-edit [bg_lazy_loaded]:not(.ss-video-placeholder, [ss-original-src-bg], [bg_preloaded_mobile] ' . $selectors_string . $parent_selectors_string . ') {
                  outline: dashed 0.25rem blue !important;
                  border: solid 0.25rem white !important;
                  outline-offset: -0.25rem !important;
                }
              }
          ';
          } else {
            $styles .= '
            .can-edit .ss-allow-bg.bg-lazy-loaded:not(.ss-video-placeholder, [ss-original-src-bg]) {
              outline: dashed 0.25rem blue !important;
              border: solid 0.25rem white !important;
              outline-offset: -0.25rem !important;
            }
          ';
          }
        }
      }
    }

    // Visually mark preloaded background images
    if (get_option('preloaded_bg_visually_mark') === '1') {
      if ($this->backgroundImageDataAvailable) {
        // If we have image data, for the specific screen size, only mark the LCP that has been preloaded
        $styles .= '
        @media screen and (min-width: 993px) {
          .can-edit [bg_preloaded_dekstop] {
            outline: dashed 0.25rem red !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }
        }
        @media screen and (min-width: 768px) and (max-width: 992px) {
          .can-edit [bg_preloaded_tablet] {
            outline: dashed 0.25rem red !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }      
        }
        @media screen and (max-width: 767px) {
          .can-edit [bg_preloaded_mobile] {
            outline: dashed 0.25rem red !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }
        }
      ';
      }
    }

    // Add style elements to the top of the HTML
    global $ss_global_variable_for_inlining_css;
    $ss_global_variable_for_inlining_css .= '<style id="superspeed-mark-preloaded-bg-img" type="delayed-style">' . $styles . '</style>
    ';
  }


  //// SET ATTRIBUTES FROM GATHERED IMAGE DATA

  // The function adds a custom, numbered ID to most HTML elements
  // If we have info about images in the database, along with the IDs, this into will be added as attributes
  public function buffer_start_for_setting_attributes_from_gathered_image_data($post_id)
  {

    // If the user has NOT enabled auto-setting LCP images
    if (!Utility::get_images_info_enabled()) {
      return;
    }

    // If the data exists
    if (Utility::image_data_is_empty()) {
      return;
    }

    // And if we should NOT be getting image data on this page
    if (!Utility::get_images_info_on_this_page()) {
      return;
    }

    // If the page is new or edited
    if (get_post_meta($post_id, 'post_edited_imaged_data', true)) {
      return;
    }

    ob_start(array($this, "setting_attributes_from_gathered_image_data"));
    $this->buffering_started_setting_attributes_from_gathered_image_data = true;
  }


  public function setting_attributes_from_gathered_image_data($html)
  {
    $dom = Utility::processHTML($html);

    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    $current_url = Utility::get_current_page_url();
    if ($current_url === null) {
      return $html;
    }

    $serialized_data = get_option('images_data_for_page_' . hash('sha256', $current_url));
    if (!$serialized_data || !($data = maybe_unserialize($serialized_data)) || !is_array($data) || !isset($data['elements'])) {
      error_log("Error retrieving or deserializing data for URL: " . $current_url);
      return $html;
    }
    $elementDataMap = [];
    foreach ($data['elements'] as $elementData) {
      $selector = $elementData['selector'];
      $visibleOn = $elementData['VisibleOn'];
      if (!isset($elementDataMap[$selector])) {
        $elementDataMap[$selector] = [];
      }
      if (!isset($elementDataMap[$selector][$visibleOn])) {
        $elementDataMap[$selector][$visibleOn] = [];
      }
      $elementDataMap[$selector][$visibleOn][] = $elementData;
    }
    $xpath = new \DOMXPath($dom);
    $targetElements = ['img', 'div', 'section', 'article', 'header', 'footer', 'aside', 'h1', 'h2', 'p', 'span'];
    foreach ($elementDataMap as $selector => $devices) {
      $elements = $xpath->query($this->cssToXpath($selector));
      foreach ($elements as $index => $element) {
        if (!in_array($element->nodeName, $targetElements)) {
          continue;
        }

        $element->setAttribute('data-selector', $selector);

        $overlayValueDesktop = null;
        $overlayValueTablet = null;
        $overlayValueMobile = null;
        $bgUrlDesktop = null;
        $bgUrlTablet = null;
        $bgUrlMobile = null;

        foreach ($devices as $device => $elementDataList) {
          if (isset($elementDataList[$index])) {
            $elementData = $elementDataList[$index];

            if ($elementData['IsLcp']) {
              $attributeName = 'is_lcp_' . strtolower($device);
              $element->setAttribute($attributeName, 'true');
            }

            $visibilityAttribute = 'visible_on_' . strtolower($device);
            $element->setAttribute($visibilityAttribute, 'true');

            if (!$elementData['isImage'] && !empty($elementData['url'])) {
              if ($device === 'Desktop') {
                $bgUrlDesktop = $elementData['url'];
              } elseif ($device === 'Tablet') {
                $bgUrlTablet = $elementData['url'];
              } elseif ($device === 'Mobile') {
                $bgUrlMobile = $elementData['url'];
              }
            }

            // Store the overlay value in a custom attribute
            if (!empty($elementData['overlay'])) {
              if ($device === 'Desktop') {
                $overlayValueDesktop = $elementData['overlay'];
              } elseif ($device === 'Tablet') {
                $overlayValueTablet = $elementData['overlay'];
              } elseif ($device === 'Mobile') {
                $overlayValueMobile = $elementData['overlay'];
              }
            }
          }
        }

        // Replace "http://" with "https://" in the background URLs
        if ($bgUrlDesktop) {
          $bgUrlDesktop = str_replace("http://", "https://", $bgUrlDesktop);
        }
        if ($bgUrlTablet) {
          $bgUrlTablet = str_replace("http://", "https://", $bgUrlTablet);
        }
        if ($bgUrlMobile) {
          $bgUrlMobile = str_replace("http://", "https://", $bgUrlMobile);
        }

        if ($bgUrlDesktop !== null && $bgUrlDesktop === $bgUrlTablet && $bgUrlDesktop === $bgUrlMobile) {
          $element->setAttribute('bg_url', $bgUrlDesktop);
          $element->setAttribute('has_background_image', 'true');
        } else {
          if ($bgUrlDesktop !== null) {
            $element->setAttribute('bg_url_desktop', $bgUrlDesktop);
            $element->setAttribute('has_background_image', 'true');
          }
          if ($bgUrlTablet !== null) {
            $element->setAttribute('bg_url_tablet', $bgUrlTablet);
            $element->setAttribute('has_background_image', 'true');
          }
          if ($bgUrlMobile !== null) {
            $element->setAttribute('bg_url_mobile', $bgUrlMobile);
            $element->setAttribute('has_background_image', 'true');
          }
        }

        // Replace "http://" with "https://" in the overlay values
        if ($overlayValueDesktop) {
          $overlayValueDesktop = str_replace("http://", "https://", $overlayValueDesktop);
        }
        if ($overlayValueTablet) {
          $overlayValueTablet = str_replace("http://", "https://", $overlayValueTablet);
        }
        if ($overlayValueMobile) {
          $overlayValueMobile = str_replace("http://", "https://", $overlayValueMobile);
        }

        if ($overlayValueDesktop !== null && $overlayValueDesktop === $overlayValueTablet && $overlayValueDesktop === $overlayValueMobile) {
          $element->setAttribute('overlay', $overlayValueDesktop);
        } else {
          if ($overlayValueDesktop !== null) {
            $element->setAttribute('overlay_desktop', $overlayValueDesktop);
          }
          if ($overlayValueTablet !== null) {
            $element->setAttribute('overlay_tablet', $overlayValueTablet);
          }
          if ($overlayValueMobile !== null) {
            $element->setAttribute('overlay_mobile', $overlayValueMobile);
          }
        }
      }
    }
    return $dom->saveHTML();
  }


  // Already performed the check, and stored it in the variable '$this->buffering_started_autosize_backgound_images'
  public function buffer_end_for_setting_attributes_from_gathered_image_data()
  {
    if ($this->buffering_started_setting_attributes_from_gathered_image_data) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_setting_attributes_from_gathered_image_data = false;
    }
  }

  public function cssToXpath($selector)
  {
    // Decode the selector to handle encoded characters
    $decodedSelector = urldecode($selector);
    // error_log("Decoded Selector: " . $decodedSelector);

    $xpath = '//*';

    // Match an ID selector.
    if (preg_match('/#([\w\-]+)/', $decodedSelector, $matches)) {
      $xpath .= "[@id='{$matches[1]}']";
      // error_log("Matched ID Selector: " . $matches[1]);
    }

    // Match class selectors up to the point where attribute selectors start
    $classSelectorPart = preg_split('/\[[^\]]+\]/', $decodedSelector, 2);
    if (isset($classSelectorPart[0])) {
      $classMatches = [];
      if (preg_match_all('/\.([\w\-]+)/', $classSelectorPart[0], $classMatches)) {
        foreach ($classMatches[1] as $class) {
          $xpath .= "[contains(concat(' ', normalize-space(@class), ' '), ' {$class} ')]";
          // error_log("Matched Class Selector: " . $class);
        }
      }
    }

    // Match attribute selectors with various operators
    $attrMatches = [];
    if (preg_match_all('/\[(\w+)([\*\^\$\|~]?=)\'([^\'\]]*)\'\]/', $decodedSelector, $attrMatches)) {
      foreach ($attrMatches[1] as $index => $attr) {
        $operator = $attrMatches[2][$index];
        $value = $attrMatches[3][$index];

        // error_log("Matched Attribute: " . $attr . ", Operator: " . $operator . ", Value: " . $value);

        switch ($operator) {
          case '*=':
            $xpath .= "[contains(@{$attr}, '{$value}')]";
            break;
          case '^=':
            $xpath .= "[starts-with(@{$attr}, '{$value}')]";
            break;
          case '$=':
            $xpath .= "[substring(@{$attr}, string-length(@{$attr}) - string-length('{$value}') + 1) = '{$value}']";
            break;
          case '|=':
            $xpath .= "[@{$attr}='{$value}' or starts-with(@{$attr}, '{$value}-')]";
            break;
          case '~=':
            $xpath .= "[contains(concat(' ', @{$attr}, ' '), ' {$value} ')]";
            break;
          default:
            $xpath .= "[@{$attr}='{$value}']";
            break;
        }
      }
    }

    // Match attribute selectors with no operator (i.e., [attr])
    if (preg_match_all('/\[(\w+)\]/', $decodedSelector, $simpleAttrMatches)) {
      foreach ($simpleAttrMatches[1] as $attr) {
        $xpath .= "[@{$attr}]";
        // error_log("Matched Attribute with no operator: " . $attr);
      }
    }

    // error_log("Final XPath: " . $xpath);
    return $xpath;
  }


  /******************************** */


  /*** CSS FOR MARKING THE LCP ***/

  public function css_for_marking_the_lcp()
  {

    // Main enable
    if (get_option('lcp_enable') !== '1') {
      return;
    }

    // Display LCP info
    if (get_option('display_lcp_info') === '1') {

      echo '<style id="superspeed-while-getting-lcp-info" type="text/css">
      div#wpadminbar {
        display: none !important;
      }
      html {
          margin-top: unset !important;

      }
      </style>';

      echo '<style id="superspeed-display-lcp-info" type="text/css">
      #lcp-info-container {
          position: absolute;
          font-size: 13px;
          background-color: rgba(0, 0, 0, 0.7);
          color: white;
          padding: 0.5em 1em;
          border-radius: 5px;
          z-index: 1000;
          margin-bottom: 5px;
          margin-right: 0.125em;
          transition: font-size ease 0.35s;
          font-weight: 400;
          font-family: system-ui;
      }

      #lcp-info-container:hover {
        font-size: 20px;
        background-color: black;
        z-index: 10000;
      }

      .lcp-info-span::selection {
        color: black;
        background: yellow;
      }

      button.lcp-close-button {
        all: unset;
        float: right;
        rotate: 45deg;
        padding: 0.5em;
        line-height: 0.5em;
        cursor: pointer;
        font-size: 2em;
        translate: 0.66em -0.33em;
    }

    .lcp-info-classes {
      display: grid;
    }

  .lcp-info-span {
    margin-left: 1em;
    }
      </style>';
    }

    // If we are not auto-preloading the LCP image, exit
    // if (get_option('auto_preload_lcp') !== '1') {
    //   return;
    // }

    if (get_option('visually_mark_lcp') !== '1') {
      return;
    }

    // Detected LCP can only be marked if Image Data Collection is tured on
    if (!$this->imageDataAvailable) {
      return;
    }

    $exclude_home_lcp = get_option('lcp_prelod_exclude_home_cbx') === '1';
    $pages_exclude_lcp = get_option('lcp_prelod_exclude_pages_cbx') === '1';
    if ($pages_exclude_lcp) {
      $pages_exclude_list_lcp = get_option('lcp_prelod_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_pages_textarea'))) : [];
    }

    // Set flag for excluding preloading on these pages
    $preload_lcp_on_this_page = true;

    $current_Url_lcp = '';
    $is_home_lcp = '';

    if ($exclude_home_lcp) {
      $is_home_lcp = Utility::is_homepage();
      if ($is_home_lcp) {
        $preload_lcp_on_this_page = false;
      }
    }

    // Run on all except these excluded pages
    elseif ($pages_exclude_lcp) {
      // Don't run on pages from the list
      if (!empty($pages_exclude_list_lcp)) {
        $current_Url_lcp = Utility::get_current_page_url();
        foreach ($pages_exclude_list_lcp as $string) {
          $normalizedString = Utility::normalizeUrlString($string);
          if (stripos($current_Url_lcp, $normalizedString) !== false) {
            $preload_lcp_on_this_page = false;
          }
        }
      }
    }

    // Visually mark LCP elements
    if (!$preload_lcp_on_this_page) {
      return;
    }

  ?>
    <style id="superspeed-lcp-marking" type="delayed-style">
      @media screen and (min-width: 993px) {
          .can-edit [is_lcp_desktop] {
            outline: dashed 0.25rem violet !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }

          .can-edit [is_lcp_desktop][bg_preloaded_dekstop],
          .can-edit [is_lcp_desktop][img_preloaded_dekstop] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem red !important;
            outline-offset: -0.33rem !important;
          }

          .can-edit [is_lcp_desktop][lazy_loaded_img],
          .can-edit [is_lcp_desktop][bg_lazy_loaded_desktop],
          .can-edit [is_lcp_desktop][bg_lazy_loaded] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem blue !important;
            outline-offset: -0.33rem !important;
          }
        }

        @media screen and (min-width: 768px) and (max-width: 992px) {
          .can-edit [is_lcp_tablet] {
            outline: dashed 0.25rem violet !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }

          .can-edit [is_lcp_tablet][bg_preloaded_tablet],
          .can-edit [is_lcp_tablet][img_preloaded_tablet] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem red !important;
            outline-offset: -0.33rem !important;
          }

          .can-edit [is_lcp_tablet][lazy_loaded_img],
          .can-edit [is_lcp_tablet][bg_lazy_loaded_tablet],
          .can-edit [is_lcp_desktop][bg_lazy_loaded] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem blue !important;
            outline-offset: -0.33rem !important;
          }
        }

        @media screen and (max-width: 767px) {
          .can-edit [is_lcp_mobile] {
            outline: dashed 0.25rem violet !important;
            border: solid 0.25rem white !important;
            outline-offset: -0.25rem !important;
          }

          .can-edit [is_lcp_mobile][bg_preloaded_mobile],
          .can-edit [is_lcp_mobile][img_preloaded_mobile] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem red !important;
            outline-offset: -0.33rem !important;
          }

          .can-edit [is_lcp_mobile][lazy_loaded_img],
          .can-edit [is_lcp_mobile][bg_lazy_loaded_mobile],
          .can-edit [is_lcp_desktop][bg_lazy_loaded] {
            outline: dashed 0.33rem violet !important;
            border: solid 0.33rem blue !important;
            outline-offset: -0.33rem !important;
          }
        }
      </style>
  <?php
  }




  /*** BACKGROUND IMAGES: AUTO-SIZE, PRELOAD ***/

  // Helper function called from the __construct()
  // Determins if we have image info, and are using it on this page
  // Gives values to global variables $this->imageDataAvailable and $this->backgroundImageDataAvailable
  private function determineImageDataAvailability()
  {

    if (!Utility::get_images_info_enabled()) {
      return;
    }

    if (!Utility::get_images_info_on_this_page()) {
      return;
    }

    // Get current URL and corresponding option name
    $current_url = Utility::get_current_page_url();
    $option_name = 'images_data_for_page_' . hash('sha256', $current_url);
    $data = get_option($option_name);

    // Check if there's no data for the page
    if (!$data || count($data) === 0) {
      return; // Exit early if no data
    }

    // Initialize flags to track found images
    $foundImage = $foundBackgroundImage = false;

    // Check for images and background images within the same loop
    foreach ($data['elements'] ?? [] as $element) {
      if (!$foundImage) {
        if (isset($element['isImage']) && $element['isImage'] == 1) {
          $this->imageDataAvailable = true;
          $foundImage = true; // Image found
        }
      }
      if (!$foundBackgroundImage) {
        if (isset($element['isImage'], $element['url']) && $element['isImage'] == 0 && $element['url'] !== 'null' && !empty($element['url'])) {
          $this->backgroundImageDataAvailable = true;
          $foundBackgroundImage = true; // Background image found
        }
      }
      if ($foundImage && $foundBackgroundImage) {
        break; // Exit loop if both types of images are found
      }
    }
  }


  // Helper function or preloading background images
  function preloadBackgroundImage($node, &$preloadLinks)
  {
    $selector = $node->getAttribute('data-selector');

    $urlDesktop = $node->getAttribute('bg_url_desktop');
    $urlTablet = $node->getAttribute('bg_url_tablet');
    $urlMobile = $node->getAttribute('bg_url_mobile');
    $urlAll = $node->getAttribute('bg_url');

    $overlayAll = $node->getAttribute('overlay');
    $overlayDesktop = $node->getAttribute('overlay_desktop');
    $overlayTablet = $node->getAttribute('overlay_tablet');
    $overlayMobile = $node->getAttribute('overlay_mobile');

    // Reset for each element
    $smallPreloadedBgImage = null;

    // Get the smallest thumbnail and preload it
    // With JS, we add the original image size inline, and remove this CSS by adding the class 'original-bg-loaded'
    // Check if all URLs are the same, if so, create a single srcset
    if ($urlAll) {
      $urlAllSmall = self::get_smallest_thumbnail($urlAll);
      $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlAllSmall . "'>";

      $declarationAll = "{background-image: url(" . $urlAllSmall . ")";

      // If the background image is mixed in with an overlay color or gradient
      if ($overlayAll) {
        $overlayAll = str_replace($urlAll, $urlAllSmall, $overlayAll);
        $declarationAll = "{background-image: " . $overlayAll;
      }

      $smallPreloadedBgImage .= "\n" . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationAll . " !important;}";
      $node->setAttribute('bg_preloaded_dekstop', '');
      $node->setAttribute('bg_preloaded_tablet', '');
      $node->setAttribute('bg_preloaded_mobile', '');
    } else {
      if ($urlDesktop) {
        $urlDesktopSmall = self::get_smallest_thumbnail($urlDesktop);
        $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlDesktopSmall . "' media='(min-width: 800.1px)'>";

        $declarationDesktop = "{background-image: url(" . $urlDesktopSmall . ")";

        // If the background image is mixed in with an overlay color or gradient
        if ($overlayDesktop) {
          $overlayDesktop = str_replace($urlDesktop, $urlDesktopSmall, $overlayDesktop);
          $declarationDesktop = "{background-image: " . $overlayDesktop;
        }

        $smallPreloadedBgImage .= "\n@media (min-width: 1025px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationDesktop . " !important;}}";
        $node->setAttribute('bg_preloaded_dekstop', '');
      }

      if ($urlTablet) {
        $urlTabletSmall = self::get_smallest_thumbnail($urlTablet);
        $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlTabletSmall . "' media='(min-width: 400.1px) and (max-width: 800px)'>";

        $declarationTablet = "{background-image: url(" . $urlTabletSmall . ")";

        // If the background image is mixed in with an overlay color or gradient
        if ($overlayTablet) {
          $overlayTablet = str_replace($urlTablet, $urlTabletSmall, $overlayTablet);
          $declarationTablet = "{background-image: " . $overlayTablet;
        }

        $smallPreloadedBgImage .= "\n@media (min-width: 601px) and (max-width: 1024px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationTablet . " !important;}}";
        $node->setAttribute('bg_preloaded_tablet', '');
      }

      if ($urlMobile) {
        $urlMobileSmall = self::get_smallest_thumbnail($urlMobile);
        $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlMobileSmall . "' media='(max-width: 400px)'>";

        $declarationMobile = "{background-image: url(" . $urlDesktopSmall . ")";

        // If the background image is mixed in with an overlay color or gradient
        if ($overlayDesktop) {
          $overlayMobile = str_replace($urlMobile, $urlMobileSmall, $overlayMobile);
          $declarationMobile = "{background-image: " . $overlayMobile;
        }

        $smallPreloadedBgImage .= "\n@media (max-width: 600px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationMobile . " !important;}}";
        $node->setAttribute('bg_preloaded_mobile', '');
      }
    }
    error_log('$smallPreloadedBgImage: ' . $smallPreloadedBgImage);

    return $smallPreloadedBgImage;
  }


  public function buffer_start_for_background_images()
  {
    // If we do not have background data or if getting data is disabled
    if (!$this->backgroundImageDataAvailable) {
      return;
    }

    // If main Enable checkbox is unchecked
    if (get_option('lazy_load_background') !== '1' && get_option('auto_preload_lcp') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      return;
    }

    // If the user has enabled auto-setting Used CSS
    // If the CSS file is missing
    // If the page is new or edited
    // And if we should be getting Used CSS on this page
    if (Utility::getting_used_css_data()) {
      return;
    }

    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    ob_start(array($this, "process_elements_with_background_images"));
    $this->buffering_started_process_bg_img = true;
  }

  public function process_elements_with_background_images($html)
  {

    // Get the page URL
    $current_url = Utility::get_current_page_url();

    if ($current_url === null) {
      return $html;
    }

    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    // Will be used to find the elements with a background image in the DOM
    $xpath = new \DOMXPath($dom);

    $nodes = $xpath->query('//*[@has_background_image]');

    // Main Enable
    $bg_images_main_enable = get_option('lazy_load_background') === '1';

    $auto_preload_lcp = get_option('auto_preload_lcp') === '1';
    if ($auto_preload_lcp) {
      $exclude_home_lcp = get_option('lcp_prelod_exclude_home_cbx') === '1';
      $pages_exclude_lcp = get_option('lcp_prelod_exclude_pages_cbx') === '1';
      if ($pages_exclude_lcp) {
        $pages_exclude_list_lcp = get_option('lcp_prelod_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_pages_textarea'))) : [];
      }
      $lcp_prelod_exclude_selector = get_option('lcp_prelod_exclude_selector') === '1';
      if ($lcp_prelod_exclude_selector) {
        $lcp_prelod_exclude_selector_textarea = get_option('lcp_prelod_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_selector_textarea'))) : [];
      }
      $lcp_prelod_exclude_parent_selector = get_option('lcp_prelod_exclude_parent_selector') === '1';
      if ($lcp_prelod_exclude_parent_selector) {
        $lcp_prelod_exclude_parent_selector_textarea = get_option('lcp_prelod_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lcp_prelod_exclude_parent_selector_textarea'))) : [];
      }
    }


    $background_auto_set_thumbnails = get_option('background_auto_set_thumbnails') === '1';
    if ($background_auto_set_thumbnails) {
      $backgound_auto_set_exclude_selector = get_option('backgound_auto_set_exclude_selector') === '1';
      if ($backgound_auto_set_exclude_selector) {
        $backgound_auto_set_exclude_selector_textarea = get_option('backgound_auto_set_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('backgound_auto_set_exclude_selector_textarea'))) : [];
      }
      $backgound_auto_set_exclude_parent_selector = get_option('backgound_auto_set_exclude_parent_selector') === '1';
      if ($backgound_auto_set_exclude_parent_selector) {
        $backgound_auto_set_exclude_parent_selector_textarea = get_option('backgound_auto_set_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('backgound_auto_set_exclude_parent_selector_textarea'))) : [];
      }
    }

    $preload_selector_list = [];
    $preload_parent_selector_list = [];

    $preload_background_images = get_option('preload_background_images') === '1';
    if ($preload_background_images) {
      $preload_selector = get_option('preload_background_selector') === '1';
      if ($preload_selector) {
        $preload_selector_list = get_option('preload_background_selector_textarea') ? array_map('trim', explode("\n", get_option('preload_background_selector_textarea'))) : [];
      }
      $preload_parent_selector = get_option('preload_background_parent_selector') === '1';
      if ($preload_parent_selector) {
        $preload_parent_selector_list = get_option('preload_background_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('preload_background_parent_selector_textarea'))) : [];
      }
    }

    $lazy_background_checkbox = get_option('lazy_background_checkbox') === '1';
    if ($lazy_background_checkbox) {
      $Radio_back_img_placeholder = get_option('back_img_placeholder_radio');
      $back_img_placeholder_none = $Radio_back_img_placeholder === 'back_img_placeholder_none';
      $back_img_placeholder_gray = $Radio_back_img_placeholder === 'back_img_placeholder_gray';
      $back_img_placeholder_thumbnail = $Radio_back_img_placeholder === 'back_img_placeholder_thumbnail';

      $exclude_home = get_option('lazy_background_exclude_home') === '1';
      $pages_exclude = get_option('lazy_background_exclude_pages_cbx') === '1';
      if ($pages_exclude) {
        $pages_exclude_list = get_option('lazy_background_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_pages_textarea'))) : [];
      }
      $lazy_background_exclude_selector = get_option('lazy_background_exclude_selector') === '1';
      if ($lazy_background_exclude_selector) {
        $lazy_background_exclude_selector_textarea = get_option('lazy_background_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_selector_textarea'))) : [];
      }
      $lazy_background_exclude_parent_selector = get_option('lazy_background_exclude_parent_selector') === '1';
      if ($lazy_background_exclude_parent_selector) {
        $exclude_bg_preload_parent_selector_list = get_option('lazy_background_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_background_exclude_parent_selector_textarea'))) : [];
      }
    }


    //// PRELOAD BACKGROUND IMAGES ////

    $preloadLinks = null;
    $backgroundStyles = null;

    if ($preload_background_images && $bg_images_main_enable) {

      // Initialize an array to keep track of processed classes for preloading
      $processedClasses = [];
      $processedParentClasses = [];

      foreach ($nodes as $node) {

        $classes = explode(' ', $node->getAttribute('class')); // Get all classes from the element

        // PRELOAD BY CLASS

        // We only preload the first instance of an image for each class
        if ($preload_selector && array_intersect($preload_selector_list, $classes)) {
          // If the classes has already been processed, skip this node
          if (array_intersect($classes, $processedClasses)) {
            continue;
          }

          $backgroundStyles .= $this->preloadBackgroundImage($node, $preloadLinks);

          // We preload only the first element with the specified cass. So, we need to keep track of processed classes, and remove them from the loop
          // Add processed classes to the array
          $processedClasses = array_merge($processedClasses, $classes);
          // Remove classes that are not in the user's list
          $processedClasses = array_intersect($processedClasses, $preload_selector_list);
          //error_log('End of the loop $processedClasses :' . print_r($processedClasses, true));

          // Exit after the first image
          continue;
        }

        // Add processed classes to the tracking array
        $processedClasses = array_merge($processedClasses, $classes);
        $processedClasses = array_intersect($processedClasses, $preload_selector_list);

        // PRELOAD BY PARENT CLASS

        // We only preload the first instance of an image for each class, and the first instance of a parent for their classes
        if ($preload_parent_selector) {

          // Get the direct parent node
          $parentNode = $node->parentNode;

          // Check if the parent node has a class attribute
          if ($parentNode && $parentNode->hasAttributes() && $parentNode->attributes->getNamedItem('class')) {

            $parentClasses = explode(' ', $parentNode->attributes->getNamedItem('class')->nodeValue);

            // Find if any of the parent's classes are in the selector list
            $matchingClass = array_intersect($parentClasses, $preload_parent_selector_list);

            if (!empty($matchingClass)) {

              // Get the first matching class
              $parentClass = reset($matchingClass);

              // If this parent class has not been processed yet
              if (!in_array($parentClass, $processedParentClasses)) {

                $backgroundStyles .= $this->preloadBackgroundImage($node, $preloadLinks, $backgroundStyles);

                // Add the processed parent class to the tracking array
                $processedParentClasses[] = $parentClass;
              }
            }
          }
        }
      }
    }


    /// AUTO-PRELOAD BACKGROUND LCP ///

    // We already have an early exit for $this->backgroundImageDataAvailable
    // Since this is an option from LCP settings, it does not require bg images enable, but main LCP enable
    if ($auto_preload_lcp && (get_option('lcp_enable') === '1')) {

      // error_log('PRELOAD BACKGROUND LCP auto_preload_lcp: '.$auto_preload_lcp);
      // Set flag for excluding preloading on these pages
      $preload_lcp_on_this_page = true;

      // Exclude on Home page
      if ($exclude_home_lcp && Utility::is_homepage()) {
        $preload_lcp_on_this_page = false;
      }

      // Exclude by URL
      if ($pages_exclude_lcp && !empty($pages_exclude_list_lcp)) {
        foreach ($pages_exclude_list_lcp as $string) {
          if (stripos($current_url, $string) !== false) {
            $preload_lcp_on_this_page = false;
          }
        }
      }

      // Preloading LCP images on this page
      if ($preload_lcp_on_this_page) {

        // Iterate over each image element
        $counterLcp = 0;
        $preloadCountLcp = 3;

        foreach ($nodes as $node) {

          $is_lcp_desktop = $node->hasAttribute('is_lcp_desktop');
          $is_lcp_tablet = $node->hasAttribute('is_lcp_tablet');
          $is_lcp_mobile = $node->hasAttribute('is_lcp_mobile');

          if (!$is_lcp_desktop && !$is_lcp_tablet && !$is_lcp_mobile) {
            continue;
          }

          $classes = explode(' ', $node->getAttribute('class')); // Get all classes from the image.

          // Exclude images by selector
          if ($lcp_prelod_exclude_selector && array_intersect($lcp_prelod_exclude_selector_textarea, $classes)) {
            continue;
          }

          // Exclude images by parent selector
          if ($lcp_prelod_exclude_parent_selector && $this->hasParentWithClass($node, $lcp_prelod_exclude_parent_selector_textarea)) {
            continue;
          }

          // Count the images
          if ($counterLcp < $preloadCountLcp) {

            $selector = $node->getAttribute('data-selector');

            $urlDesktop = $node->getAttribute('bg_url_desktop');
            $urlTablet = $node->getAttribute('bg_url_tablet');
            $urlMobile = $node->getAttribute('bg_url_mobile');
            $urlAll = $node->getAttribute('bg_url');

            $overlayAll = $node->getAttribute('overlay');
            $overlayDesktop = $node->getAttribute('overlay_desktop');
            $overlayTablet = $node->getAttribute('overlay_tablet');
            $overlayMobile = $node->getAttribute('overlay_mobile');

            // Reset for each element
            $smallPreloadedBgImage = null;

            // If images are the same for each screen size
            // Exit if elements have the background image property, but do not have an image
            if ($urlAll) {

              // Get the smallest thumbnail and preload it
              // With JS, we add the original image size inline, and remove this CSS by adding the class 'original-bg-loaded'

              $urlAllSmall = self::get_smallest_thumbnail($urlAll);
              $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlAllSmall . "'>";

              $declarationAll = "{background-image: url(" . $urlAllSmall . ")";

              // If the background image is mixed in with an overlay color or gradient
              if ($overlayAll) {
                $overlayAll = str_replace($urlAll, $urlAllSmall, $overlayAll);
                $declarationAll = "{background-image: " . $overlayAll;
              }

              $smallPreloadedBgImage .= "\n" . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationAll . " !important;}";
              $node->setAttribute('bg_preloaded_dekstop', '');
              $node->setAttribute('bg_preloaded_tablet', '');
              $node->setAttribute('bg_preloaded_mobile', '');
            } else {

              // Exit if elements have the background image property, but do not have an image
              if ($is_lcp_desktop && $urlDesktop !== 'null') {
                $urlDesktopSmall = self::get_smallest_thumbnail($urlDesktop);
                $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlDesktopSmall . "' media='(min-width: 800.1px)'>";

                $declarationDesktop = "{background-image: url(" . $urlDesktopSmall . ")";

                // If the background image is mixed in with an overlay color or gradient
                if ($overlayDesktop) {
                  $overlayDesktop = str_replace($urlDesktop, $urlDesktopSmall, $overlayDesktop);
                  $declarationDesktop = "{background-image: " . $overlayDesktop;
                }

                $smallPreloadedBgImage .= "\n@media (min-width: 1025px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationDesktop . " !important;}}";
                $node->setAttribute('bg_preloaded_dekstop', '');
              }

              if ($is_lcp_tablet && $urlTablet !== 'null') {
                $urlTabletSmall = self::get_smallest_thumbnail($urlTablet);
                $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlTabletSmall . "' media='(min-width: 400.1px) and (max-width: 800px)'>";

                $declarationTablet = "{background-image: url(" . $urlTabletSmall . ")";

                // If the background image is mixed in with an overlay color or gradient
                if ($overlayTablet) {
                  $overlayTablet = str_replace($urlTablet, $urlTabletSmall, $overlayTablet);
                  $declarationTablet = "{background-image: " . $overlayTablet;
                }

                $smallPreloadedBgImage .= "\n@media (min-width: 601px) and (max-width: 1024px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationTablet . " !important;}}";
                $node->setAttribute('bg_preloaded_tablet', '');
              }

              if ($is_lcp_mobile && $urlMobile !== 'null') {
                $urlMobileSmall = self::get_smallest_thumbnail($urlMobile);
                $preloadLinks .= "\n<link rel='preload' as='image' fetchpriority='high' href='" . $urlMobileSmall . "' media='(max-width: 400px)'>";

                $declarationMobile = "{background-image: url(" . $urlDesktopSmall . ")";

                // If the background image is mixed in with an overlay color or gradient
                if ($overlayDesktop) {
                  $overlayMobile = str_replace($urlMobile, $urlMobileSmall, $overlayMobile);
                  $declarationMobile = "{background-image: " . $overlayMobile;
                }

                $smallPreloadedBgImage .= "\n@media (max-width: 600px) { " . $selector . ":not(.original-bg-loaded, .ss-allow-bg) " . $declarationMobile . " !important;}}";
                $node->setAttribute('bg_preloaded_mobile', '');
              }
            }


            $backgroundStyles .= $smallPreloadedBgImage;

            if ($is_lcp_desktop && $is_lcp_tablet &&  $is_lcp_mobile) {
              break;
            } else {
              $counterLcp++;
              continue;
            }
          }
        }
      }
    }



    //// LAZY LOAD BACKGROUND IMAGES ////

    // The image url is already in the HTML
    // We prevent images from loading is done with CSS
    // We load, or allow to load, the image with JS
    // Here, with PHP, we are only setting placeholders and attributes

    if ($lazy_background_checkbox && $bg_images_main_enable) {

      $runLazyLoadOnPage = true;

      if ($exclude_home && Utility::is_homepage()) {
        $runLazyLoadOnPage = false;
      }

      // Run on all except these excluded pages
      elseif ($pages_exclude) {
        // Don't run on pages from the list
        if (!empty($pages_exclude_list)) {
          $current_Url = Utility::get_current_page_url();
          foreach ($pages_exclude_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_Url, $normalizedString) !== false) {
              $runLazyLoadOnPage = false;
            }
          }
        }
      }


      if ($runLazyLoadOnPage) {

        foreach ($nodes as $node) {

          // Skip preloaded background images
          if (
            $node->hasAttribute('bg_preloaded_dekstop') ||
            $node->hasAttribute('bg_preloaded_tablet') ||
            $node->hasAttribute('bg_preloaded_mobile')
          ) {
            continue;
          }

          $classes = explode(' ', $node->getAttribute('class'));

          // Exclude by class
          if ($lazy_background_exclude_selector && array_intersect($lazy_background_exclude_selector_textarea, $classes)) {
            continue;
          }

          // Exclude by parent class
          if ($lazy_background_exclude_parent_selector && $this->hasParentWithClass($node, $exclude_bg_preload_parent_selector_list)) {
            continue;
          }

          $lazyBgPlaceholderStyles = '';
          $selector = $node->getAttribute('data-selector');
          $urlDesktop = $node->getAttribute('bg_url_desktop');
          $urlTablet = $node->getAttribute('bg_url_tablet');
          $urlMobile = $node->getAttribute('bg_url_mobile');
          $urlAll = $node->getAttribute('bg_url');


          if ($urlDesktop && $urlDesktop !== 'null') {
            $node->setAttribute('bg_lazy_loaded_desktop', '');
          }
          if ($urlTablet && $urlTablet !== 'null') {
            $node->setAttribute('bg_lazy_loaded_tablet', '');
          }
          if ($urlMobile && $urlMobile !== 'null') {
            $node->setAttribute('bg_lazy_loaded_mobile', '');
          }
          if ($urlAll && $urlAll !== 'null') {
            $node->setAttribute('bg_lazy_loaded', '');
          }

          // LAZY LOAD PLACEHOLDERS //

          // EMPTY PLACEHOLDER FOR BACKGROUND IMAGES
          if ($back_img_placeholder_none) {

            $node->setAttribute('bg_placeholder', 'none');

            if ($urlAll) {
              $lazyBgPlaceholderStyles .= "\n" . $selector . ":not(.ss-allow-bg, [is_lcp_desktop], [is_lcp_tablet], [is_lcp_mobile]) {background-image: unset !important;}";
            } else {
              // We are excluding the LCP for each screen size separately
              if ($urlDesktop) {
                $lazyBgPlaceholderStyles .= "\n@media (min-width: 1025px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_desktop]) {background-image: unset !important;}}";
              }
              if ($urlTablet) {
                $lazyBgPlaceholderStyles .= "\n@media (min-width: 601px) and (max-width: 1024px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_tablet]) {background-image: unset !important;}}";
              }
              if ($urlMobile) {
                $lazyBgPlaceholderStyles .= "\n@media (max-width: 600px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_mobile]) {background-image: unset !important;}}";
              }
            }
          }


          // GRAY PLACEHOLDER FOR BACKGROUND IMAGES
          if ($back_img_placeholder_gray) {

            $node->setAttribute('bg_placeholder', 'gray');

            $grayPlaceholder = 'data:image/svg+xml;utf8,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%3E%3Crect%20x%3D%220%22%20y%3D%220%22%20width%3D%2210000%22%20height%3D%2210000%22%20fill%3D%22gray%22%3E%3Canimate%20attributeName%3D%22fill%22%20values%3D%22lightgray%3B%20darkgray%3B%20lightgray%22%20dur%3D%222s%22%20repeatCount%3D%22indefinite%22%20%2F%3E%3C%2Frect%3E%3C%2Fsvg%3E';

            // We are excluding the LCP for each screen size separately
            if ($urlDesktop) {
              $lazyBgPlaceholderStyles .= "\n@media (min-width: 1025px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_desktop]) {background-image: url(" . $grayPlaceholder . ") !important;}}";
            }
            if ($urlTablet) {
              $lazyBgPlaceholderStyles .= "\n@media (min-width: 601px) and (max-width: 1024px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_tablet]) {background-image: url(" . $grayPlaceholder . ") !important;}}";
            }
            if ($urlMobile) {
              $lazyBgPlaceholderStyles .= "\n@media (max-width: 600px) { " . $selector . ":not(.ss-allow-bg, [is_lcp_mobile]) {background-image: url(" . $grayPlaceholder . ") !important;}}";
            }
          }

          // LOW-REZ PLACEHOLDER FOR BACKGROUND IMAGES
          if ($back_img_placeholder_thumbnail) {

            $node->setAttribute('bg_placeholder', 'low-rez');

            // We don't need to exclude the LCP. A low resolution thumbnail is a good setting

            if (($urlDesktop === $urlTablet) && ($urlDesktop === $urlMobile)) {
              $urlDesktop = self::get_smallest_thumbnail($urlDesktop);
              $lazyBgPlaceholderStyles .= "\n" . $selector . ":not(.ss-allow-bg) {background-image: url(" . $urlDesktop . ") !important;}";
            } else {
              if ($urlDesktop) {
                $urlDesktop = self::get_smallest_thumbnail($urlDesktop);
                $lazyBgPlaceholderStyles .= "\n@media (min-width: 1025px) { " . $selector . ":not(.ss-allow-bg) {background-image: url(" . $urlDesktop . ") !important;}}";
              }

              if ($urlTablet) {
                $urlTablet = self::get_smallest_thumbnail($urlTablet);
                $lazyBgPlaceholderStyles .= "\n@media (min-width: 601px) and (max-width: 1024px) { " . $selector . ":not(.ss-allow-bg) {background-image: url(" . $urlTablet . ") !important;}}";
              }

              if ($urlMobile) {
                $urlMobile = self::get_smallest_thumbnail($urlMobile);
                $lazyBgPlaceholderStyles .= "\n@media (max-width: 600px) { " . $selector . ":not(.ss-allow-bg) {background-image: url(" . $urlMobile . ") !important;}}";
              }
            }
          }
          $backgroundStyles .= $lazyBgPlaceholderStyles;
        }
      }
    }


    //// AUTO-SET THUMBNAILS FOR BACKGROUND IMAGES ////

    if ($background_auto_set_thumbnails && $bg_images_main_enable) {

      foreach ($nodes as $node) {

        $urlDesktop = $node->getAttribute('bg_url_desktop');
        $urlTablet = $node->getAttribute('bg_url_tablet');
        $urlMobile = $node->getAttribute('bg_url_mobile');
        $urlAll = $node->getAttribute('bg_url');

        // Check if all URLs are 'null'
        if ((!$urlDesktop || $urlDesktop === 'null') &&
          (!$urlTablet || $urlTablet === 'null') &&
          (!$urlMobile || $urlMobile === 'null') &&
          (!$urlAll || $urlAll === 'null')
        ) {
          continue;
        }

        $classes = explode(' ', $node->getAttribute('class')); // Get all classes from the element

        // Exclude by class
        if ($backgound_auto_set_exclude_selector && array_intersect($backgound_auto_set_exclude_selector_textarea, $classes)) {
          continue;
        }

        // Exclude by parent class
        if ($backgound_auto_set_exclude_parent_selector && $this->hasParentWithClass($node, $backgound_auto_set_exclude_parent_selector_textarea)) {
          continue;
        }

        // If there is only the 'bg_url' attribute, create a single srcset
        if ($urlAll) {
          $this->create_custom_srcset($node, ['all' => $urlAll]);
        } else {
          $this->create_custom_srcset($node, ['desktop' => $urlDesktop, 'tablet' => $urlTablet, 'mobile' => $urlMobile]);
        }
      }
    }

    // Add preloaded images to the global variable
    global $ss_global_variable_for_preloading_images;
    $ss_global_variable_for_preloading_images .= $preloadLinks;

    // Add style elements to the top of the HTML
    global $ss_global_variable_for_inlining_css;
    $ss_global_variable_for_inlining_css .= "\n<style id='superspeed-auto-set-bg-img-thumbs' type='text/css'>" . $backgroundStyles . "\n</style>";

    return $dom->saveHTML();
  }

  // Already performed the check, and stored it in the variable '$this->buffering_started_autosize_backgound_images'
  public function buffer_end_for_background_images()
  {
    if ($this->buffering_started_process_bg_img) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_process_bg_img = false;
    }
  }


  // Helper function for creating custom 'srcset' for background images
  public function create_custom_srcset($node, $urls)
  {
    // error_log('create_custom_srcset helper');
    $sizesToLabels = [
      'full' =>       '8192',
      'xxx_large' =>  '4096',
      'xx_large' =>   '2048',
      'x_large' =>    '1536',
      'large' =>      '1024',
      'xx_medium' =>  '768',
      'x_medium' =>   '512',
      'small' =>      '256',
      'x_small' =>    '0'
    ];

    // Initialize an array to hold srcset strings for different devices
    $srcsetStrings = [];

    foreach ($urls as $deviceType => $url) {

      // URL-decode the link
      $url = urldecode($url);
      // error_log('$url: ' . $url);

      if (empty($url)) continue; // Skip if URL is empty
      // error_log('$url: ' . $url);

      $bgImgUrl = (substr($url, 0, 1) === '/') ? home_url($url) : $url;
      // error_log('$bgImgUrl: ' . $bgImgUrl);

      $imageId = attachment_url_to_postid($bgImgUrl);
      if (!$imageId) {
        $fullSizeSrc = preg_replace('/-\d+x\d+(?=\.\w+$)/', '', $bgImgUrl);
        $imageId = attachment_url_to_postid($fullSizeSrc);
      }
      // error_log('$imageId: ' . $imageId);

      if (!$imageId) continue; // Skip if image ID is not found
      // $node->setAttribute('create_custom_srcset-4', ''); // ???

      // Retrieve full image and x_small image URLs
      $imageFull = wp_get_attachment_image_src($imageId, 'full')[0];
      $x_small = wp_get_attachment_image_src($imageId, 'x_small')[0];

      // error_log('$imageFull: ' . $imageFull);
      // error_log('$x_small: ' . $x_small);

      $srcsetParts = [];
      foreach ($sizesToLabels as $size => $label) {
        $imageSrc = wp_get_attachment_image_src($imageId, $size);
        // error_log('For each: $imageId = ' . $imageId . ' size = ' . $size);
        // error_log('imageSrc: ' . print_r($imageSrc, true));

        // Corrected fallback logic
        if ($x_small === $imageFull && isset($imageSrc[0])) {
          if ($size === 'xxx_large') {
            $imageSrc = wp_get_attachment_image_src($imageId, 'full');
            // error_log('Creating the SrcSet: size = ' . $size . ' ; Thumb: ' . $imageSrc);
          }
          if ($size === 'xx_large' || $size === 'x_large') {
            $imageSrc = wp_get_attachment_image_src($imageId, 'large');
            // error_log('Creating the SrcSet: size = ' . $size . ' ; Thumb: ' . $imageSrc);
          }
          if ($size === 'xx_medium' || $size === 'x_medium') {
            $imageSrc = wp_get_attachment_image_src($imageId, 'medium_large');
            // error_log('Creating the SrcSet: size = ' . $size . ' ; Thumb: ' . $imageSrc);
          }
          if ($size === 'small' || $size === 'x_small') {
            $imageSrc = wp_get_attachment_image_src($imageId, 'medium');
            // error_log('Creating the SrcSet: size = ' . $size . ' ; Thumb: ' . $imageSrc);
          }
        }

        if ($imageSrc && isset($imageSrc[0])) {
          $srcsetParts[] = $imageSrc[0] . '|' . $label;
        }
      }

      $srcsetString = implode(', ', $srcsetParts);
      $srcsetStrings[$deviceType] = $srcsetString;
      // error_log('$srcsetString: ' . $srcsetString);
    }

    // Apply srcset strings to the node for each device type
    foreach ($srcsetStrings as $deviceType => $srcsetString) {
      $node->setAttribute('bg-srcset-' . $deviceType, $srcsetString);
    }

    $lowRezThumbnail = '';
    // If the image is not preloaded or lazy loaded, place the low resolution thumbnails
    if (
      !$node->hasAttribute('bg_lazy_loaded_desktop') ||
      !$node->hasAttribute('bg_lazy_loaded_tablet') ||
      !$node->hasAttribute('bg_lazy_loaded_mobile') ||
      !$node->hasAttribute('bg_preloaded_dekstop') ||
      !$node->hasAttribute('bg_preloaded_tablet') ||
      !$node->hasAttribute('bg_preloaded_mobile')
    ) {

      $selector = $node->getAttribute('data-selector');

      if ($node->hasAttribute('bg_url_desktop')) {
        $lowRezThumbDesktop = $node->getAttribute('bg_url_desktop');
        $lowRezThumbDesktop = self::get_smallest_thumbnail($lowRezThumbDesktop);
        $lowRezThumbnail .=  $selector . ':not(.background-auto-set) {background-image: url(' . $lowRezThumbDesktop . ') !important;}';
      }

      if ($node->hasAttribute('bg_url_tablet')) {
        $lowRezThumbTablet = $node->getAttribute('bg_url_tablet');
        $lowRezThumbTablet = self::get_smallest_thumbnail($lowRezThumbTablet);
        $lowRezThumbnail .=  '@media (min-width: 768px) and (max-width: 1023px) {' . $selector . ':not(.background-auto-set) {background-image: url(' . $lowRezThumbDesktop . ') !important;}';
      }

      if ($node->hasAttribute('bg_url_mobile')) {
        $lowRezThumbMobile = $node->getAttribute('bg_url_mobile');
        $lowRezThumbMobile = self::get_smallest_thumbnail($lowRezThumbMobile);
        $lowRezThumbnail .=  '@media (max-width: 767px) {' . $selector . ':not(.background-auto-set) {background-image: url(' . $lowRezThumbMobile . ') !important;}';
      }

      $lowRezThumbnails = '<style id="superspeed-loz-rez-bg-tbumns" type="text/css">' . $lowRezThumbnail . '</style>';

      // Add style elements to the top of the HTML
      global $ss_global_variable_for_inlining_css;
      $ss_global_variable_for_inlining_css .= $lowRezThumbnails;
    }
  }


  // Add an HTML element that will hold the settings for auto-setting image thumbnails
  public function auto_set_background_image_thumbs_settings()
  {

    // Option enable
    if (get_option('background_auto_set_thumbnails') !== '1') {
      return;
    }

    $image_info = get_option('background_auto_set_thumbs_info') === '1';

    if (!$image_info) {
      return;
    }

    $background_custom_pixel_density_desktop = get_option('background_custom_pixel_density_desktop') === '1';
    $background_custom_pixel_density_tablet = get_option('background_custom_pixel_density_tablet') === '1';
    $background_custom_pixel_density_mobile = get_option('background_custom_pixel_density_mobile') === '1';

    $image_info_value = 'false';
    if ($image_info) {
      $image_info_value = 'true';
    }

    $custom_or_native_desktop = 'native';
    if ($background_custom_pixel_density_desktop) {
      $custom_or_native_desktop = 'custom';
    }

    $custom_or_native_tablet = 'native';
    if ($background_custom_pixel_density_tablet) {
      $custom_or_native_tablet = 'custom';
    }

    $custom_or_native_mobile = 'native';
    if ($background_custom_pixel_density_mobile) {
      $custom_or_native_mobile = 'custom';
    }

    $pixel_density_desktop = floatval(get_option('backgound_auto_set_fixed_pixels_desktop'));
    $pixel_density_tablet = floatval(get_option('backgound_auto_set_fixed_pixels_tablet'));
    $pixel_density_mobile = floatval(get_option('backgound_auto_set_fixed_pixels_mobile'));

    echo '<script type="application/json" id="superspeed-fixed-pixel-density-bg-img">
   {
       "custom_or_native_desktop": "' . $custom_or_native_desktop . '",
       "pixel_density_desktop": ' . $pixel_density_desktop . ',
       "custom_or_native_tablet": "' . $custom_or_native_tablet . '",
       "pixel_density_tablet": ' . $pixel_density_tablet . ',
       "custom_or_native_mobile": "' . $custom_or_native_mobile . '",
       "pixel_density_mobile": ' . $pixel_density_mobile . ',
       "image_info": ' . $image_info_value . '
   }
   </script>
';
  }




  /********************** */

  // Dynamically load JS files of frontend tools for getting Images Data
  public function register_rest_routes()
  {
    // Define a common permission callback for all routes
    $permission_callback = function () {
      return current_user_can('manage_options');
    };

    // An array to define all the routes in one place
    $routes = [
      '/get-images-data/' => 'handle_get_images_data',
      '/delete-image-data/' => 'handle_refresh_image_data',
      '/delete-all-images-data/' => 'handle_delete_all_image_data',
    ];

    // Register each route with a loop
    foreach ($routes as $route => $callback) {
      register_rest_route('super-speed/v1', $route, [
        'methods' => 'POST',
        'callback' => [$this, $callback],
        'permission_callback' => $permission_callback,
      ]);
    }
  }


  // Dynamically load files JS of frontend tools for CSS
  public function enqueue_scripts_dynamically()
  {
    if (!current_user_can('manage_options')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      echo '<style id="getNewImagesData" getNewImagesData="true"></style>';
    } else {
      echo '<style id="getNewImagesData" getNewImagesData="false"></style>';
    }

    // Server paths to the JS files
    $script_base_path = plugin_dir_path(dirname(__FILE__, 2)) . 'assets/js/';

    global $ss_global_variable_for_version_control;

    $get_images_data_script_version = $ss_global_variable_for_version_control;
    $refresh_images_data_script_version = $ss_global_variable_for_version_control;

    // Determine script versions
    if (!$ss_global_variable_for_version_control) {

      $get_images_data_script_path = $script_base_path . 'get-images-data.js';
      $get_images_data_script_version = filemtime($get_images_data_script_path);

      $refresh_images_data_script_path = $script_base_path . 'refresh-images-data.js';
      $refresh_images_data_script_version = filemtime($refresh_images_data_script_path);
    }

    // Base URL for the scripts
    $script_base_url = plugins_url('assets/js/', dirname(__FILE__, 2));

  ?>
    <script id="super-speed-data-for-dynamically-loaded-image-scripts">
      window.dynamicallyLoadedFilesDataImages = {
        getImagesDataUrl: '<?php echo esc_js(rest_url('super-speed/v1/get-images-data/')); ?>',
        deleteImageDataUrl: '<?php echo esc_js(rest_url('super-speed/v1/delete-image-data/')); ?>',
        deleteAllImagesDataUrl: '<?php echo esc_js(rest_url('super-speed/v1/delete-all-images-data/')); ?>',
        imagesNonce: '<?php echo wp_create_nonce('wp_rest'); ?>',
      };
    </script>
    <?php

    ?>
    <script id="super-speed-enqueue-images-data-scripts-dynamically" defer>
      (function() {
        function initializeScript() {

          if (!document.body.classList.contains('can-edit') || !document.getElementById('wpadminbar')) {
            return;
          }

          // Function to load scripts
          function loadScript(src) {
            var script = document.createElement('script');
            script.src = src;
            document.body.appendChild(script);
          }

          const scriptBaseUrl = '<?php echo $script_base_url; ?>';

          // Don't load in an iframe, which is used for getting image data
          function handleRefreshImagesScriptLoading() {

            if (window === window.top) {
              const refreshImagesFileVersion = '<?php echo $refresh_images_data_script_version; ?>';
              var refreshImagesDataScriptSrc = scriptBaseUrl + 'refresh-images-data.js?ver=' + refreshImagesFileVersion;
              loadScript(refreshImagesDataScriptSrc);
            }
          }

          // Retrieve the element to check if new images data is needed
          const getNewImagesDataCheck = document.getElementById('getNewImagesData');

          if (!getNewImagesDataCheck) {
            console.error("Critical error: #getNewImagesData element not found");
            return;
          }

          const getNewImagesData = getNewImagesDataCheck.getAttribute('getNewImagesData');

          // Function to dynamically load the script if new images data is needed
          function getImagesData() {

            // Correctly compare string value to "true"
            if (getNewImagesData === "true") {
              const getImagesFileVersion = '<?php echo $get_images_data_script_version; ?>';
              var getImagesDataScriptSrc = scriptBaseUrl + 'get-images-data.js?ver=' + getImagesFileVersion;
              loadScript(getImagesDataScriptSrc);
            }
          }

          // Function to handle conditional loading
          function handleGetImagesScriptLoading() {
            if (window.UniqueScriptExecutionState && window.UniqueScriptExecutionState.cssLoaded) {
              getImagesData();
            } else if (!document.querySelector('script[src*="get-used-css.js"]')) {
              getImagesData();
            } else {
              setTimeout(getImagesData, 100);
            }
          }

          handleRefreshImagesScriptLoading();
          handleGetImagesScriptLoading();
        }

        if (document.readyState === 'complete') {
          initializeScript();
        } else {
          window.addEventListener('load', initializeScript);
        }
      })();
    </script>

<?php


  }


  /********************** */

  public function handle_get_images_data(\WP_REST_Request $request)
  {
    $nonce = $request->get_header('X-WP-Nonce');

    if (!wp_verify_nonce($nonce, 'wp_rest')) {
      return new \WP_Error('nonce_verification_failed', 'Nonce verification failed', ['status' => 403]);
    }

    $data = json_decode($request->get_body(), true);

    if (!isset($data['pageUrl'])) {
      return new \WP_Error('missing_data', 'The page URL is missing from the request.', ['status' => 400]);
    }

    $option_name = 'images_data_for_page_' . hash('sha256', $data['pageUrl']);

    // First, check if the option already exists
    if (false !== get_option($option_name)) {
      // Option exists, delete it first
      delete_option($option_name);
    }

    // Now, add the new data. Note that add_option returns true if the option was successfully added
    $add_result = add_option($option_name, $data, '', 'no'); // '' is a deprecated argument, included for compatibility

    // Check if adding the new option was successful
    if (!$add_result) {
      return new \WP_Error('option_add_failed', 'Failed to add option', ['status' => 500]);
    }

    return new \WP_REST_Response('Success', 200); // Ensure a successful response is appropriately returned
  }


  // Add the url of the post or page, so that it can be used to remove imges data if the post was edited
  public function add_meta_with_post_url()
  {
    $current_url = Utility::get_current_page_url();

    // Save the $template_type as post meta so that we know if the post has been edited
    update_post_meta(get_the_ID(), 'post_edited_imaged_data', $current_url);
  }

  // When the post is edited, remove images data
  public function if_post_is_edited_remove_images_data($post_id, $post, $update)
  {

    // Exit if doing an autosave or if it's a post revision.
    if (defined('DOING_AUTOSAVE') && DOING_AUTOSAVE || wp_is_post_revision($post_id)) {
      return;
    }

    // This check ensures the function only runs on updates, not post creation.
    if (!$update) {
      return;
    }

    $current_url = get_post_meta($post_id, 'post_edited_imaged_data', true);
    $option_name = 'images_data_for_page_' . hash('sha256', $current_url);
    delete_option($option_name);
  }


  // Remove image data for specific page from the frontend
  public function handle_refresh_image_data(\WP_REST_Request $request)
  {
    $currentUrlHash = sanitize_text_field($request->get_param('currentUrl'));
    $optionName = 'images_data_for_page_' . $currentUrlHash;

    if (get_option($optionName) !== false) {
      if (delete_option($optionName)) {
        // Correctly use WP_REST_Response for successful operations
        return new \WP_REST_Response(['message' => 'Image data removed successfully', 'success' => true], 200);
      } else {
        return new \WP_Error('failed_to_remove_image_data', 'Failed to remove image data', ['status' => 500]);
      }
    } else {
      return new \WP_Error('no_option', 'Failed to remove image data - Option does not exist', ['status' => 404]);
    }
  }



  // Function to delete all image data
  public function handle_delete_all_image_data(\WP_REST_Request $request)
  {
    $nonce = $request->get_header('X-WP-Nonce');
    if (!wp_verify_nonce($nonce, 'wp_rest')) {
      return new \WP_Error('nonce_verification_failed', 'Nonce verification failed', ['status' => 403]);
    }

    if (!current_user_can('delete_others_posts')) {
      wp_send_json_error('Insufficient permissions');
      return;
    }

    global $wpdb;
    $pattern = $wpdb->esc_like('images_data_for_page_') . '%';
    $options = $wpdb->get_col("SELECT option_name FROM $wpdb->options WHERE option_name LIKE '$pattern'");

    foreach ($options as $option) {
      delete_option($option);
    }

    // Use WP_REST_Response for REST API success message
    return new \WP_REST_Response(['message' => 'All image data removed successfully', 'success' => true], 200);
  }



  /********************* */


  /*** LAZY LOAD IFRAMES ***/

  public function buffer_start_lazyload_iframes()
  {

    if (get_option('lazy_load_iframes') !== '1') {
      return;
    }

    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    // Exit early, if it should not run on this page
    $pages_exclude_list = get_option('lazy_iframe_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lazy_iframe_exclude_pages_textarea'))) : [];
    $pages_exclude = get_option('lazy_iframe_exclude_pages_cbx') === '1';
    $exclude_home = get_option('lazy_iframe_exclude_home_cbx') === '1';

    if ($exclude_home && Utility::is_homepage()) {
      return;
    }

    // Run on all except these excluded pages
    elseif ($pages_exclude) {
      // Don't run on pages from the list
      if (!empty($pages_exclude_list)) {
        $current_Url = Utility::get_current_page_url();
        foreach ($pages_exclude_list as $string) {
          $normalizedString = Utility::normalizeUrlString($string);
          if (stripos($current_Url, $normalizedString) !== false) {
            return;
          }
        }
      }
    }

    ob_start(array($this, "lazy_load_iframes"));
    $this->buffering_started_lazyload_iframes = true;
  }

  public function lazy_load_iframes($html)
  {
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    // Create a new DOMXPath instance
    $xpath = new \DOMXPath($dom);

    // Fetch all iFrame elements that do not contain "youtu" in their src attribute
    $iframes = $xpath->query('//iframe[not(contains(@src, "youtu"))]');

    // Fetch options
    $exclude_selector = get_option('lazy_iframe_exclude_selector') === '1';
    $exclude_parent_selector = get_option('lazy_iframe_exclude_parent_selector') === '1';
    $exclude_selector_list = get_option('lazy_iframe_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_iframe_exclude_selector_textarea'))) : [];
    $exclude_parent_selector_list = get_option('lazy_iframe_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_iframe_exclude_parent_selector_textarea'))) : [];
    $gray_placeholder = get_option('lazy_iframe_placeholder') === '1';


    // Iterate over each iframe element
    foreach ($iframes as $iframe) {
      $classes = explode(' ', $iframe->getAttribute('class')); // Get all classes from the iframe

      // Exclude iframes by selector
      if ($exclude_selector && array_intersect($exclude_selector_list, $classes)) {
        $iframe->removeAttribute('loading');
        continue;
      }

      // Exclude iframes by parent selector
      if ($exclude_parent_selector && $this->hasParentWithClass($iframe, $exclude_parent_selector_list)) {
        $iframe->removeAttribute('loading');
        continue;
      }

      // Lazy-load all iframes that are not excluded
      $iframe->setAttribute('loading', 'lazy');
      $iframe->setAttribute('lazy_loaded_iframe', '');


      // If the option to add the gray placeholder is enabled, wrap the iframes and add the gray placeholder div
      if ($gray_placeholder) {

        $iframe->setAttribute('iframe_placeholder', 'gray');

        // Check if $classes is an array and not empty
        $classString = '';
        if (is_array($classes) && !empty($classes)) {
          $classString = implode(' ', $classes);
        }

        // Add our class, with space before it, in case there are existing classes
        $classString .= ' ss-iframe-placeholder';

        $placeholderWidth = $iframe->getAttribute('width');
        $placeholderHeight = $iframe->getAttribute('height');
        $placeholderStyle = $iframe->getAttribute('style');

        // Gray SVG for the gray placeholder
        $placeholderSrc = 'data:image/svg+xml;utf8,%3Csvg%20xmlns%3D%22http%3A%2F%2Fwww.w3.org%2F2000%2Fsvg%22%3E%3Crect%20width%3D%2210000%22%20height%3D%2210000%22%20fill%3D%22gray%22%3E%3Canimate%20attributeName%3D%22fill%22%20values%3D%22lightgray%3B%20darkgray%3B%20lightgray%22%20dur%3D%222s%22%20repeatCount%3D%22indefinite%22%20%2F%3E%3C%2Frect%3E%3C%2Fsvg%3E';

        // Create the cloned iframe placeholder
        $clonedIframePlaceholder = $dom->createElement('iframe');

        // Set attributes
        $clonedIframePlaceholder->setAttribute('class', $classString);
        $clonedIframePlaceholder->setAttribute('width', $placeholderWidth);
        $clonedIframePlaceholder->setAttribute('height', $placeholderHeight);
        $clonedIframePlaceholder->setAttribute('style', $placeholderStyle);
        $clonedIframePlaceholder->setAttribute('src', $placeholderSrc);

        // Check if the original iframe has a next sibling
        if ($iframe->nextSibling) {

          // Place the cloned iframe placeholder after the original iframe
          $iframe->parentNode->insertBefore($clonedIframePlaceholder, $iframe->nextSibling);
        } else {

          // If the original iframe is the last child, append the cloned iframe
          $iframe->parentNode->appendChild($clonedIframePlaceholder);
        }
      }
    }

    return $dom->saveHTML();
  }

  // Already performed the check if we are on the right page, and stored it in the variable '$this->buffering_started_critical_css'
  public function buffer_end_lazyload_iframes()
  {
    if ($this->buffering_started_lazyload_iframes) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_lazyload_iframes = false;
    }
  }

  // CSS for iFrames options: placeholder, mark excluded
  public function css_for_lazy_loaded_iframe()
  {

    if (get_option('lazy_load_iframes') !== '1') {
      return;
    }

    // Visually mark lazy load iframes
    if (get_option('lazyloaded_iframe_visually_mark') === '1') {
      echo '<style id="superspeed-mark-lazy-iframes" type="delayed-style">
      .can-edit iframe[lazy_loaded_iframe]:not([src*="youtu"]) {
        outline: dashed 0.25rem blue !important;
        border: solid 0.25rem white !important;
        outline-offset: -0.25rem !important;
     }
  </style>';
    }

    // Display a gray placeholder until the iframe is loaded
    if (get_option('lazy_iframe_placeholder') === '1') {
      echo '<style id="superspeed-lazy-iframe-placeholder" type="text/css">
        /* Needs to be relative, for the gray placeholder */
        iframe[lazy_loaded_iframe]:not([src*="youtu"], .ss-iframe-loaded) {
          translate: -1000vw;
          position: absolute;
          opacity: 0;
        }
        </style>';
    }
  }




  /*** LAZY LOAD VIDEOS ***/

  public function buffer_start_lazyload_videos()
  {
    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    if (get_option('enable_lazy_load_videos') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    $exclude_home = get_option('lazy_video_exclude_home_cbx') === '1';
    $exclude_pages = get_option('lazy_video_exclude_pages_cbx') === '1';
    $exclude_list = get_option('lazy_video_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('lazy_video_exclude_pages_textarea'))) : [];

    if ($exclude_home && Utility::is_homepage()) {
      return;
    }

    // Run on all except these excluded pages
    elseif ($exclude_pages) {
      // Don't run on pages from the list
      if (!empty($exclude_list)) {
        $current_Url = Utility::get_current_page_url();
        foreach ($exclude_list as $string) {
          $normalizedString = Utility::normalizeUrlString($string);
          if (stripos($current_Url, $normalizedString) !== false) {
            return;
          }
        }
      }
    }

    ob_start(array($this, "lazy_load_videos"));
    $this->buffering_started_lazyload_videos = true;
  }

  public function lazy_load_videos($html)
  {
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    // Create a new DOMXPath instance
    $xpath = new \DOMXPath($dom);

    // Fetch all video elements that contain "youtu" in their src attribute
    $videos = $xpath->query('//iframe[contains(@src, "youtu")]');

    // Fetch options
    $exclude_selector = get_option('lazy_video_exclude_selector') === '1';
    $exclude_parent_selector = get_option('lazy_video_exclude_parent_selector') === '1';
    $exclude_selector_list = get_option('lazy_video_exclude_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_video_exclude_selector_textarea'))) : [];
    $exclude_parent_selector_list = get_option('lazy_video_exclude_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_video_exclude_parent_selector_textarea'))) : [];
    $preload_selector = get_option('lazy_video_thumb_preload_selector') === '1';
    $preload_parent_selector = get_option('lazy_video_thumb_preload_parent_selector') === '1';
    $preload_selector_list = get_option('lazy_video_thumb_preload_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_video_thumb_preload_selector_textarea'))) : [];
    $preload_parent_selector_list = get_option('lazy_video_thumb_preload_parent_selector_textarea') ? array_map('trim', explode("\n", get_option('lazy_video_thumb_preload_parent_selector_textarea'))) : [];

    // Create the SVG DOM element from the SVG XML string
    $svgDom = new \DOMDocument();
    $svgDom->loadXML('<svg height="100%" version="1.1" viewBox="0 0 68 48" width="65" style="width: 65px; height: auto;"><path class="ytp-large-play-button-bg" d="M66.52,7.74c-0.78-2.93-2.49-5.41-5.42-6.19C55.79,.13,34,0,34,0S12.21,.13,6.9,1.55 C3.97,2.33,2.27,4.81,1.48,7.74C0.06,13.05,0,24,0,24s0.06,10.95,1.48,16.26c0.78,2.93,2.49,5.41,5.42,6.19 C12.21,47.87,34,48,34,48s21.79-0.13,27.1-1.55c2.93-0.78,4.64-3.26,5.42-6.19C67.94,34.95,68,24,68,24S67.94,13.05,66.52,7.74z" fill="#f00"></path><path d="M 45,24 27,14 27,34" fill="#fff"></path></svg>');
    $svg = $dom->importNode($svgDom->documentElement, true);

    // Attached to he head, this variable is used for preloading
    global $ss_global_variable_for_preloading_images;

    // Initialize an empty array to hold schema markups
    $schema_markups = [];

    // Generate a style tag to hold our CSS
    $styleTag = $dom->createElement('style');

    $has_preloaded_class = false;
    $has_preloaded_parent_class = false;

    foreach ($videos as $video) {

      // Mark, so that users can know that the plugin can't lazy load the video, since it's not in the HTML
      $video->setAttribute('embeded', 'true');

      // Get width and height to calculate aspect ratio
      $width = $video->getAttribute('width');
      $height = $video->getAttribute('height');
      $aspectRatio = "$width/$height";
      $video->setAttribute('style', "aspect-ratio: $aspectRatio;");

      $classes = explode(' ', $video->getAttribute('class')); // Get all classes from the video

      // Exclude videos by selector
      if ($exclude_selector && array_intersect($exclude_selector_list, $classes)) {
        $video->removeAttribute('loading');
        continue;
      }

      // Exclude videos by parent selector
      if ($exclude_parent_selector && $this->hasParentWithClass($video, $exclude_parent_selector_list)) {
        $video->removeAttribute('loading');
        continue;
      }

      // Extract video ID from the src attribute
      $src = $video->getAttribute('src');
      preg_match('/youtu.*?(?:\/|v=|vi=)([a-zA-Z0-9_-]{11})/', $src, $matches);
      $videoId = $matches[1] ?? null;

      // Check for matching classes in the video iframe
      if (!$has_preloaded_class && ($preload_selector && !empty($preload_selector_list)) && (array_intersect($preload_selector_list, $classes))) {
        // Add preload links to the global variable
        $ss_global_variable_for_preloading_images .= "\n<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/default.webp\" as=\"image\" fetchpriority=\"high\" media=\"(max-width: 767px)\">\n";
        $ss_global_variable_for_preloading_images .= "<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/mqdefault.webp\" as=\"image\" fetchpriority=\"high\" media=\"(min-width: 990px) and (max-width: 1024px)\">\n";
        $ss_global_variable_for_preloading_images .= "<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/hqdefault.webp\" as=\"image\" fetchpriority=\"high\" media=\"(min-width: 1024.1px)\">\n";

        $video->setAttribute('yt-bg-preloaded', '');
        $has_preloaded_class = true;
      }

      // Check for matching classes in the parent element
      if (!$has_preloaded_parent_class && ($preload_parent_selector && !empty($preload_parent_selector_list))) {

        if ($this->hasParentWithClass($video, $preload_parent_selector_list)) {
          // Add preload links to the global variable
          $ss_global_variable_for_preloading_images .= "\n<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/default.webp\" as=\"image\" fetchpriority=\"high\" media=\"(max-width: 767px)\">\n";
          $ss_global_variable_for_preloading_images .= "<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/mqdefault.webp\" as=\"image\" fetchpriority=\"high\" media=\"(min-width: 990px) and (max-width: 1024px)\">\n";
          $ss_global_variable_for_preloading_images .= "<link rel=\"preload\" href=\"https://i.ytimg.com/vi_webp/$videoId/hqdefault.webp\" as=\"image\" fetchpriority=\"high\" media=\"(min-width: 1024.1px)\">\n";

          $video->setAttribute('yt-bg-preloaded', '');
          $has_preloaded_parent_class = true;
        }
      }

      $title = $video->getAttribute('title');
      $classList = $video->getAttribute('class');

      // Create the placeholder div
      $placeholder = $dom->createElement('div');
      $placeholder->setAttribute('title', $title);
      $placeholder->setAttribute('class', "ss-video-placeholder ss-video-thumb-" . $videoId . ' ' . $classList);
      $placeholder->setAttribute('style', "aspect-ratio: $aspectRatio;");

      // Store the necessary iframe attributes as data attributes on the placeholder
      $placeholder->setAttribute('data-src', $src);
      $placeholder->setAttribute('data-title', $title);
      $placeholder->setAttribute('data-width', $width);
      $placeholder->setAttribute('data-height', $height);
      $placeholder->setAttribute('data-style', "aspect-ratio: $aspectRatio;");

      // If we are preloading the background image
      if ($video->hasAttribute('yt-bg-preloaded')) {
        $placeholder->setAttribute('yt-bg-preloaded', '');
      }

      // Create inner elements for the placeholder
      $innerTitle = $dom->createElement('div', $title);
      $innerTitle->setAttribute('class', 'ss-yt-title');
      $placeholder->appendChild($innerTitle);

      // Clone the pre-made SVG and append it to the placeholder
      $svgClone = $svg->cloneNode(true);
      $placeholder->appendChild($svgClone);

      // Append the placeholder to the DOM, right after the video
      $video->parentNode->insertBefore($placeholder, $video->nextSibling);

      // Remove the iframe itself or just hide it using CSS
      $video->parentNode->removeChild($video);

      // Create the CSS rules
      $css = <<<CSS
      /* Desktop */
      .ss-video-thumb-$videoId {
        background-image: url('https://i.ytimg.com/vi_webp/$videoId/hqdefault.webp');
      }

      /* Tablet */
      @media (min-width: 768px) and (max-width: 1023px) {
        .ss-video-thumb-$videoId {
          background-image: url('https://i.ytimg.com/vi_webp/$videoId/mqdefault.webp');
        }
      }

      /* Mobile */
      @media (max-width: 767px) {
        .ss-video-thumb-$videoId {
          background-image: url('https://i.ytimg.com/vi_webp/$videoId/default.webp');
        }
      }
      CSS;

      // Add these CSS rules to the style tag
      $styleTag->nodeValue .= $css;

      // Create schema markup for each video
      $schema_markup = [
        "@context" => "http://schema.org",
        "@type" => "VideoObject",
        "name" => $title,
        "thumbnailUrl" => "https://img.youtube.com/vi/{$videoId}/0.jpg",
        "uploadDate" => "", // Ideally, this should be the post's creation date. Fill in accordingly.
        "embedUrl" => $src,
        "contentUrl" => "https://www.youtube.com/watch?v={$videoId}"
      ];

      // Append to the array of schema markups
      $schema_markups[] = $schema_markup;
    }


    // After the loop
    // Append each schema markup to the HTML head
    $head = $dom->getElementsByTagName('head')->item(0);
    foreach ($schema_markups as $markup) {
      $scriptTag = $dom->createElement('script');
      $scriptTag->setAttribute('type', 'application/ld+json');
      $scriptTag->nodeValue = json_encode($markup, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT);
      $head->appendChild($scriptTag);
    }

    // Append the style tag to the head element
    $head = $dom->getElementsByTagName('head')->item(0);
    $head->appendChild($styleTag);

    return $dom->saveHTML();
  }


  // Already performed the check if we are on the right page, and stored it in the variable '$this->buffering_started_critical_css'
  public function buffer_end_lazyload_videos()
  {
    if ($this->buffering_started_lazyload_videos) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_lazyload_videos = false;
    }
  }


  // CSS for Videos options: placeholder, mark excluded, mark preloaded thumbnail
  public function css_for_lazy_loaded_videos()
  {
    if (get_option('enable_lazy_load_videos') !== '1') {
      return;
    }

    $youtubeStyles = '';

    // Visually mark non-embeded YouTube videos
    if (get_option('visually_mark_non_embeded_yt') === '1') {
      $youtubeStyles .= '  
        .can-edit iframe[src*="youtu"]:not([embeded]) {
          outline: dashed 0.25rem green !important;
          border: solid 0.25rem white !important;
          outline-offset: -0.25rem !important;
        }
    ';
    }

    // Visually mark videos lazy loaded using JS
    if (get_option('lazyloaded_video_visually_mark_js') === '1') {
      $youtubeStyles .= '
      .can-edit iframe[src*="youtu"][video-lazyloaded],
      .can-edit .ss-video-placeholder[data-src] {
        outline: dashed 0.25rem blue !important;
        border: solid 0.25rem white !important;
        outline-offset: -0.25rem !important;
      }
        ';
    }

    // Visually mark preloaded thumbnails
    if (get_option('preloaded_thumbnail_visually_mark') === '1') {
      $youtubeStyles .= '
      .can-edit iframe[src*="youtu"][yt-bg-preloaded],
      .can-edit .ss-video-placeholder[yt-bg-preloaded] {
        outline: dashed 0.25rem red !important;
        border: solid 0.25rem white !important;
        outline-offset: -0.25rem !important;
      }

      /* blue + red */
      /* Video lazy-loaded. Thumbnail preloaded */
      .can-edit iframe[src*="youtu"][video-lazyloaded][yt-bg-preloaded], 
      .can-edit .ss-video-placeholder[yt-bg-preloaded] {
        outline: dashed 0.33rem blue !important;
        border: solid 0.33rem red !important;
        outline-offset: -0.33rem !important;
    }
        ';
    }

    // Display only the thumbnail until user click intent
    $youtubeStyles .= '
      iframe[src*="youtu"] {
        width: 100%;
        height: auto;
        background-color: black;
        vertical-align: bottom;
      }
      
      .video-thumbnail {
        display: flex;
        align-items: center;
        justify-content: center;
        cursor: pointer;
        background-repeat: no-repeat;
        background-color: black;
        background-position: center;
        background-size: cover;
      }
      
      .ss-video-placeholder {
        background-repeat: no-repeat;
        background-size: cover;
        background-position: center;
        position: relative;
        display: flex;
        align-items: center;
        justify-content: center;
        cursor: pointer;
        background-color: black;
        width: 100%;
      }
      
      .ss-yt-title {
        font-family: "YouTube Noto", Roboto, Arial, Helvetica, sans-serif;
        color: #eee;
        text-align: left;
        line-height: 1.3;
        -webkit-font-smoothing: antialiased;
        -webkit-tap-highlight-color: rgba(0, 0, 0, 0);
        text-shadow: 0 0 2px rgba(0, 0, 0, 0.5);
        white-space: nowrap;
        word-wrap: normal;
        font-size: 18px;
        display: block;
        text-overflow: ellipsis;
        overflow: hidden;
        text-decoration: none;
        -webkit-transition: color 0.1s cubic-bezier(0, 0, 0.2, 1);
        -o-transition: color 0.1s cubic-bezier(0, 0, 0.2, 1);
        transition: color 0.1s cubic-bezier(0, 0, 0.2, 1);
        outline: 0;
        position: absolute;
        top: 0;
        left: 0;
        width: 100%;
        background: linear-gradient(#00000061, transparent);
        padding: 25px 12.5px 0;
      }
        ';

    if ($youtubeStyles) {
      echo '<style id="superspeed-youtube" type="text/css">' . $youtubeStyles . '</style>
      ';
    }
  }



  // Adds a dynamic class to featured images based on the type of post (including CPTs)
  public function add_class_to_featured_images($html, $post_id)
  {
    // Get the post using the ID
    $post = get_post($post_id);
    $suffix = 'featured_image_';

    // Check if the post is valid and determine the post type
    if ($post) {
      $post_type = $post->post_type;
      $new_class = "{$suffix}-{$post_type}";

      // Modify the HTML to add the new class
      $html = str_replace('class="', 'class="' . $new_class . ' ', $html);
    }

    return $html;
  }

  // Add a class to the WooCommerce main product image
  public function add_class_to_main_woo_product_image($html)
  {
    return str_replace('class="', 'class="featured_image_woo_product ', $html);
  }
}
