<?php

namespace SuperSpeed;

use SuperSpeed\Utility;

class ScriptHandler
{

  // cache the result of checking which page we are on, before removing scripts
  private $buffering_started_remove_scripts = false;

  // cache the result of checking which page we are on, before defering scripts
  private $buffering_started_defer_scripts = false;

  // cache the result of checking which page we are on, before delaying scripts
  private $buffering_started_delay_scripts = false;

  public function __construct()
  {
    // Exit if we could not get the current URL. Skipping static resource because it is not a page.
    $current_url = Utility::get_current_page_url();
    if (!$current_url) {
      return;
    }

    if (!Utility::is_page_editor_or_wp_backend()) {

      // Remove scripts
      add_action('template_redirect', array($this, 'buffer_start_remove_scripts'), 9999);
      add_action('shutdown', array($this, 'buffer_end_remove_scripts'), PHP_INT_MAX);

      // Defer external scripts
      add_action('wp_print_scripts', array($this, 'collect_all_script_dependencies'));
      add_filter('script_loader_tag', array($this, 'add_attributes_to_dependent_scripts'), 10, 2);

      // Defer scripts
      add_action('template_redirect', array($this, 'buffer_start_defer_scripts'), 9999);
      add_action('shutdown', array($this, 'buffer_end_defer_scripts'), PHP_INT_MAX);

      // Delay  scripts
      add_action('template_redirect', array($this, 'buffer_start_delay_scripts'), 9999);
      add_action('shutdown', array($this, 'buffer_end_delay_scripts'), PHP_INT_MAX);
    }
  }

  /*** DEQUEUE FRONTEND SCRIPTS ***/

  // Start buffer
  public function buffer_start_remove_scripts()
  {

    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    if (get_option('ss_enable_dequeue') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      return;
    }

    // If the user has enabled auto-setting Used CSS
    // If the CSS file is missing
    // If the page is new or edited
    // And if we should be getting Used CSS on this page
    if (Utility::getting_used_css_data()) {
      return;
    }

    // PAGES - Get radio buttons and text areas
    $checkedDeqRadio = get_option('ss_deq_radio_checked');
    $exclude_list = get_option('ss_exc_deq_js_textarea') ? array_map('trim', explode("\n", get_option('ss_exc_deq_js_textarea'))) : [];
    $include_list = get_option('ss_deq_spec_js_textarea') ? array_map('trim', explode("\n", get_option('ss_deq_spec_js_textarea'))) : [];

    // Add home page settings checkboxses
    $exclude_home = get_option('ss_dequeue_js_exclude_home') === '1';
    $include_home = get_option('ss_dequeue_js_include_home') === '1';

    $all_pages = $checkedDeqRadio === 'dequeue_all_pages';
    $exclude_pages = $checkedDeqRadio === 'exclude_remove_pages';
    $include_pages = $checkedDeqRadio === 'dequeue_specific_pages';

    $current_URL = '';
    $is_home = '';

    if (!$all_pages) {
      $is_home = Utility::is_homepage();
    }

    // Run on all pages
    if ($all_pages) {
      ob_start(array($this, "remove_external_scripts"));
      $this->buffering_started_remove_scripts = true;
      return;
    }

    // Run on all except these excluded pages
    elseif ($exclude_pages) {
      if ($is_home && $exclude_home) {
        // Don't run on the homepage
        $this->buffering_started_remove_scripts = false;
        return;
      } else {
        // Don't run on pages from the list
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              $this->buffering_started_remove_scripts = false;
              return;
            }
          }
        }
      }

      // Run if we are not on an excluded page
      $this->buffering_started_remove_scripts = true;
      ob_start(array($this, "remove_external_scripts"));
      return;
    }

    // Run only on specified pages
    elseif ($include_pages) {
      if ($is_home && $include_home) {
        // Run on Home page
        ob_start(array($this, "remove_external_scripts"));
        $this->buffering_started_remove_scripts = true;
        return;
      } else {
        if (!empty($include_list)) {
          foreach ($include_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              ob_start(array($this, "remove_external_scripts"));
              $this->buffering_started_remove_scripts = true;
              return;
            }
          }
        }
      }
    }
  }


  // Removes <script> tags from HTML based on user-configured settings in WordPress options
  public function remove_external_scripts($html)
  {
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }
    $xpath = new \DOMXPath($dom);

    // Exclusion list that should never be dequeued
    $never_remove_list = array(
      'superspeed',
      'super-speed',
      'wp-includes/js/jquery/jquery',
    );

    // Retrieve options from WordPress
    $checkedDeqLinksJSRadio = get_option('ss_dequeue_js_links_radio_checked');
    $exclude_list = get_option('ss_dequeue_exclude_links_js_textarea') ? array_map('trim', explode("\n", get_option('ss_dequeue_exclude_links_js_textarea'))) : [];
    $include_list = get_option('ss_dequeue_spacific_links_js_textarea') ? array_map('trim', explode("\n", get_option('ss_dequeue_spacific_links_js_textarea'))) : [];
    $exclude_3rd_party = get_option('dequeue_scripts_exclude_3rd_party') === '1';


    // Exclude 3rd party styles
    if ($exclude_3rd_party) {
      // Get the site URL and parse it to get components
      $parsed_url = parse_url(get_site_url());
      $base_path = $parsed_url['scheme'] . '://' . $parsed_url['host'];      

      // This will select script tags that either do not have a src attribute (inline scripts)
      // or have a src attribute that starts with the site's domain (local scripts)
      $nodes = $xpath->query("//script[not(@src) or starts-with(@src, '{$base_path}')]");
    } else {
      // Select inline and external scripts
      $nodes = $xpath->query("//script");
    }


    foreach ($nodes as $node) {
      $src = $node->getAttribute('src');
      $id = $node->getAttribute('id');
      $combinedAttributes = $src . ' ' . $id; // Combine src and id into a single string

      $shouldSkip = false; // Flag to check if current node should be skipped

      // Early skip if the string is in the never remove list
      foreach ($never_remove_list as $never_remove_item) {
        if (stripos($combinedAttributes, trim($never_remove_item)) !== false) {
          $shouldSkip = true;
        }
      }

      if ($shouldSkip) {
        continue; // Skip to the next node in the outer loop
      }

      $remove = false; // Flag to determine if the node should be removed

      if ($checkedDeqLinksJSRadio === 'dequeue_all_links_js') {
        $remove = true;
      } elseif ($checkedDeqLinksJSRadio === 'dequeue_exclude_links_js') {
        $remove = true; // Assume removal unless an exclusion match is found
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $exclude) {
            if (stripos($combinedAttributes, trim($exclude)) !== false) {
              $remove = false;
              break;
            }
          }
        }
      } elseif ($checkedDeqLinksJSRadio === 'dequeue_specific_links_js') {
        if (!empty($include_list)) {
          foreach ($include_list as $include) {
            if (stripos($combinedAttributes, trim($include)) !== false) {
              $remove = true;
              break;
            }
          }
        }
      }

      if ($remove) {
        $node->parentNode->removeChild($node);
      }
    }

    return $dom->saveHTML();
  }

  // End buffer for removing scripts
  public function buffer_end_remove_scripts()
  {
    if ($this->buffering_started_remove_scripts) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_remove_scripts = false;
    }
  }



  /*** DEFER FRONTEND SCRIPTS ***/

  // Collect all script dependencies
  public $script_supports = array();

  public function collect_all_script_dependencies()
  {

    if (get_option('ss_enable_defer_ext_js') !== '1') {
      return;
    }

    global $wp_scripts;

    foreach ($wp_scripts->queue as $handle) {
      $deps = $wp_scripts->registered[$handle]->deps;
      if ($deps) {
        $this->script_supports = array_merge($this->script_supports, $deps);
      }
    }

    $this->script_supports = array_unique($this->script_supports); // remove duplicates
  }


  // Mark support and dependant scripts
  public function add_attributes_to_dependent_scripts($tag, $handle)
  {
    global $wp_scripts;

    if (get_option('ss_enable_defer_ext_js') !== '1') {
      return $tag;
    }

    $support_scripts = get_option('ss_defer_exclude_support') === '1';
    $dependand_scripts = get_option('ss_defer_exclude_dependent') === '1';

    // If checkmarks are not used
    if (!$support_scripts && !$dependand_scripts) {
      return $tag;
    }

    if ($dependand_scripts) {
      // Ensure the script is registered before trying to access its properties
      if (wp_script_is($handle, 'registered')) {
        // Get the dependencies of the script.
        $dependencies = $wp_scripts->registered[$handle]->deps;

        // If the script has dependencies, add the attribute to its tag.
        if (!empty($dependencies)) {
          $dependency_str = implode(',', $dependencies);
          $tag = str_replace('<script', "<script loading-dependency='{$dependency_str}'", $tag);
        }
      }

      // jQuery migrate is dependent on jQuery core
      if ($handle === 'jquery-migrate') {
        $tag = str_replace('<script', '<script loading-dependency="jquery-core"', $tag);
      }
    }

    if ($support_scripts) {
      // If the script is a dependency for another script, add the 'is-support' attribute.
      if (in_array($handle, $this->script_supports)) {
        $tag = str_replace('<script', '<script is-support="true"', $tag);
      }

      // Add 'is-support' attribute to the 'jquery-core' and 'jquery-migrate' handles manually
      if ($handle === 'jquery-core' || $handle === 'jquery-migrate') {
        $tag = str_replace('<script', '<script is-support="true"', $tag);
      }
    }

    return $tag;
  }


  /*** DEFER FRONTEND SCRIPTS ***/

  // Start buffer
  public function buffer_start_defer_scripts()
  {

    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    if (get_option('ss_enable_defer_ext_js') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    ob_start(array($this, "defer_scripts"));
    $this->buffering_started_defer_scripts = true;
  }


  // Defer scritps
  public function defer_scripts($html)
  {
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    $xpath = new \DOMXPath($dom);
    $nodes = $xpath->query("//script");

    // Exclusion list that should never be defered
    $never_remove_list = array(
      'superspeed',
      'jquery',
      'admin-bar.min.js',
    );

    // Retrieve options from WordPress
    $DeferRadioButtons = get_option('ss_defer_radio_checked');
    $exclude_list = get_option('ss_exc_defer_textarea') ? array_map('trim', explode("\n", get_option('ss_exc_defer_textarea'))) : [];
    $include_list = get_option('ss_defer_specific_textarea') ? array_map('trim', explode("\n", get_option('ss_defer_specific_textarea'))) : [];

    $exclude_support_scripts = get_option('ss_defer_exclude_support') === '1';
    $exclude_dependand_scripts = get_option('ss_defer_exclude_dependent') === '1';

    foreach ($nodes as $node) {

      if ($node->hasAttribute('defer')) {
        continue;
      }

      if ($exclude_support_scripts && $node->hasAttribute('is-support')) {
        continue;
      }

      if ($exclude_dependand_scripts && $node->hasAttribute('loading-dependency')) {
        continue;
      }

      $src = $node->getAttribute('src');
      $id = $node->getAttribute('id');
      $combinedAttributes = $src . ' ' . $id; // Combine src and id into a single string

      $shouldSkip = false; // Flag to check if current node should be skipped

      // Early skip if the string is in the never remove list
      foreach ($never_remove_list as $never_remove_item) {
        if (stripos($combinedAttributes, trim($never_remove_item)) !== false) {
          $shouldSkip = true;
        }
      }

      if ($shouldSkip) {
        continue; // Skip to the next node in the outer loop
      }

      $defer = false; // Flag to determine if the node should be removed

      if ($DeferRadioButtons === 'defer_all_scripts') {
        $defer = true;
      } elseif ($DeferRadioButtons === 'exclude_from_defer') {
        $defer = true; // Assume removal unless an exclusion match is found
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $exclude) {
            if (stripos($combinedAttributes, trim($exclude)) !== false) {
              $defer = false;
            }
          }
        }
      } elseif ($DeferRadioButtons === 'defer_specific_links') {
        if (!empty($include_list)) {
          foreach ($include_list as $include) {
            if (stripos($combinedAttributes, trim($include)) !== false) {
              $defer = true;
            }
          }
        }
      }

      if ($defer) {
        $node->setAttribute('defer', '');
      }
    }

    return $dom->saveHTML();
  }

  // End buffer for removing scripts
  public function buffer_end_defer_scripts()
  {

    if ($this->buffering_started_defer_scripts) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_defer_scripts = false;
    }
  }


  /*** DELAY SCRIPTS ***/

  // Start buffer
  public function buffer_start_delay_scripts()
  {
    if ($_SERVER['REQUEST_METHOD'] !== 'GET') {
      return;
    }

    if (get_option('ss_enable_delay_ext_js') !== '1') {
      return;
    }

    // Exit if `$post` is not a WP_Post object.
    global $post;
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (Utility::getting_images_data()) {
      return;
    }

    // If the user has enabled auto-setting Used CSS
    // If the CSS file is missing
    // If the page is new or edited
    // And if we should be getting Used CSS on this page
    if (Utility::getting_used_css_data()) {
      return;
    }

    // Pages
    $checkedDelExRadioPages = get_option('ss_delay_ext_pages_radio_checked');
    $exclude_list = get_option('ss_delay_ext_excl_pages_textarea') ? array_map('trim', explode("\n", get_option('ss_delay_ext_excl_pages_textarea'))) : [];
    $include_list = get_option('ss_delay_ext_specific_pages_textarea') ? array_map('trim', explode("\n", get_option('ss_delay_ext_specific_pages_textarea'))) : [];


    // Add home page settings checkboxses
    $exclude_home = get_option('ss_delay_js_exclude_home') === '1';
    $include_home = get_option('ss_delay_js_include_home') === '1';

    $all_pages = $checkedDelExRadioPages === 'delay_ext_all_pages_radio';
    $exclude_pages = $checkedDelExRadioPages === 'delay_ext_excl_pages_radio';
    $include_pages = $checkedDelExRadioPages === 'delay_ext_specific_pages_radio';

    $current_URL = '';
    $is_home = '';

    if (!$all_pages) {
      $is_home = Utility::is_homepage();
    }

    // Run on all pages
    if ($all_pages) {
      ob_start(array($this, "delay_scripts"));
      $this->buffering_started_delay_scripts = true;
      return;
    }

    // Run on all except these excluded pages
    elseif ($exclude_pages) {
      if ($is_home && $exclude_home) {
        // Don't run on the homepage
        $this->buffering_started_delay_scripts = false;
        return;
      } else {
        // Don't run on pages from the list
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              $this->buffering_started_delay_scripts = false;
              return;
            }
          }
        }
      }

      // Run if we are not on an excluded page
      $this->buffering_started_delay_scripts = true;
      ob_start(array($this, "delay_scripts"));
      return;
    }

    // Run only on specified pages
    elseif ($include_pages) {
      if ($is_home && $include_home) {
        // Run on Home page
        ob_start(array($this, "delay_scripts"));
        $this->buffering_started_delay_scripts = true;
        return;
      } else {
        if (!empty($include_list)) {
          foreach ($include_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              ob_start(array($this, "delay_scripts"));
              $this->buffering_started_delay_scripts = true;
              return;
            }
          }
        }
      }
    }
  }


  // Delay scritps
  public function delay_scripts($html)
  {
    
    $dom = Utility::processHTML($html);

    // Check if the returned value is an instance of DOMDocument
    if (!$dom instanceof \DOMDocument) {
      return $html;
    }

    $xpath = new \DOMXPath($dom);
    $nodes = $xpath->query("//script");

    // Exclusion list that should never be delayed
    $never_remove_list = array(
      'superspeed',
    );

    // Retrieve options from WordPress
    $checkedDelExRadioLinks = get_option('ss_delay_js_links_radio_checked');
    $exclude_list = get_option('ss_delay_ext_exclude_links_textarea') ? array_map('trim', explode("\n", get_option('ss_delay_ext_exclude_links_textarea'))) : [];
    $include_list = get_option('ss_delay_ext_specific_links_textarea') ? array_map('trim', explode("\n", get_option('ss_delay_ext_specific_links_textarea'))) : [];

    $all_strings = $checkedDelExRadioLinks === 'delay_ext_all_links_radio';
    $exclude_strings = $checkedDelExRadioLinks === 'delay_ext_exclude_links_radio';
    $include_strings = $checkedDelExRadioLinks === 'delay_ext_specific_links_radio';

    foreach ($nodes as $node) {

      $src = $node->getAttribute('src');
      $id = $node->getAttribute('id');
      $combinedAttributes = $src . ' ' . $id; // Combine src and id into a single string

      $shouldSkip = false; // Flag to check if current node should be skipped

      // Early skip if the string is in the never remove list
      foreach ($never_remove_list as $never_remove_item) {
        if (stripos($combinedAttributes, trim($never_remove_item)) !== false) {
          $shouldSkip = true;
        }
      }

      if ($shouldSkip) {
        continue; // Skip to the next node in the outer loop
      }

      $run = false; // Flag

      if ($all_strings) {
        $run = true;
      } elseif ($exclude_strings) {
        $run = true; // Assume true unless an exclusion match is found
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $exclude) {
            if (stripos($combinedAttributes, trim($exclude)) !== false) {
              $run = false;
            }
          }
        }
      } elseif ($include_strings) {
        if (!empty($include_list)) {
          foreach ($include_list as $include) {
            if (stripos($combinedAttributes, trim($include)) !== false) {
              $run = true;
            }
          }
        }
      }

      if ($run) {
        $node->removeAttribute('type');
        $node->setAttribute('type', 'ss-delay-script');
      }
    }

    return $dom->saveHTML();
  }

  // End buffer for removing scripts
  public function buffer_end_delay_scripts()
  {
    if ($this->buffering_started_delay_scripts) {
      if (ob_get_length() !== false) {
        ob_end_flush();
      }
      $this->buffering_started_delay_scripts = false;
    }
  }
}
