<?php

namespace SuperSpeed;

class Utility
{

  // Set a flag if the page or post is newly created or edited
  public static function init()
  {
    // Exit if we could not get the current URL
    $current_url = self::get_current_page_url();
    if (!$current_url) {
      return false;
    }

    // Prevent cache on first open, if we need to get background image and other info
    add_action('template_redirect', array('SuperSpeed\Utility', 'prevent_cache_on_first_open'), -9999);

    // Delete image data info from archives when a post or CPT is updated
    add_action('save_post', array('SuperSpeed\Utility', 'custom_save_post_cleanup'), 10, 2);
  }


  // Exit these pages: Backend, Elementor editor, WordPress editor
  public static function is_page_editor_or_wp_backend()
  {
    // Check if we are in the admin area
    if (is_admin()) {
      return true;
    }

    if (defined('ELEMENTOR_VERSION') && isset($_GET['elementor-preview'])) {
      return true;
    }

    // These pages are the WordPress post editor
    global $pagenow;
    if (in_array($pagenow, ['post.php', 'post-new.php'], true)) {
      return true;
    }

    // Construct the current URL
    $scheme = filter_input(INPUT_SERVER, 'REQUEST_SCHEME', FILTER_SANITIZE_URL) ?? 'http';
    $host = filter_input(INPUT_SERVER, 'HTTP_HOST', FILTER_SANITIZE_URL);
    $requestUri = filter_input(INPUT_SERVER, 'REQUEST_URI', FILTER_SANITIZE_URL);

    // Return false if required server parameters are missing
    if ($host === null || $requestUri === null) {
      return false;
    }

    $current_url = $scheme . "://" . $host . $requestUri;

    // Define excluded items
    $excluded_items = ['divi', 'wp-admin', 'action=elementor'];

    // Check if the URL contains any excluded items
    foreach ($excluded_items as $item) {
      if (stripos($current_url, $item) !== false) {
        return true;
      }
    }

    return false;
  }


  // Prevents caching for a post when it is first opened, if the page is new or has been edited
  //On second open, we use the collected data
  public static function prevent_cache_on_first_open()
  {
    if (self::is_page_editor_or_wp_backend()) {
      return false; // Return false to indicate no update is necessary
    }

    // Exit if we could not get the current URL
    $current_url = self::get_current_page_url();
    if (!$current_url) {
      return false;
    }

    global $post;

    if (!is_a($post, 'WP_Post')) {
      if (!defined('DONOTCACHEPAGE')) define('DONOTCACHEPAGE', false);
      return;
    }

    // If the user has enabled auto-setting LCP images
    // If the data is missing
    // If the page is new or edited
    // And if we should be getting image data on this page
    if (
      self::getting_images_data() ||

      // If the user has enabled auto-setting Used CSS
      // If the CSS file is missing
      // If the page is new or edited
      // And if we should be getting Used CSS on this page
      (self::getting_used_css_data())
    ) {

      header("Cache-Control: no-cache, must-revalidate, max-age=0");
      header("Pragma: no-cache");
      header("Expires: Wed, 11 Jan 1984 05:00:00 GMT");

      if (!defined('DONOTCACHEPAGE')) define('DONOTCACHEPAGE', true);
    } else {
      if (!defined('DONOTCACHEPAGE')) define('DONOTCACHEPAGE', false);
    }
  }


  public static function get_current_page_url()
  {
    $uri = $_SERVER['REQUEST_URI'];
    $parsed_url = wp_parse_url($uri);
    if (!$parsed_url) {
      return false;
    }

    $path = $parsed_url['path'] ?? '';  // Use null coalescing operator to handle absence of 'path'

    if (preg_match('/\.(jpg|jpeg|png|gif|svg|ico|css|js|woff2|woff|ttf|map)$/i', $path)) {
      // error_log("Static resource detected, exiting. Path: $path");
      return false;  // Ensure this return is enabled if no URL is needed for static resources
    }

    $protocol = is_ssl() ? 'https' : 'http';
    $host = $_SERVER['HTTP_HOST'];
    $baseUrl = $protocol . '://' . $host . $path;

    return $baseUrl ?: false;  // Return false if $baseUrl is empty
  }


  // For normalising the Home page, we remove the query, hashtag, and slash
  public static function is_homepage()
  {
    $current_url = self::get_current_page_url();
    $home_url = home_url('/');

    // Parse URLs to components
    $parsed_current_url = parse_url($current_url);
    $parsed_home_url = parse_url($home_url);

    // Normalize the path by removing the trailing slash if present
    $current_path = rtrim($parsed_current_url['path'] ?? '/', '/');
    $home_path = rtrim($parsed_home_url['path'] ?? '/', '/');

    // Compare paths only for homepage check, ignore query and fragment
    return $current_path === $home_path;
  }


  // Helper function to normalizes a URL for comparison
  public static function normalizeUrlString($url)
  {
    // Normalize the URL: remove protocol, query string, hash, and trailing slash
    $normalizedUrl = preg_replace(['#^https?://#', '/[?#].*$/', '#/$#'], ['', '', ''], $url);
    return $normalizedUrl;
  }

  /********** */

  // Process and load HTML content, returning original content on failure.
  public static function processHTML($html)
  {
    // Check for null or effectively empty HTML
    if ($html === null || empty(trim($html))) {
      return null;
    }

    // Ensure the input is a proper string
    if (!is_string($html)) {
      error_log("Invalid HTML content type");
      return null;
    }

    // Optional: Check for the presence of necessary HTML structure if required
    if (stripos($html, '<html') === false) {
      error_log("Incomplete HTML content");
      return null;
    }

    // Load HTML into DOMDocument
    $dom = new \DOMDocument();
    libxml_use_internal_errors(true);
    if (!$dom->loadHTML($html, LIBXML_HTML_NOIMPLIED | LIBXML_HTML_NODEFDTD)) {
      $errors = libxml_get_errors();
      foreach ($errors as $error) {
        error_log("HTML Load Error: " . $error->message);
      }
      libxml_clear_errors();
      libxml_use_internal_errors(false);
      return null;
    }

    libxml_clear_errors();
    libxml_use_internal_errors(false);
    return $dom;
  }

  /********** */

  /*** GETTING IMAGES DATA ***/

  // Used to stop functions that are getting in the way of properly getting the background images
  public static function get_images_info_enabled()
  {
    // Exit if we could not get the current URL
    $current_url = self::get_current_page_url();
    if (!$current_url) {
      return false;
    }

    if (get_option('image_data_collection') === '1') {
      return true;
    } else {
      return false;
    }
  }

  // Read Post Saved flag state for background data
  public static function image_data_is_empty()
  {

    // Exit if the option is not in use
    if (!self::get_images_info_enabled()) {
      return false;
    }

    $current_url = self::get_current_page_url();

    $option_name = 'images_data_for_page_' . hash('sha256', $current_url);

    // Set flag to true if the data is empty
    if (empty(get_option($option_name))) {
      return true;
    }

    return false;
  }


  // Check if we should get images info on this page
  public static function get_images_info_on_this_page()
  {

    // Exit if the option is not in use
    if (!self::get_images_info_enabled()) {
      return false;
    }

    $exclude_home = get_option('image_data_collection_exclude_home') === '1';
    $exclude_pages = get_option('image_data_collection_exclude_pages_cbx') === '1';
    $exclude_list = get_option('image_data_collection_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('image_data_collection_exclude_pages_textarea'))) : [];

    $exclude_posts = get_option('image_data_collection_exclude_posts') === '1';
    $exclude_products = get_option('image_data_collection_exclude_products') === '1';
    $exclude_cpts = get_option('image_data_collection_exclude_cpts') === '1';
    $exclude_cpt_list = get_option('image_data_collection_exclude_cpts_textarea') ? array_map('trim', explode("\n", get_option('image_data_collection_exclude_cpts_textarea'))) : [];

    if ($exclude_home) {
      $is_home = self::is_homepage();
      if ($is_home) {
        return false;
      }
    }

    // Run on all except these excluded pages
    elseif ($exclude_pages) {
      // Don't run on pages from the list
      if (!empty($exclude_list)) {
        $current_Url = self::get_current_page_url();
        foreach ($exclude_list as $string) {
          $normalizedString = self::normalizeUrlString($string);
          if (stripos($current_Url, $normalizedString) !== false) {
            return false;
          }
        }
      }
    }

    // Do not run on single blog posts
    elseif ($exclude_posts && is_single() && get_post_type() === 'post') {
      return false;
    }

    // Do not run on single product pages
    elseif ($exclude_products && is_singular('product')) {
      return false;
    }

    // Excluding by CPT ingle post
    // We don't need a foreach loop because the is_singular() function can accept an array of post type names directly
    elseif ($exclude_cpts && !empty($exclude_cpt_list) && is_singular($exclude_cpt_list)) {
      return false;
    }

    return true;
  }


  // Cache the result
  // If the user has enabled auto-setting LCP images
  // If the data is missing
  // If the page is new or edited
  // And if we should be getting image data on this page
  static function getting_images_data()
  {
    global $post;
    // Exit if `$post` is not a WP_Post object.
    if (!is_a($post, 'WP_Post')) {
      return;
    }

    // Static cache variable to store results keyed by post ID
    static $cache = [];

    // Check if result is already cached for this post ID
    if (isset($cache[$post->ID])) {
      return $cache[$post->ID];
    }

    // Perform the checks and cache the result
    $cache[$post->ID] = self::get_images_info_enabled() &&
      self::image_data_is_empty($post->ID) &&
      self::get_images_info_on_this_page();

    return $cache[$post->ID];
  }


  // Delete image data from archives when a post or CPT is updated
  public static function custom_save_post_cleanup($post_ID, $post)
  {

    // Exclude pages and check for revisions or auto-drafts
    if (get_post_type($post_ID) !== 'page' && !wp_is_post_revision($post_ID) && $post->post_status !== 'auto-draft') {
      self::clear_related_post_data($post_ID);

      // Check if the home page displays posts
      if ('posts' === get_option('show_on_front')) {
        self::clear_home_page_data_and_cache();
      }
    }
  }

  // Helper functions that finds on which archive and paginated page the post is displayed, and clears data from the database
  public static function clear_related_post_data($post_ID)
  {

    global $wpdb;
    // This is for options, but you might target postmeta or other tables
    $pattern = $wpdb->esc_like('data_related_to_post_' . $post_ID) . '%';
    // Since the data is stored in options (and not a custom table)
    $options = $wpdb->get_col("SELECT option_name FROM $wpdb->options WHERE option_name LIKE '$pattern'");

    foreach ($options as $option) {
      delete_option($option);
    }
  }

  // If the home page displays posts, delete image data and clear cache from some plugins
  protected static function clear_home_page_data_and_cache()
  {

    // Normalize the home URL by removing any query strings
    $home_url = trailingslashit(home_url());

    // Ensure the URL does not end with file extensions specified
    // This may not be strictly necessary for the home URL, but it ensures consistency
    if (!preg_match('/\.(jpg|jpeg|png|gif|ico|svg|css|js|woff|woff2|ttf|eot|otf|pdf)(\?.*)?$/i', $home_url)) {
      $option_name = 'images_data_for_page_' . hash('sha256', $home_url);

      // Delete the specific option related to the home page
      delete_option($option_name);
    }

    // Home page identifier
    $home_page_id = get_option('page_on_front');

    global $wpdb;

    if ($home_page_id) {
      // If your data relates specifically to posts displayed on the home page, adjust the pattern accordingly
      $pattern = $wpdb->esc_like('data_related_to_home_page') . '%'; // Adjust based on how data is stored
      $options = $wpdb->get_col("SELECT option_name FROM $wpdb->options WHERE option_name LIKE '$pattern'");

      foreach ($options as $option) {
        delete_option($option);
      }
    }

    // W3 Total Cache
    if (function_exists('w3tc_flush_post')) {
      w3tc_flush_post($home_page_id);
    }

    // WP Super Cache
    if (function_exists('wpsc_delete_post_cache')) {
      wpsc_delete_post_cache($home_page_id);
    }

    // WP-Optimize
    if (class_exists('WPO_Page_Cache')) {
      WPO_Page_Cache::delete_single_post_cache($home_page_id);
    }

    // LiteSpeed Cache
    if (function_exists('litespeed_purge_post')) {
      do_action('litespeed_purge_post', $home_page_id);
    }

    // WP Fastest Cache
    if (function_exists('wpfc_clear_post_cache_by_id')) {
      wpfc_clear_post_cache_by_id($home_page_id);
    }

    // Comet Cache
    if (class_exists('comet_cache')) {
      comet_cache::clear($home_page_id);
    }

    // Cache Enabler
    if (function_exists('cache_enabler_clear_home_page_cache')) {
      cache_enabler_clear_home_page_cache();
    }
  }


  /*** USED CSS ***/

  // Is the Used CSS option in use
  public static function get_used_css_enabled()
  {
    if (get_option('enable_used_css') === '1') {
      return true;
    } else {
      return false;
    }
  }


  public static function used_css_file_is_missing($post_id)
  {
    // Exit if the option is not in use
    if (!self::get_used_css_enabled()) {
      return false;
    }

    $template_type = get_post_meta($post_id, 'post_edited_used_css', true);

    if ($template_type !== null) {
      $upload_dir = wp_upload_dir();
      $file_path = $upload_dir['basedir'] . '/superspeed-used-css/' . $template_type . '.css';

      if (file_exists($file_path)) {
        return false;
      }
    }

    return true;
  }

  // Check if we should use Critical CSS on this page
  static function getting_used_css_on_page()
  {
    // Exit if the option is not in use
    if (!self::get_used_css_enabled()) {

      return false;
    }

    $exclude_pages = null;
    $exclude_home = null;
    $exclude_list = null;
    $include_pages = null;
    $include_home = null;
    $include_list = null;

    $loadUsedCssOnPagesRadio = get_option('load_used_css_on_pages_radio');
    $all_pages = $loadUsedCssOnPagesRadio === 'used_css_load_all_pages';
    if (!$all_pages) {
      $exclude_pages = $loadUsedCssOnPagesRadio === 'used_css_load_exclude_pages';
      if ($exclude_pages) {
        $exclude_home = get_option('used_css_exclude_home') === '1';
        $exclude_list = get_option('used_css_exclude_pages_textarea') ? array_map('trim', explode("\n", get_option('used_css_exclude_pages_textarea'))) : [];
      }
      $include_pages = $loadUsedCssOnPagesRadio === 'used_css_load_specific_pages';
      if ($include_pages) {
        $include_home = get_option('used_css_include_home') === '1';
        $include_list = get_option('used_css_spacific_pages_textarea') ? array_map('trim', explode("\n", get_option('used_css_spacific_pages_textarea'))) : [];
      }
    }

    $current_URL = '';
    $is_home = '';

    if (!$all_pages) {
      $is_home = Utility::is_homepage();
    }

    // Run on all pages
    if ($all_pages) {
      return true;
    }

    // Run on all except these excluded pages
    elseif ($exclude_pages) {
      if ($is_home && $exclude_home) {
        // Don't run on the homepage
        if ($exclude_home && Utility::is_homepage()) {
          return false;
        }
      } else {
        // Don't run on pages from the list
        if (!empty($exclude_list)) {
          foreach ($exclude_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              return false;
            }
          }
        }
      }

      // Run if we are not on an excluded page
      return true;
    }

    // Run only on specified pages
    elseif ($include_pages) {
      if ($is_home && $include_home) {
        // Run on Home page
        if ($include_home && Utility::is_homepage()) {
          return true;
        }
      } else {
        if (!empty($include_list)) {
          foreach ($include_list as $string) {
            $normalizedString = Utility::normalizeUrlString($string);
            if (stripos($current_URL, $normalizedString) !== false) {
              return true;
            }
          }
        }
      }
    }
  }


  // If the user has enabled auto-setting Used CSS
  // If the CSS file is missing
  // If the page is new or edited
  // And if we should be getting Used CSS on this page
  static function getting_used_css_data()
  {
    global $post;

    // Directly return the result of the checks
    return self::get_used_css_enabled() &&
      self::used_css_file_is_missing($post->ID) &&
      self::getting_used_css_on_page();
  }


  // Excluding or including a page, we strip the domain, query, and compare only the core path
  static function normalizeURLPath($url)
  {
    // Parse the URL to extract components
    $parsedUrl = parse_url($url);

    // Extract the path and ensure it does not have leading/trailing slashes
    $path = isset($parsedUrl['path']) ? trim($parsedUrl['path'], '/') : '';

    return $path;
  }
}
